@@ -1,4 +1,4 @@
-module ::Kernel
+class ::Object
   SIGNALS = {
     1  => "SIGHUP", 2  => "SIGINT", 3  => "SIGQUIT", 4  => "SIGILL", 5  => "SIGTRAP", 6  => "SIGABRT",
     7  => "SIGPOLL", 8  => "SIGFPE", 9  => "SIGKILL", 10 => "SIGBUS", 11 => "SIGSEGV", 12 => "SIGSYS",
@@ -6,30 +6,26 @@ module ::Kernel
     19 => "SIGCONT", 20 => "SIGCHLD", 21 => "SIGTTIN", 22 => "SIGTTOU", 24 => "SIGXCPU", 25 => "SIGXFSZ",
     26 => "SIGVTALRM", 27 => "SIGPROF", 30 => "SIGUSR1", 31 => "SIGUSR2"
   }
+  
   begin
+    JavaSignal = Java::sun.misc.Signal
     def __jtrap(*args, &block)
-      sig = args.first
+      sig = args.shift
       sig = SIGNALS[sig] if sig.kind_of?(Fixnum)
       sig = sig.to_s.sub(/^SIG(.+)/,'\1')
-      signal_class = Java::sun.misc.Signal
-      signal_class.send :attr_accessor, :prev_handler
-      signal_object = signal_class.new(sig) rescue nil
+
+      block = args.shift unless args.empty?
+
+      signal_object = JavaSignal.new(sig) rescue nil
       return unless signal_object
-      signal_handler = Java::sun.misc.SignalHandler.impl do
-        begin
-          block.call
-        rescue Exception => e
-          Thread.main.raise(e) rescue nil
-        ensure
-          # re-register the handler
-          signal_class.handle(signal_object, signal_handler)
-        end
-      end
-      signal_object.prev_handler = signal_class.handle(signal_object, signal_handler)
+      
+      Signal::__jtrap_kernel(block, signal_object, sig)
+    rescue Exception
+      warn "The signal #{sig} is in use by the JVM and will not work correctly on this platform"
     end
   rescue NameError
     def __jtrap(*args, &block)
-      warn "trap not supported by this VM"
+      warn "trap not supported or not allowed by this VM"
     end
   end
 end
