@@ -36,10 +36,9 @@ package org.jruby;
 import java.io.IOException;
 import java.math.BigDecimal;
 import java.math.BigInteger;
-import org.jruby.runtime.Arity;
-import org.jruby.runtime.Block;
-import org.jruby.runtime.CallType;
-import org.jruby.runtime.CallbackFactory;
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
+import org.jruby.common.IRubyWarnings.ID;
 import org.jruby.runtime.ClassIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
@@ -51,42 +50,15 @@ import org.jruby.runtime.marshal.UnmarshalStream;
  *
  * @author  jpetersen
  */
+@JRubyClass(name="Bignum", parent="Integer")
 public class RubyBignum extends RubyInteger {
     public static RubyClass createBignumClass(Ruby runtime) {
-        RubyClass bignum = runtime.defineClass("Bignum", runtime.getClass("Integer"),
+        RubyClass bignum = runtime.defineClass("Bignum", runtime.getInteger(),
                 ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
+        runtime.setBignum(bignum);
         bignum.index = ClassIndex.BIGNUM;
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyBignum.class);
-
-        bignum.defineFastMethod("to_s", callbackFactory.getFastOptMethod("to_s"));
-        bignum.defineFastMethod("coerce", callbackFactory.getFastMethod("coerce", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("-@", callbackFactory.getFastMethod("uminus"));
-        bignum.defineFastMethod("+", callbackFactory.getFastMethod("plus", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("-", callbackFactory.getFastMethod("minus", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("*", callbackFactory.getFastMethod("mul", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("/", callbackFactory.getFastMethod("div", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("%", callbackFactory.getFastMethod("mod", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("div", callbackFactory.getFastMethod("div", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("divmod", callbackFactory.getFastMethod("divmod", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("modulo", callbackFactory.getFastMethod("mod", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("remainder", callbackFactory.getFastMethod("remainder", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("quo", callbackFactory.getFastMethod("quo", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("**", callbackFactory.getFastMethod("pow", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("&", callbackFactory.getFastMethod("and", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("|", callbackFactory.getFastMethod("or", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("^", callbackFactory.getFastMethod("xor", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("~", callbackFactory.getFastMethod("neg"));
-        bignum.defineFastMethod("<<", callbackFactory.getFastMethod("lshift", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod(">>", callbackFactory.getFastMethod("rshift", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("[]", callbackFactory.getFastMethod("aref", RubyKernel.IRUBY_OBJECT));
-
-        bignum.defineFastMethod("<=>", callbackFactory.getFastMethod("cmp", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("==", callbackFactory.getFastMethod("equal", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("eql?", callbackFactory.getFastMethod("eql_p", RubyKernel.IRUBY_OBJECT));
-        bignum.defineFastMethod("hash", callbackFactory.getFastMethod("hash"));
-        bignum.defineFastMethod("to_f", callbackFactory.getFastMethod("to_f"));
-        bignum.defineFastMethod("abs", callbackFactory.getFastMethod("abs"));
-        bignum.defineFastMethod("size", callbackFactory.getFastMethod("size"));
+        
+        bignum.defineAnnotatedMethods(RubyBignum.class);
 
         return bignum;
     }
@@ -95,66 +67,21 @@ public class RubyBignum extends RubyInteger {
     private static final long MAX = (1L << (BIT_SIZE - 1)) - 1;
     private static final BigInteger LONG_MAX = BigInteger.valueOf(MAX);
     private static final BigInteger LONG_MIN = BigInteger.valueOf(-MAX - 1);
-    
-    public static final byte OP_PLUS_SWITCHVALUE = 1;
-    public static final byte OP_MINUS_SWITCHVALUE = 2;
-    public static final byte OP_LT_SWITCHVALUE = 3;
-    public static final byte TO_S_SWITCHVALUE = 4;
-    public static final byte TO_I_SWITCHVALUE = 5;
-    public static final byte HASH_SWITCHVALUE = 6;
-    public static final byte OP_TIMES_SWITCHVALUE = 7;
-    public static final byte EQUALEQUAL_SWITCHVALUE = 8;
-    public static final byte OP_SPACESHIP_SWITCHVALUE = 9;
-    public static final byte INSPECT_SWITCHVALUE = 10;
 
     private final BigInteger value;
 
     public RubyBignum(Ruby runtime, BigInteger value) {
-        super(runtime, runtime.getClass("Bignum"));
+        super(runtime, runtime.getBignum());
         this.value = value;
     }
     
-    public IRubyObject callMethod(ThreadContext context, RubyModule rubyclass, int methodIndex, String name,
-            IRubyObject[] args, CallType callType, Block block) {
-        // If tracing is on, don't do STI dispatch
-        if (context.getRuntime().hasEventHooks()) return super.callMethod(context, rubyclass, name, args, callType, block);
-        
-        switch (getRuntime().getSelectorTable().table[rubyclass.index][methodIndex]) {
-        case OP_PLUS_SWITCHVALUE:
-            if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
-            return plus(args[0]);
-        case OP_MINUS_SWITCHVALUE:
-            if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
-            return minus(args[0]);
-        case TO_S_SWITCHVALUE:
-            return to_s(args);
-        case TO_I_SWITCHVALUE:
-            if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
-            return to_i();
-        case HASH_SWITCHVALUE:
-            if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
-            return hash();
-        case OP_TIMES_SWITCHVALUE:
-            if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
-            return mul(args[0]);
-        case EQUALEQUAL_SWITCHVALUE:
-            if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
-            return equal(args[0]);
-        case OP_SPACESHIP_SWITCHVALUE:
-            if (args.length != 1) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 1 + ")");
-            return cmp(args[0]);
-        case INSPECT_SWITCHVALUE:
-            if (args.length != 0) throw context.getRuntime().newArgumentError("wrong number of arguments(" + args.length + " for " + 0 + ")");
-            return inspect();
-        case 0:
-        default:
-            return super.callMethod(context, rubyclass, name, args, callType, block);
-        }
-    }
-    
     public int getNativeTypeIndex() {
         return ClassIndex.BIGNUM;
     }
+    
+    public Class<?> getJavaClass() {
+        return BigInteger.class;
+    }
 
     public static RubyBignum newBignum(Ruby runtime, long value) {
         return newBignum(runtime, BigInteger.valueOf(value));
@@ -211,22 +138,47 @@ public class RubyBignum extends RubyInteger {
 
         if (big.compareTo(LONG_MIN) < 0 || big.compareTo(LONG_MAX) > 0) {
             throw value.getRuntime().newRangeError("bignum too big to convert into `long'");
-    }
-        return big.longValue();
         }
+        return big.longValue();
+    }
 
     /** rb_big2dbl
      * 
      */
     public static double big2dbl(RubyBignum value) {
         BigInteger big = value.getValue();
-        double dbl = big.doubleValue();
+        double dbl = convertToDouble(big);
         if (dbl == Double.NEGATIVE_INFINITY || dbl == Double.POSITIVE_INFINITY) {
-            value.getRuntime().getWarnings().warn("Bignum out of Float range");
+            value.getRuntime().getWarnings().warn(ID.BIGNUM_FROM_FLOAT_RANGE, "Bignum out of Float range");
     }
         return dbl;
     }
+    
+    private IRubyObject checkShiftDown(RubyBignum other) {
+        if (other.value.signum() == 0) return RubyFixnum.zero(getRuntime()); 
+        if (value.compareTo(LONG_MIN) < 0 || value.compareTo(LONG_MAX) > 0) {
+            return other.value.signum() >= 0 ? RubyFixnum.zero(getRuntime()) : RubyFixnum.minus_one(getRuntime());
+        }
+        return getRuntime().getNil();
+    }
 
+    /**
+     * BigInteger#doubleValue is _really_ slow currently.
+     * This is faster, and mostly correct (?)
+     */
+    static double convertToDouble(BigInteger bigint) {
+        byte[] arr = bigint.toByteArray();
+        double res = 0;
+        double acc = 1;
+        for (int i = arr.length - 1; i > 0 ; i--)
+        {
+            res += (double) (arr[i] & 0xff) * acc;
+            acc *= 256;
+        }
+        res += (double) arr[0] * acc; // final byte sign is significant
+        return res;
+    }
+    
     /** rb_int2big
      * 
      */
@@ -242,8 +194,8 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_to_s
      * 
      */
+    @JRubyMethod(name = "to_s", optional = 1)
     public IRubyObject to_s(IRubyObject[] args) {
-        Arity.checkArgumentCount(getRuntime(), args, 0, 1);
         int base = args.length == 0 ? 10 : num2int(args[0]);
         if (base < 2 || base > 36) {
             throw getRuntime().newArgumentError("illegal radix " + base);
@@ -254,6 +206,7 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_coerce
      * 
      */
+    @JRubyMethod(name = "coerce", required = 1)
     public IRubyObject coerce(IRubyObject other) {
         if (other instanceof RubyFixnum) {
             return getRuntime().newArray(newBignum(getRuntime(), ((RubyFixnum) other).getLongValue()), this);
@@ -267,44 +220,78 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_uminus
      * 
      */
-    public IRubyObject uminus() {
+    @JRubyMethod(name = "-@")
+    public IRubyObject op_uminus() {
         return bignorm(getRuntime(), value.negate());
     }
 
     /** rb_big_plus
      * 
      */
-    public IRubyObject plus(IRubyObject other) {
+    @JRubyMethod(name = "+", required = 1)
+    public IRubyObject op_plus(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return bignorm(getRuntime(), value.add(fix2big(((RubyFixnum) other))));
-        }
-        if (other instanceof RubyBignum) {
-            return bignorm(getRuntime(), value.add(((RubyBignum) other).value));
+            return addFixnum((RubyFixnum)other);
+        } else if (other instanceof RubyBignum) {
+            return addBignum((RubyBignum)other);
         } else if (other instanceof RubyFloat) {
-            return RubyFloat.newFloat(getRuntime(), big2dbl(this) + ((RubyFloat) other).getDoubleValue());
+            return addFloat((RubyFloat)other);
+        }
+        return addOther(context, other);
     }
-        return coerceBin("+", other);
+    
+    private IRubyObject addFixnum(RubyFixnum other) {
+        return bignorm(getRuntime(), value.add(fix2big(other)));
+    }
+    
+    private IRubyObject addBignum(RubyBignum other) {
+        return bignorm(getRuntime(), value.add(other.value));
+    }
+    
+    private IRubyObject addFloat(RubyFloat other) {
+        return RubyFloat.newFloat(getRuntime(), big2dbl(this) + other.getDoubleValue());
+    }
+    
+    private IRubyObject addOther(ThreadContext context, IRubyObject other) {
+        return coerceBin(context, "+", other);
     }
 
     /** rb_big_minus
      * 
      */
-    public IRubyObject minus(IRubyObject other) {
+    @JRubyMethod(name = "-", required = 1)
+    public IRubyObject op_minus(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
-            return bignorm(getRuntime(), value.subtract(fix2big(((RubyFixnum) other))));
-    }
-        if (other instanceof RubyBignum) {
-            return bignorm(getRuntime(), value.subtract(((RubyBignum) other).value));
+            return subtractFixnum((RubyFixnum)other);
+        } else if (other instanceof RubyBignum) {
+            return subtractBignum((RubyBignum)other);
         } else if (other instanceof RubyFloat) {
-            return RubyFloat.newFloat(getRuntime(), big2dbl(this) - ((RubyFloat) other).getDoubleValue());
+            return subtractFloat((RubyFloat)other);
+        }
+        return subtractOther(context, other);
     }
-        return coerceBin("-", other);
+    
+    private IRubyObject subtractFixnum(RubyFixnum other) {
+        return bignorm(getRuntime(), value.subtract(fix2big(((RubyFixnum) other))));
+    }
+    
+    private IRubyObject subtractBignum(RubyBignum other) {
+        return bignorm(getRuntime(), value.subtract(((RubyBignum) other).value));
+    }
+    
+    private IRubyObject subtractFloat(RubyFloat other) {
+        return RubyFloat.newFloat(getRuntime(), big2dbl(this) - ((RubyFloat) other).getDoubleValue());
+    }
+    
+    private IRubyObject subtractOther(ThreadContext context, IRubyObject other) {
+        return coerceBin(context, "-", other);
     }
 
     /** rb_big_mul
      * 
      */
-    public IRubyObject mul(IRubyObject other) {
+    @JRubyMethod(name = "*", required = 1)
+    public IRubyObject op_mul(ThreadContext context, IRubyObject other) {
         if (other instanceof RubyFixnum) {
             return bignorm(getRuntime(), value.multiply(fix2big(((RubyFixnum) other))));
         }
@@ -313,22 +300,30 @@ public class RubyBignum extends RubyInteger {
         } else if (other instanceof RubyFloat) {
             return RubyFloat.newFloat(getRuntime(), big2dbl(this) * ((RubyFloat) other).getDoubleValue());
         }
-        return coerceBin("*", other);
+        return coerceBin(context, "*", other);
     }
 
-    /** rb_big_div
-     * 
+    /**
+     * rb_big_divide. Shared part for both "/" and "div" operations.
      */
-    public IRubyObject div(IRubyObject other) {
+    private IRubyObject op_divide(ThreadContext context, IRubyObject other, String op) {
+        assert ("/".equals(op) || "div".equals(op));
+
         final BigInteger otherValue;
         if (other instanceof RubyFixnum) {
             otherValue = fix2big((RubyFixnum) other);
         } else if (other instanceof RubyBignum) {
             otherValue = ((RubyBignum) other).value;
         } else if (other instanceof RubyFloat) {
-            return RubyFloat.newFloat(getRuntime(), big2dbl(this) / ((RubyFloat) other).getDoubleValue());
+            double div = big2dbl(this) / ((RubyFloat) other).getDoubleValue();
+            if ("/".equals(op)) {
+                return RubyFloat.newFloat(getRuntime(),
+                        big2dbl(this) / ((RubyFloat) other).getDoubleValue());
+            } else {
+                return RubyNumeric.dbl2num(getRuntime(), div);
+            }
         } else {
-            return coerceBin("/", other);
+            return coerceBin(context, op, other);
         }
 
         if (otherValue.equals(BigInteger.ZERO)) {
@@ -337,23 +332,40 @@ public class RubyBignum extends RubyInteger {
 
         BigInteger[] results = value.divideAndRemainder(otherValue);
 
-        if (results[0].signum() == -1 && results[1].signum() != 0) {
+        if ((value.signum() * otherValue.signum()) == -1 && results[1].signum() != 0) {
             return bignorm(getRuntime(), results[0].subtract(BigInteger.ONE));
         }
         return bignorm(getRuntime(), results[0]);
     }
 
+    /** rb_big_div
+     *
+     */
+    @JRubyMethod(name = {"/"}, required = 1)
+    public IRubyObject op_div(ThreadContext context, IRubyObject other) {
+        return op_divide(context, other, "/");
+    }
+
+    /** rb_big_idiv
+     *
+     */
+    @JRubyMethod(name = {"div"}, required = 1)
+    public IRubyObject op_idiv(ThreadContext context, IRubyObject other) {
+        return op_divide(context, other, "div");
+    }
+
     /** rb_big_divmod
      * 
      */
-    public IRubyObject divmod(IRubyObject other) {
+    @JRubyMethod(name = "divmod", required = 1)
+    public IRubyObject divmod(ThreadContext context, IRubyObject other) {
         final BigInteger otherValue;
         if (other instanceof RubyFixnum) {
             otherValue = fix2big((RubyFixnum) other);
         } else if (other instanceof RubyBignum) {
             otherValue = ((RubyBignum) other).value;
         } else {
-            return coerceBin("divmod", other);
+            return coerceBin(context, "divmod", other);
         }
 
         if (otherValue.equals(BigInteger.ZERO)) {
@@ -362,8 +374,9 @@ public class RubyBignum extends RubyInteger {
 
         BigInteger[] results = value.divideAndRemainder(otherValue);
 
-        if (results[0].signum() == -1 && results[1].signum() != 0) {
-            return bignorm(getRuntime(), results[0].subtract(BigInteger.ONE));
+        if ((value.signum() * otherValue.signum()) == -1 && results[1].signum() != 0) {
+            results[0] = results[0].subtract(BigInteger.ONE);
+            results[1] = otherValue.add(results[1]);
     	}
         final Ruby runtime = getRuntime();
         return RubyArray.newArray(getRuntime(), bignorm(runtime, results[0]), bignorm(runtime, results[1]));
@@ -372,20 +385,21 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_modulo
      * 
      */
-    public IRubyObject mod(IRubyObject other) {
+    @JRubyMethod(name = {"%", "modulo"}, required = 1)
+    public IRubyObject op_mod(ThreadContext context, IRubyObject other) {
         final BigInteger otherValue;
         if (other instanceof RubyFixnum) {
             otherValue = fix2big((RubyFixnum) other);
         } else if (other instanceof RubyBignum) {
             otherValue = ((RubyBignum) other).value;
         } else {
-            return coerceBin("%", other);
+            return coerceBin(context, "%", other);
         }
         if (otherValue.equals(BigInteger.ZERO)) {
             throw getRuntime().newZeroDivisionError();
         }
         BigInteger result = value.mod(otherValue.abs());
-        if (otherValue.signum() == -1) {
+        if (otherValue.signum() == -1 && result.signum() != 0) {
             result = otherValue.add(result);
         }
         return bignorm(getRuntime(), result);
@@ -395,14 +409,15 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_remainder
      * 
      */
-    public IRubyObject remainder(IRubyObject other) {
+    @JRubyMethod(name = "remainder", required = 1)
+    public IRubyObject remainder(ThreadContext context, IRubyObject other) {
         final BigInteger otherValue;
         if (other instanceof RubyFixnum) {
             otherValue = fix2big(((RubyFixnum) other));
         } else if (other instanceof RubyBignum) {
             otherValue = ((RubyBignum) other).value;
         } else {
-            return coerceBin("remainder", other);
+            return coerceBin(context, "remainder", other);
         }
         if (otherValue.equals(BigInteger.ZERO)) {
             throw getRuntime().newZeroDivisionError();
@@ -414,60 +429,100 @@ public class RubyBignum extends RubyInteger {
 
      * 
      */
-    public IRubyObject quo(IRubyObject other) {
+    @JRubyMethod(name = "quo", required = 1)
+    public IRubyObject quo(ThreadContext context, IRubyObject other) {
     	if (other instanceof RubyNumeric) {
             return RubyFloat.newFloat(getRuntime(), big2dbl(this) / ((RubyNumeric) other).getDoubleValue());
         } else {
-            return coerceBin("quo", other);
+            return coerceBin(context, "quo", other);
     	}
     }
 
     /** rb_big_pow
      * 
      */
-    public IRubyObject pow(IRubyObject other) {
+    @JRubyMethod(name = {"**", "power"}, required = 1)
+    public IRubyObject op_pow(ThreadContext context, IRubyObject other) {
         double d;
         if (other instanceof RubyFixnum) {
             RubyFixnum fix = (RubyFixnum) other;
             long fixValue = fix.getLongValue();
             // MRI issuses warning here on (RBIGNUM(x)->len * SIZEOF_BDIGITS * yy > 1024*1024)
             if (((value.bitLength() + 7) / 8) * 4 * Math.abs(fixValue) > 1024 * 1024) {
-                getRuntime().getWarnings().warn("in a**b, b may be too big");
-    	}
+                getRuntime().getWarnings().warn(ID.MAY_BE_TOO_BIG, "in a**b, b may be too big", fixValue);
+            }
             if (fixValue >= 0) {
                 return bignorm(getRuntime(), value.pow((int) fixValue)); // num2int is also implemented
             } else {
                 return RubyFloat.newFloat(getRuntime(), Math.pow(big2dbl(this), (double)fixValue));
             }
         } else if (other instanceof RubyBignum) {
-            getRuntime().getWarnings().warn("in a**b, b may be too big");
             d = ((RubyBignum) other).getDoubleValue();
+            getRuntime().getWarnings().warn(ID.MAY_BE_TOO_BIG, "in a**b, b may be too big", d);
         } else if (other instanceof RubyFloat) {
             d = ((RubyFloat) other).getDoubleValue();
         } else {
-            return coerceBin("**", other);
+            return coerceBin(context, "**", other);
 
-    }
+        }
         return RubyFloat.newFloat(getRuntime(), Math.pow(big2dbl(this), d));
     }
 
+    /** rb_big_pow
+     * 
+     */
+    @JRubyMethod(name = {"**", "power"}, required = 1, compat = CompatVersion.RUBY1_9)
+    public IRubyObject op_pow_19(ThreadContext context, IRubyObject other) {
+        Ruby runtime = context.getRuntime();
+        if (other == RubyFixnum.zero(runtime)) return RubyFixnum.one(runtime);
+        double d;
+        if (other instanceof RubyFixnum) {
+            RubyFixnum fix = (RubyFixnum) other;
+            long fixValue = fix.getLongValue();
+            
+            if (fixValue < 0) {
+                return RubyRational.newRationalRaw(runtime, this).callMethod(context, "**", other);
+            }
+            // MRI issuses warning here on (RBIGNUM(x)->len * SIZEOF_BDIGITS * yy > 1024*1024)
+            if (((value.bitLength() + 7) / 8) * 4 * Math.abs(fixValue) > 1024 * 1024) {
+                getRuntime().getWarnings().warn(ID.MAY_BE_TOO_BIG, "in a**b, b may be too big", fixValue);
+            }
+            if (fixValue >= 0) {
+                return bignorm(runtime, value.pow((int) fixValue)); // num2int is also implemented
+            } else {
+                return RubyFloat.newFloat(runtime, Math.pow(big2dbl(this), (double)fixValue));
+            }
+        } else if (other instanceof RubyBignum) {
+            d = ((RubyBignum) other).getDoubleValue();
+            getRuntime().getWarnings().warn(ID.MAY_BE_TOO_BIG, "in a**b, b may be too big", d);
+        } else if (other instanceof RubyFloat) {
+            d = ((RubyFloat) other).getDoubleValue();
+        } else {
+            return coerceBin(context, "**", other);
+
+        }
+        return RubyNumeric.dbl2num(runtime, Math.pow(big2dbl(this), d));
+    }    
+    
     /** rb_big_and
      * 
      */
-    public IRubyObject and(IRubyObject other) {
+    @JRubyMethod(name = "&", required = 1)
+    public IRubyObject op_and(ThreadContext context, IRubyObject other) {
         other = other.convertToInteger();
         if (other instanceof RubyBignum) {
             return bignorm(getRuntime(), value.and(((RubyBignum) other).value));
         } else if(other instanceof RubyFixnum) {
             return bignorm(getRuntime(), value.and(fix2big((RubyFixnum)other)));
         }
-        return coerceBin("&", other);
+        return coerceBin(context, "&", other);
     }
 
     /** rb_big_or
      * 
      */
-    public IRubyObject or(IRubyObject other) {
+    @JRubyMethod(name = "|", required = 1)
+    public IRubyObject op_or(ThreadContext context, IRubyObject other) {
         other = other.convertToInteger();
         if (other instanceof RubyBignum) {
             return bignorm(getRuntime(), value.or(((RubyBignum) other).value));
@@ -475,77 +530,120 @@ public class RubyBignum extends RubyInteger {
         if (other instanceof RubyFixnum) { // no bignorm here needed
             return bignorm(getRuntime(), value.or(fix2big((RubyFixnum)other)));
         }
-        return coerceBin("|", other);
+        return coerceBin(context, "|", other);
     }
 
     /** rb_big_xor
      * 
      */
-    public IRubyObject xor(IRubyObject other) {
+    @JRubyMethod(name = "^", required = 1)
+    public IRubyObject op_xor(ThreadContext context, IRubyObject other) {
         other = other.convertToInteger();
         if (other instanceof RubyBignum) {
             return bignorm(getRuntime(), value.xor(((RubyBignum) other).value));
 		}
         if (other instanceof RubyFixnum) {
             return bignorm(getRuntime(), value.xor(BigInteger.valueOf(((RubyFixnum) other).getLongValue())));
-    }
-
-        return coerceBin("^", other);
+        }
+        return coerceBin(context, "^", other);
     }
 
     /** rb_big_neg     
      * 
      */
-    public IRubyObject neg() {
+    @JRubyMethod(name = "~")
+    public IRubyObject op_neg() {
         return RubyBignum.newBignum(getRuntime(), value.not());
-    	}
+    }
 
     /** rb_big_lshift     
      * 
      */
-    public IRubyObject lshift(IRubyObject other) {
-        int width = num2int(other);
-        if (width < 0) {
-            return rshift(RubyFixnum.newFixnum(getRuntime(), -width));
+    @JRubyMethod(name = "<<", required = 1)
+    public IRubyObject op_lshift(IRubyObject other) {
+        long shift;
+        boolean neg = false;
+
+        for (;;) {
+            if (other instanceof RubyFixnum) {
+                shift = ((RubyFixnum)other).getLongValue();
+                if (shift < 0) {
+                    neg = true;
+                    shift = -shift;
+                }
+                break;
+            } else if (other instanceof RubyBignum) {
+                RubyBignum otherBignum = (RubyBignum)other;
+                if (otherBignum.value.signum() < 0) {
+                    IRubyObject tmp = otherBignum.checkShiftDown(this);
+                    if (!tmp.isNil()) return tmp;
+                    neg = true;
+                }
+                shift = big2long(otherBignum);
+                break;
+            }
+            other = other.convertToInteger();
         }
-    	
-        return bignorm(getRuntime(), value.shiftLeft(width));
+
+        return bignorm(getRuntime(), neg ? value.shiftRight((int)shift) : value.shiftLeft((int)shift));
     }
 
     /** rb_big_rshift     
      * 
      */
-    public IRubyObject rshift(IRubyObject other) {
-        int width = num2int(other);
+    @JRubyMethod(name = ">>", required = 1)
+    public IRubyObject op_rshift(IRubyObject other) {
+        long shift;
+        boolean neg = false;
 
-        if (width < 0) {
-            return lshift(RubyFixnum.newFixnum(getRuntime(), -width));
+        for (;;) {
+            if (other instanceof RubyFixnum) {
+                shift = ((RubyFixnum)other).getLongValue();
+                if (shift < 0) {
+                    neg = true;
+                    shift = -shift;
+                }
+                break;
+            } else if (other instanceof RubyBignum) {
+                RubyBignum otherBignum = (RubyBignum)other;
+                if (otherBignum.value.signum() >= 0) {
+                    IRubyObject tmp = otherBignum.checkShiftDown(this);
+                    if (!tmp.isNil()) return tmp;
+                } else {
+                    neg = true;
+                }
+                shift = big2long(otherBignum);
+                break;
+            }
+            other = other.convertToInteger();
         }
-        return bignorm(getRuntime(), value.shiftRight(width));
+        return bignorm(getRuntime(), neg ? value.shiftLeft((int)shift) : value.shiftRight((int)shift));
     }
 
-    /** rb_big_aref     
-     * 
+    /** rb_big_aref
+     *
      */
-    public RubyFixnum aref(IRubyObject other) {
+    @JRubyMethod(name = "[]", required = 1)
+    public RubyFixnum op_aref(IRubyObject other) {
         if (other instanceof RubyBignum) {
             if (((RubyBignum) other).value.signum() >= 0 || value.signum() == -1) {
                 return RubyFixnum.zero(getRuntime());
             }
             return RubyFixnum.one(getRuntime());
         }
-        int position = num2int(other);
-        if (position < 0) {
+        long position = num2long(other);
+        if (position < 0 || position > Integer.MAX_VALUE) {
             return RubyFixnum.zero(getRuntime());
         }
         
-        return value.testBit(num2int(other)) ? RubyFixnum.one(getRuntime()) : RubyFixnum.zero(getRuntime());
+        return value.testBit((int)position) ? RubyFixnum.one(getRuntime()) : RubyFixnum.zero(getRuntime());
     }
 
     /** rb_big_cmp     
      * 
      */
-    public IRubyObject cmp(IRubyObject other) {
+    @JRubyMethod(name = "<=>", required = 1)
+    public IRubyObject op_cmp(ThreadContext context, IRubyObject other) {
         final BigInteger otherValue;
         if (other instanceof RubyFixnum) {
             otherValue = fix2big((RubyFixnum) other);
@@ -554,7 +652,7 @@ public class RubyBignum extends RubyInteger {
         } else if (other instanceof RubyFloat) {
             return dbl_cmp(getRuntime(), big2dbl(this), ((RubyFloat) other).getDoubleValue());
         } else {
-            return coerceCmp("<=>", other);
+            return coerceCmp(context, "<=>", other);
         }
 
         // wow, the only time we can use the java protocol ;)        
@@ -564,7 +662,8 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_eq     
      * 
      */
-    public IRubyObject equal(IRubyObject other) {
+    @JRubyMethod(name = "==", required = 1)
+    public IRubyObject op_equal(IRubyObject other) {
         final BigInteger otherValue;
         if (other instanceof RubyFixnum) {
             otherValue = fix2big((RubyFixnum) other);
@@ -577,7 +676,7 @@ public class RubyBignum extends RubyInteger {
             }
             return RubyBoolean.newBoolean(getRuntime(), a == big2dbl(this));
         } else {
-            return super.equal(other);
+            return other.op_eqq(getRuntime().getCurrentContext(), this);
         }
         return RubyBoolean.newBoolean(getRuntime(), value.compareTo(otherValue) == 0);
     }
@@ -585,6 +684,7 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_eql     
      * 
      */
+    @JRubyMethod(name = {"eql?", "==="}, required = 1)
     public IRubyObject eql_p(IRubyObject other) {
         if (other instanceof RubyBignum) {
             return value.compareTo(((RubyBignum)other).value) == 0 ? getRuntime().getTrue() : getRuntime().getFalse();
@@ -595,6 +695,7 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_hash
      * 
      */
+    @JRubyMethod(name = "hash")
     public RubyFixnum hash() {
         return getRuntime().newFixnum(value.hashCode());
         }
@@ -602,6 +703,7 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_to_f
      * 
      */
+    @JRubyMethod(name = "to_f")
     public IRubyObject to_f() {
         return RubyFloat.newFloat(getRuntime(), getDoubleValue());
     }
@@ -609,6 +711,7 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_abs
      * 
      */
+    @JRubyMethod(name = "abs")
     public IRubyObject abs() {
         return RubyBignum.newBignum(getRuntime(), value.abs());
     }
@@ -616,11 +719,14 @@ public class RubyBignum extends RubyInteger {
     /** rb_big_size
      * 
      */
-    public RubyFixnum size() {
+    @JRubyMethod(name = "size")
+    public IRubyObject size() {
         return getRuntime().newFixnum((value.bitLength() + 7) / 8);
     }
 
     public static void marshalTo(RubyBignum bignum, MarshalStream output) throws IOException {
+        output.registerLinkTarget(bignum);
+
         output.write(bignum.value.signum() >= 0 ? '+' : '-');
         
         BigInteger absValue = bignum.value.abs();
