@@ -33,30 +33,35 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby;
 
-import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
+import org.jruby.runtime.Binding;
 import org.jruby.runtime.Frame;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
+import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 
 /**
  * @author  jpetersen
  */
+@JRubyClass(name="Binding")
 public class RubyBinding extends RubyObject {
-    private Block block;
+    private Binding binding;
 
-    public RubyBinding(Ruby runtime, RubyClass rubyClass, Block block) {
+    public RubyBinding(Ruby runtime, RubyClass rubyClass, Binding binding) {
         super(runtime, rubyClass);
         
-        this.block = block;
+        this.binding = binding;
+    }
+
+    private RubyBinding(Ruby runtime, RubyClass rubyClass) {
+        super(runtime, rubyClass);
     }
     
     private static ObjectAllocator BINDING_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
-            RubyBinding instance = runtime.newBinding();
-            
-            instance.setMetaClass(klass);
+            RubyBinding instance = new RubyBinding(runtime, klass);
             
             return instance;
         }
@@ -64,21 +69,21 @@ public class RubyBinding extends RubyObject {
     
     public static RubyClass createBindingClass(Ruby runtime) {
         RubyClass bindingClass = runtime.defineClass("Binding", runtime.getObject(), BINDING_ALLOCATOR);
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyBinding.class);   
+        runtime.setBinding(bindingClass);
         
-        bindingClass.getMetaClass().defineMethod("of_caller", callbackFactory.getSingletonMethod("of_caller"));
+        bindingClass.defineAnnotatedMethods(RubyBinding.class);
         
         return bindingClass;
     }
 
-    public Block getBlock() {
-        return block;
+    public Binding getBinding() {
+        return binding;
     }
 
     // Proc class
     
-    public static RubyBinding newBinding(Ruby runtime, Block block) {
-        return new RubyBinding(runtime, runtime.getClass("Binding"), block);
+    public static RubyBinding newBinding(Ruby runtime, Binding binding) {
+        return new RubyBinding(runtime, runtime.getBinding(), binding);
     }
 
     public static RubyBinding newBinding(Ruby runtime) {
@@ -86,18 +91,16 @@ public class RubyBinding extends RubyObject {
         
         // FIXME: We should be cloning, not reusing: frame, scope, dynvars, and potentially iter/block info
         Frame frame = context.getCurrentFrame();
-        Block bindingBlock = Block.createBinding(frame, context.getCurrentScope());
+        Binding binding = new Binding(frame, context.getImmediateBindingRubyClass(), context.getCurrentScope());
         
-        return new RubyBinding(runtime, runtime.getClass("Binding"), bindingBlock);
+        return new RubyBinding(runtime, runtime.getBinding(), binding);
     }
 
     /**
      * Create a binding appropriate for a bare "eval", by using the previous (caller's) frame and current
      * scope.
      */
-    public static RubyBinding newBindingForEval(Ruby runtime) {
-        ThreadContext context = runtime.getCurrentContext();
-        
+    public static RubyBinding newBindingForEval(ThreadContext context) {
         // This requires some explaining.  We use Frame values when executing blocks to fill in 
         // various values in ThreadContext and EvalState.eval like rubyClass, cref, and self.
         // Largely, for an eval that is using the logical binding at a place where the eval is 
@@ -115,24 +118,30 @@ public class RubyBinding extends RubyObject {
         currentFrame.setKlazz(previousFrame.getKlazz());
         
         // Set jump target to whatever the previousTarget thinks is good.
-        currentFrame.setJumpTarget(previousFrame.getJumpTarget() != null ? previousFrame.getJumpTarget() : previousFrame);
+//        currentFrame.setJumpTarget(previousFrame.getJumpTarget() != null ? previousFrame.getJumpTarget() : previousFrame);
         
-        Block bindingBlock = Block.createBinding(previousFrame, context.getCurrentScope());
+        Binding binding = new Binding(previousFrame, context.getEvalBindingRubyClass(), context.getCurrentScope());
+        Ruby runtime = context.getRuntime();
         
-        return new RubyBinding(runtime, runtime.getClass("Binding"), bindingBlock);
-    }
-
-    public static RubyBinding newBindingOfCaller(Ruby runtime) {
-        ThreadContext context = runtime.getCurrentContext();
-        
-        // FIXME: We should be cloning, not reusing: frame, scope, dynvars, and potentially iter/block info
-        Frame frame = context.getPreviousFrame();
-        Block bindingBlock = Block.createBinding(frame, context.getPreviousScope());
-        
-        return new RubyBinding(runtime, runtime.getClass("Binding"), bindingBlock);
+        return new RubyBinding(runtime, runtime.getBinding(), binding);
     }
     
-    public static IRubyObject of_caller(IRubyObject recv, Block aBlock) {
-        return RubyBinding.newBindingOfCaller(recv.getRuntime());
+    @JRubyMethod(name = "initialize", visibility = Visibility.PRIVATE)
+    public IRubyObject initialize(ThreadContext context) {
+        // FIXME: We should be cloning, not reusing: frame, scope, dynvars, and potentially iter/block info
+        Frame frame = context.getCurrentFrame();
+        binding = new Binding(frame, context.getImmediateBindingRubyClass(), context.getCurrentScope());
+        
+        return this;
+    }
+    
+    @JRubyMethod(name = "initialize_copy", required = 1, visibility = Visibility.PRIVATE)
+    @Override
+    public IRubyObject initialize_copy(IRubyObject other) {
+        RubyBinding otherBinding = (RubyBinding)other;
+        
+        binding = otherBinding.binding;
+        
+        return this;
     }
 }
