@@ -31,16 +31,19 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby;
 
+import static org.jruby.RubyEnumerator.enumeratorize;
+
 import java.io.File;
+import java.io.FileInputStream;
 import java.io.IOException;
 import java.util.ArrayList;
-import java.util.Iterator;
 import java.util.List;
+import org.jruby.anno.JRubyMethod;
+import org.jruby.anno.JRubyClass;
+import org.jruby.ext.posix.util.Platform;
 
 import org.jruby.javasupport.JavaUtil;
-import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
@@ -53,6 +56,7 @@ import org.jruby.util.ByteList;
  *
  * @author  jvoegele
  */
+@JRubyClass(name="Dir", include="Enumerable")
 public class RubyDir extends RubyObject {
 	// What we passed to the constructor for method 'path'
     private RubyString    path;
@@ -65,7 +69,7 @@ public class RubyDir extends RubyObject {
         super(runtime, type);
     }
     
-    private static ObjectAllocator DIR_ALLOCATOR = new ObjectAllocator() {
+    private static final ObjectAllocator DIR_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
             return new RubyDir(runtime, klass);
         }
@@ -73,43 +77,22 @@ public class RubyDir extends RubyObject {
 
     public static RubyClass createDirClass(Ruby runtime) {
         RubyClass dirClass = runtime.defineClass("Dir", runtime.getObject(), DIR_ALLOCATOR);
+        runtime.setDir(dirClass);
 
-        dirClass.includeModule(runtime.getModule("Enumerable"));
-
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyDir.class);
-
-        dirClass.getMetaClass().defineMethod("glob", callbackFactory.getOptSingletonMethod("glob"));
-        dirClass.getMetaClass().defineFastMethod("entries", callbackFactory.getFastSingletonMethod("entries", RubyKernel.IRUBY_OBJECT));
-        dirClass.getMetaClass().defineMethod("[]", callbackFactory.getOptSingletonMethod("glob"));
-        dirClass.getMetaClass().defineMethod("chdir", callbackFactory.getOptSingletonMethod("chdir"));
-        dirClass.getMetaClass().defineFastMethod("chroot", callbackFactory.getFastSingletonMethod("chroot", RubyKernel.IRUBY_OBJECT));
-        //dirClass.defineSingletonMethod("delete", callbackFactory.getSingletonMethod(RubyDir.class, "delete", RubyString.class));
-        dirClass.getMetaClass().defineMethod("foreach", callbackFactory.getSingletonMethod("foreach", RubyKernel.IRUBY_OBJECT));
-        dirClass.getMetaClass().defineFastMethod("getwd", callbackFactory.getFastSingletonMethod("getwd"));
-        dirClass.getMetaClass().defineFastMethod("pwd", callbackFactory.getFastSingletonMethod("getwd"));
-        // dirClass.defineAlias("pwd", "getwd");
-        dirClass.getMetaClass().defineFastMethod("mkdir", callbackFactory.getFastOptSingletonMethod("mkdir"));
-        dirClass.getMetaClass().defineMethod("open", callbackFactory.getSingletonMethod("open", RubyKernel.IRUBY_OBJECT));
-        dirClass.getMetaClass().defineFastMethod("rmdir", callbackFactory.getFastSingletonMethod("rmdir", RubyKernel.IRUBY_OBJECT));
-        dirClass.getMetaClass().defineFastMethod("unlink", callbackFactory.getFastSingletonMethod("rmdir", RubyKernel.IRUBY_OBJECT));
-        dirClass.getMetaClass().defineFastMethod("delete", callbackFactory.getFastSingletonMethod("rmdir", RubyKernel.IRUBY_OBJECT));
-        // dirClass.defineAlias("unlink", "rmdir");
-        // dirClass.defineAlias("delete", "rmdir");
-
-        dirClass.defineFastMethod("close", callbackFactory.getFastMethod("close"));
-        dirClass.defineMethod("each", callbackFactory.getMethod("each"));
-        dirClass.defineFastMethod("entries", callbackFactory.getFastMethod("entries"));
-        dirClass.defineFastMethod("path", callbackFactory.getFastMethod("path"));
-        dirClass.defineFastMethod("tell", callbackFactory.getFastMethod("tell"));
-        dirClass.defineAlias("pos", "tell");
-        dirClass.defineFastMethod("seek", callbackFactory.getFastMethod("seek", RubyKernel.IRUBY_OBJECT));
-        dirClass.defineFastMethod("pos=", callbackFactory.getFastMethod("setPos", RubyKernel.IRUBY_OBJECT));
-        dirClass.defineFastMethod("read", callbackFactory.getFastMethod("read"));
-        dirClass.defineFastMethod("rewind", callbackFactory.getFastMethod("rewind"));
-        dirClass.defineMethod("initialize", callbackFactory.getMethod("initialize", RubyKernel.IRUBY_OBJECT));
+        dirClass.includeModule(runtime.getEnumerable());
+        
+        dirClass.defineAnnotatedMethods(RubyDir.class);
 
         return dirClass;
     }
+    
+    private final void checkDir() {
+        if (!isTaint() && getRuntime().getSafeLevel() >= 4) throw getRuntime().newSecurityError("Insecure: operation on untainted Dir");
+     
+        testFrozen("");
+        
+        if (!isOpen) throw getRuntime().newIOError("closed directory");
+    }    
 
     /**
      * Creates a new <code>Dir</code>.  This method takes a snapshot of the
@@ -118,16 +101,21 @@ public class RubyDir extends RubyObject {
      * <code>Dir</code> object returned, so a new <code>Dir</code> instance
      * must be created to reflect changes to the underlying file system.
      */
+    @JRubyMethod(name = "initialize", required = 1, frame = true)
     public IRubyObject initialize(IRubyObject _newPath, Block unusedBlock) {
         RubyString newPath = _newPath.convertToString();
         getRuntime().checkSafeString(newPath);
-        dir = JRubyFile.create(getRuntime().getCurrentDirectory(),newPath.toString());
+
+        String adjustedPath = RubyFile.adjustRootPathOnWindows(getRuntime(), newPath.toString(), null);
+        checkDirIsTwoSlashesOnWindows(getRuntime(), adjustedPath);
+
+        dir = JRubyFile.create(getRuntime().getCurrentDirectory(), adjustedPath);
         if (!dir.isDirectory()) {
             dir = null;
             throw getRuntime().newErrnoENOENTError(newPath.toString() + " is not a directory");
         }
         path = newPath;
-		List snapshotList = new ArrayList();
+		List<String> snapshotList = new ArrayList<String>();
 		snapshotList.add(".");
 		snapshotList.add("..");
 		snapshotList.addAll(getContents(dir));
@@ -138,48 +126,87 @@ public class RubyDir extends RubyObject {
     }
 
 // ----- Ruby Class Methods ----------------------------------------------------
+    
+    private static List<ByteList> dirGlobs(String cwd, IRubyObject[] args, int flags) {
+        List<ByteList> dirs = new ArrayList<ByteList>();
+        
+        for (int i = 0; i < args.length; i++) {
+            ByteList globPattern = args[i].convertToString().getByteList();
+            dirs.addAll(Dir.push_glob(cwd, globPattern, flags));
+        }
+        
+        return dirs;
+    }
+    
+    private static IRubyObject asRubyStringList(Ruby runtime, List<ByteList> dirs) {
+        List<RubyString> allFiles = new ArrayList<RubyString>();
 
+        for (ByteList dir: dirs) {
+            allFiles.add(RubyString.newString(runtime, dir));
+        }            
+
+        IRubyObject[] tempFileList = new IRubyObject[allFiles.size()];
+        allFiles.toArray(tempFileList);
+         
+        return runtime.newArrayNoCopy(tempFileList);
+    }
+    
+    private static String getCWD(Ruby runtime) {
+        try {
+            return new org.jruby.util.NormalizedFile(runtime.getCurrentDirectory()).getCanonicalPath();
+        } catch(Exception e) {
+            return runtime.getCurrentDirectory();
+        }
+    }
+
+    @JRubyMethod(name = "[]", required = 1, rest=true, meta = true)
+    public static IRubyObject aref(IRubyObject recv, IRubyObject[] args) {
+        List<ByteList> dirs;
+        if (args.length == 1) {
+            ByteList globPattern = args[0].convertToString().getByteList();
+            dirs = Dir.push_glob(getCWD(recv.getRuntime()), globPattern, 0);
+        } else {
+            dirs = dirGlobs(getCWD(recv.getRuntime()), args, 0);
+        }
+
+        return asRubyStringList(recv.getRuntime(), dirs);
+    }
+    
     /**
      * Returns an array of filenames matching the specified wildcard pattern
      * <code>pat</code>. If a block is given, the array is iterated internally
      * with each filename is passed to the block in turn. In this case, Nil is
      * returned.  
      */
-    public static IRubyObject glob(IRubyObject recv, IRubyObject[] args, Block block) {
-        String cwd = recv.getRuntime().getCurrentDirectory();
-        int flags = 0;
-        if(Arity.checkArgumentCount(recv.getRuntime(),args,1,2) == 2) {
-            flags = RubyNumeric.num2int(args[1]);
-        }
-        ByteList pt = args[0].convertToString().getByteList();
+    @JRubyMethod(name = "glob", required = 1, optional = 1, frame = true, meta = true)
+    public static IRubyObject glob(ThreadContext context, IRubyObject recv, IRubyObject[] args, Block block) {
+        Ruby runtime = recv.getRuntime();
+        int flags = args.length == 2 ?  RubyNumeric.num2int(args[1]) : 0;
 
-        String cwd2;
-        try {
-            cwd2 = new org.jruby.util.NormalizedFile(cwd).getCanonicalPath();
-        } catch(Exception e) {
-            cwd2 = cwd;
+        List<ByteList> dirs;
+        IRubyObject tmp = args[0].checkArrayType();
+        if (tmp.isNil()) {
+            ByteList globPattern = args[0].convertToString().getByteList();
+            dirs = Dir.push_glob(recv.getRuntime().getCurrentDirectory(), globPattern, flags);
+        } else {
+            dirs = dirGlobs(getCWD(runtime), ((RubyArray) tmp).toJavaArray(), flags);
         }
-
-        List l = Dir.push_glob(cwd2, pt.bytes, 0, pt.realSize, flags);
         
-        if(block.isGiven()) {
-            ThreadContext context = recv.getRuntime().getCurrentContext();
-            for(Iterator iter = l.iterator(); iter.hasNext(); ) {
-                block.yield(context, RubyString.newString(recv.getRuntime(),(ByteList)iter.next()));
+        if (block.isGiven()) {
+            for (int i = 0; i < dirs.size(); i++) {
+                block.yield(context, RubyString.newString(runtime, dirs.get(i)));
             }
+        
             return recv.getRuntime().getNil();
         }
-        IRubyObject[] l2 = new IRubyObject[l.size()];
-        int i=0;
-        for(Iterator iter = l.iterator(); iter.hasNext(); i++) {
-            l2[i] = RubyString.newString(recv.getRuntime(),(ByteList)iter.next());
-        }
-        return recv.getRuntime().newArrayNoCopy(l2);
+
+        return asRubyStringList(recv.getRuntime(), dirs);
     }
 
     /**
      * @return all entries for this Dir
      */
+    @JRubyMethod(name = "entries")
     public RubyArray entries() {
         return getRuntime().newArrayNoCopy(JavaUtil.convertJavaArrayToRuby(getRuntime(), snapshot));
     }
@@ -187,25 +214,43 @@ public class RubyDir extends RubyObject {
     /**
      * Returns an array containing all of the filenames in the given directory.
      */
+    @JRubyMethod(name = "entries", required = 1, meta = true)
     public static RubyArray entries(IRubyObject recv, IRubyObject path) {
-        final JRubyFile directory = JRubyFile.create(recv.getRuntime().getCurrentDirectory(),path.convertToString().toString());
-        
+        Ruby runtime = recv.getRuntime();
+
+        String adjustedPath = RubyFile.adjustRootPathOnWindows(
+                runtime, path.convertToString().toString(), null);
+        checkDirIsTwoSlashesOnWindows(runtime, adjustedPath);
+
+        final JRubyFile directory = JRubyFile.create(
+                recv.getRuntime().getCurrentDirectory(), adjustedPath);
+
         if (!directory.isDirectory()) {
             throw recv.getRuntime().newErrnoENOENTError("No such directory");
         }
-        List fileList = getContents(directory);
-		fileList.add(0,".");
-		fileList.add(1,"..");
+        List<String> fileList = getContents(directory);
+		fileList.add(0, ".");
+		fileList.add(1, "..");
         Object[] files = fileList.toArray();
         return recv.getRuntime().newArrayNoCopy(JavaUtil.convertJavaArrayToRuby(recv.getRuntime(), files));
     }
+    
+    // MRI behavior: just plain '//' or '\\\\' are considered illegal on Windows.
+    private static void checkDirIsTwoSlashesOnWindows(Ruby runtime, String path) {
+        if (Platform.IS_WINDOWS && ("//".equals(path) || "\\\\".equals(path))) {
+            throw runtime.newErrnoEINVALError("Invalid argument - " + path);
+        }
+    }
 
     /** Changes the current directory to <code>path</code> */
-    public static IRubyObject chdir(IRubyObject recv, IRubyObject[] args, Block block) {
-        Arity.checkArgumentCount(recv.getRuntime(), args, 0, 1);
+    @JRubyMethod(name = "chdir", optional = 1, frame = true, meta = true)
+    public static IRubyObject chdir(ThreadContext context, IRubyObject recv, IRubyObject[] args, Block block) {
         RubyString path = args.length == 1 ? 
-            (RubyString) args[0].convertToString() : getHomeDirectoryPath(recv); 
-        JRubyFile dir = getDir(recv.getRuntime(), path.toString(), true);
+            (RubyString) args[0].convertToString() : getHomeDirectoryPath(context);
+        String adjustedPath = RubyFile.adjustRootPathOnWindows(
+                recv.getRuntime(), path.toString(), null);
+        checkDirIsTwoSlashesOnWindows(recv.getRuntime(), adjustedPath);
+        JRubyFile dir = getDir(recv.getRuntime(), adjustedPath, true);
         String realPath = null;
         String oldCwd = recv.getRuntime().getCurrentDirectory();
         
@@ -223,8 +268,9 @@ public class RubyDir extends RubyObject {
         	// FIXME: Don't allow multiple threads to do this at once
             recv.getRuntime().setCurrentDirectory(realPath);
             try {
-                result = block.yield(recv.getRuntime().getCurrentContext(), path);
+                result = block.yield(context, path);
             } finally {
+                dir = getDir(recv.getRuntime(), oldCwd, true);
                 recv.getRuntime().setCurrentDirectory(oldCwd);
             }
         } else {
@@ -239,6 +285,7 @@ public class RubyDir extends RubyObject {
      * Changes the root directory (only allowed by super user).  Not available
      * on all platforms.
      */
+    @JRubyMethod(name = "chroot", required = 1, meta = true)
     public static IRubyObject chroot(IRubyObject recv, IRubyObject path) {
         throw recv.getRuntime().newNotImplementedError("chroot not implemented: chroot is non-portable and is not supported.");
     }
@@ -247,6 +294,7 @@ public class RubyDir extends RubyObject {
      * Deletes the directory specified by <code>path</code>.  The directory must
      * be empty.
      */
+    @JRubyMethod(name = {"rmdir", "unlink", "delete"}, required = 1, meta = true)
     public static IRubyObject rmdir(IRubyObject recv, IRubyObject path) {
         JRubyFile directory = getDir(recv.getRuntime(), path.convertToString().toString(), true);
         
@@ -261,20 +309,29 @@ public class RubyDir extends RubyObject {
      * Executes the block once for each file in the directory specified by
      * <code>path</code>.
      */
-    public static IRubyObject foreach(IRubyObject recv, IRubyObject _path, Block block) {
+    @JRubyMethod(name = "foreach", required = 1, frame = true, meta = true)
+    public static IRubyObject foreach(ThreadContext context, IRubyObject recv, IRubyObject _path, Block block) {
         RubyString path = _path.convertToString();
         recv.getRuntime().checkSafeString(path);
 
-        RubyClass dirClass = recv.getRuntime().getClass("Dir");
-        RubyDir dir = (RubyDir) dirClass.newInstance(new IRubyObject[] { path }, block);
+        RubyClass dirClass = recv.getRuntime().getDir();
+        RubyDir dir = (RubyDir) dirClass.newInstance(context, new IRubyObject[] { path }, block);
         
-        dir.each(block);
+        dir.each(context, block);
         return recv.getRuntime().getNil();
     }
 
+    @JRubyMethod(name = "foreach", frame = true, compat = CompatVersion.RUBY1_9)
+    public static IRubyObject foreach19(ThreadContext context, IRubyObject recv, IRubyObject _path, Block block) {
+        return block.isGiven() ? foreach(context, recv, _path, block) : enumeratorize(context.getRuntime(), recv, "foreach", _path);
+    }
+
     /** Returns the current directory. */
+    @JRubyMethod(name = {"getwd", "pwd"}, meta = true)
     public static RubyString getwd(IRubyObject recv) {
-        return recv.getRuntime().newString(recv.getRuntime().getCurrentDirectory());
+        Ruby ruby = recv.getRuntime();
+        
+        return RubyString.newUnicodeString(ruby, ruby.getCurrentDirectory());
     }
 
     /**
@@ -282,24 +339,25 @@ public class RubyDir extends RubyObject {
      * <code>mode</code> parameter is provided only to support existing Ruby
      * code, and is ignored.
      */
+    @JRubyMethod(name = "mkdir", required = 1, optional = 1, meta = true)
     public static IRubyObject mkdir(IRubyObject recv, IRubyObject[] args) {
-        if (args.length < 1) {
-            throw recv.getRuntime().newArgumentError(args.length, 1);
-        }
-        if (args.length > 2) {
-            throw recv.getRuntime().newArgumentError(args.length, 2);
-        }
-
-        recv.getRuntime().checkSafeString(args[0]);
+        Ruby runtime = recv.getRuntime();
+        runtime.checkSafeString(args[0]);
         String path = args[0].toString();
 
-        File newDir = getDir(recv.getRuntime(), path, false);
+        File newDir = getDir(runtime, path, false);
         if (File.separatorChar == '\\') {
             newDir = new File(newDir.getPath());
         }
         
-        return newDir.mkdirs() ? RubyFixnum.zero(recv.getRuntime()) :
-            RubyFixnum.one(recv.getRuntime());
+        int mode = args.length == 2 ? ((int) args[1].convertToInteger().getLongValue()) : 0777;
+
+        if (runtime.getPosix().mkdir(newDir.getAbsolutePath(), mode) < 0) {
+            // FIXME: This is a system error based on errno
+            throw recv.getRuntime().newSystemCallError("mkdir failed");
+        }
+        
+        return RubyFixnum.zero(recv.getRuntime());
     }
 
     /**
@@ -307,20 +365,19 @@ public class RubyDir extends RubyObject {
      * provided, a new directory object is passed to the block, which closes the
      * directory object before terminating.
      */
-    public static IRubyObject open(IRubyObject recv, IRubyObject path, Block block) {
+    @JRubyMethod(name = "open", required = 1, frame = true, meta = true)
+    public static IRubyObject open(ThreadContext context, IRubyObject recv, IRubyObject path, Block block) {
         RubyDir directory = 
-            (RubyDir) recv.getRuntime().getClass("Dir").newInstance(
+            (RubyDir) recv.getRuntime().getDir().newInstance(context,
                     new IRubyObject[] { path }, Block.NULL_BLOCK);
 
         if (!block.isGiven()) return directory;
         
         try {
-            block.yield(recv.getRuntime().getCurrentContext(), directory);
+            return block.yield(context, directory);
         } finally {
             directory.close();
         }
-            
-        return recv.getRuntime().getNil();
     }
 
 // ----- Ruby Instance Methods -------------------------------------------------
@@ -328,8 +385,11 @@ public class RubyDir extends RubyObject {
     /**
      * Closes the directory stream.
      */
+    @JRubyMethod(name = "close")
     public IRubyObject close() {
         // Make sure any read()s after close fail.
+        checkDir();
+        
         isOpen = false;
 
         return getRuntime().getNil();
@@ -338,19 +398,28 @@ public class RubyDir extends RubyObject {
     /**
      * Executes the block once for each entry in the directory.
      */
-    public IRubyObject each(Block block) {
+    @JRubyMethod(name = "each", frame = true)
+    public IRubyObject each(ThreadContext context, Block block) {
+        checkDir();
+        
         String[] contents = snapshot;
-        ThreadContext context = getRuntime().getCurrentContext();
         for (int i=0; i<contents.length; i++) {
             block.yield(context, getRuntime().newString(contents[i]));
         }
         return this;
     }
 
+    @JRubyMethod(name = "each", frame = true, compat = CompatVersion.RUBY1_9)
+    public IRubyObject each19(ThreadContext context, Block block) {
+        return block.isGiven() ? each(context, block) : enumeratorize(context.getRuntime(), this, "each");
+    }
+
     /**
      * Returns the current position in the directory.
      */
+    @JRubyMethod(name = {"tell", "pos"})
     public RubyInteger tell() {
+        checkDir();
         return getRuntime().newFixnum(pos);
     }
 
@@ -358,29 +427,31 @@ public class RubyDir extends RubyObject {
      * Moves to a position <code>d</code>.  <code>pos</code> must be a value
      * returned by <code>tell</code> or 0.
      */
+    @JRubyMethod(name = "seek", required = 1)
     public IRubyObject seek(IRubyObject newPos) {
-        setPos(newPos);
+        checkDir();
+        
+        set_pos(newPos);
         return this;
     }
     
-    public IRubyObject setPos(IRubyObject newPos) {
+    @JRubyMethod(name = "pos=", required = 1)
+    public IRubyObject set_pos(IRubyObject newPos) {
         this.pos = RubyNumeric.fix2int(newPos);
         return newPos;
     }
 
-    public IRubyObject path() {
-        if (!isOpen) {
-            throw getRuntime().newIOError("closed directory");
-        }
+    @JRubyMethod(name = "path")
+    public IRubyObject path(ThreadContext context) {
+        checkDir();
         
-        return path;
+        return path.strDup(context.getRuntime());
     }
 
     /** Returns the next entry from this directory. */
+    @JRubyMethod(name = "read")
     public IRubyObject read() {
-	if (!isOpen) {
-	    throw getRuntime().newIOError("Directory already closed");
-	}
+        checkDir();
 
         if (pos >= snapshot.length) {
             return getRuntime().getNil();
@@ -391,9 +462,13 @@ public class RubyDir extends RubyObject {
     }
 
     /** Moves position in this directory to the first entry. */
+    @JRubyMethod(name = "rewind")
     public IRubyObject rewind() {
+        if (!isTaint() && getRuntime().getSafeLevel() >= 4) throw getRuntime().newSecurityError("Insecure: can't close");
+        checkDir();
+
         pos = 0;
-        return getRuntime().newFixnum(pos);
+        return this;
     }
 
 // ----- Helper Methods --------------------------------------------------------
@@ -407,10 +482,13 @@ public class RubyDir extends RubyObject {
      */
     protected static JRubyFile getDir(final Ruby runtime, final String path, final boolean mustExist) {
         JRubyFile result = JRubyFile.create(runtime.getCurrentDirectory(),path);
+        if (mustExist && !result.exists()) {
+            throw runtime.newErrnoENOENTError("No such file or directory - " + path);
+        }
         boolean isDirectory = result.isDirectory();
         
         if (mustExist && !isDirectory) {
-            throw runtime.newErrnoENOENTError(path + " is not a directory");
+            throw runtime.newErrnoENOTDIRError(path + " is not a directory");
         } else if (!mustExist && isDirectory) {
             throw runtime.newErrnoEEXISTError("File exists - " + path); 
         }
@@ -422,9 +500,9 @@ public class RubyDir extends RubyObject {
      * Returns the contents of the specified <code>directory</code> as an
      * <code>ArrayList</code> containing the names of the files as Java Strings.
      */
-    protected static List getContents(File directory) {
+    protected static List<String> getContents(File directory) {
         String[] contents = directory.list();
-        List result = new ArrayList();
+        List<String> result = new ArrayList<String>();
 
         // If an IO exception occurs (something odd, but possible)
         // A directory may return null.
@@ -440,8 +518,8 @@ public class RubyDir extends RubyObject {
      * Returns the contents of the specified <code>directory</code> as an
      * <code>ArrayList</code> containing the names of the files as Ruby Strings.
      */
-    protected static List getContents(File directory, Ruby runtime) {
-        List result = new ArrayList();
+    protected static List<RubyString> getContents(File directory, Ruby runtime) {
+        List<RubyString> result = new ArrayList<RubyString>();
         String[] contents = directory.list();
         
         for (int i = 0; i < contents.length; i++) {
@@ -450,30 +528,62 @@ public class RubyDir extends RubyObject {
         return result;
     }
 	
-	/*
-	 * Poor mans find home directory.  I am not sure how windows ruby behaves with '~foo', but
-	 * this mostly will work on any unix/linux/cygwin system.  When someone wants to extend this
-	 * to include the windows way, we should consider moving this to an external ruby file.
-	 */
-	public static IRubyObject getHomeDirectoryPath(IRubyObject recv, String user) {
-		// TODO: Having a return where I set user inside readlines created a JumpException.  It seems that
-		// evalScript should catch that and return?
-		return recv.getRuntime().evalScript("File.open('/etc/passwd') do |f| f.readlines.each do" +
-				"|l| f = l.split(':'); return f[5] if f[0] == '" + user + "'; end; end; nil");
-	}
-	
-	public static RubyString getHomeDirectoryPath(IRubyObject recv) {
-		RubyHash hash = (RubyHash) recv.getRuntime().getObject().getConstant("ENV_JAVA");
-		IRubyObject home = hash.aref(recv.getRuntime().newString("user.home"));
-		
-		if (home == null || home.isNil()) {
-			home = hash.aref(recv.getRuntime().newString("LOGDIR"));
-		}
-		
-		if (home == null || home.isNil()) {
-			throw recv.getRuntime().newArgumentError("user.home/LOGDIR not set");
-		}
-		
-		return (RubyString) home;
-	}
+    /**
+     * Returns the home directory of the specified <code>user</code> on the
+     * system. If the home directory of the specified user cannot be found,
+     * an <code>ArgumentError it thrown</code>.
+     */
+    public static IRubyObject getHomeDirectoryPath(ThreadContext context, String user) {
+        /*
+         * TODO: This version is better than the hackish previous one. Windows
+         *       behavior needs to be defined though. I suppose this version
+         *       could be improved more too.
+         * TODO: /etc/passwd is also inadequate for MacOSX since it does not
+         *       use /etc/passwd for regular user accounts
+         */
+
+        String passwd = null;
+        try {
+            FileInputStream stream = new FileInputStream("/etc/passwd");
+            int totalBytes = stream.available();
+            byte[] bytes = new byte[totalBytes];
+            stream.read(bytes);
+            stream.close();
+            passwd = new String(bytes);
+        } catch (IOException e) {
+            return context.getRuntime().getNil();
+        }
+
+        String[] rows = passwd.split("\n");
+        int rowCount = rows.length;
+        for (int i = 0; i < rowCount; i++) {
+            String[] fields = rows[i].split(":");
+            if (fields[0].equals(user)) {
+                return context.getRuntime().newString(fields[5]);
+            }
+        }
+
+        throw context.getRuntime().newArgumentError("user " + user + " doesn't exist");
+    }
+
+    public static RubyString getHomeDirectoryPath(ThreadContext context) {
+        Ruby runtime = context.getRuntime();
+        RubyHash systemHash = (RubyHash) runtime.getObject().fastGetConstant("ENV_JAVA");
+        RubyHash envHash = (RubyHash) runtime.getObject().fastGetConstant("ENV");
+        IRubyObject home = envHash.op_aref(context, runtime.newString("HOME"));
+
+        if (home == null || home.isNil()) {
+            home = systemHash.op_aref(context, runtime.newString("user.home"));
+        }
+
+        if (home == null || home.isNil()) {
+            home = envHash.op_aref(context, runtime.newString("LOGDIR"));
+        }
+
+        if (home == null || home.isNil()) {
+            throw runtime.newArgumentError("user.home/LOGDIR not set");
+        }
+
+        return (RubyString) home;
+    }
 }
