@@ -38,33 +38,35 @@ package org.jruby;
 
 import java.io.IOException;
 import java.io.PrintStream;
-import java.util.HashMap;
-import java.util.Map;
+import java.util.List;
 
-import org.jruby.runtime.Arity;
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
-import org.jruby.runtime.Frame;
-import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ObjectMarshal;
 import org.jruby.runtime.ThreadContext;
+import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
+import org.jruby.runtime.builtin.Variable;
+import org.jruby.runtime.component.VariableEntry;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
+import org.jruby.util.SafePropertyAccessor;
 
 /**
  *
  * @author  jpetersen
  */
+@JRubyClass(name="Exception")
 public class RubyException extends RubyObject {
-
-    private Frame[] backtraceFrames;
+    private StackTraceElement[] backtraceFrames;
+    private StackTraceElement[] javaStackTrace;
     private IRubyObject backtrace;
     public IRubyObject message;
-	public static final int TRACE_HEAD = 8;
-	public static final int TRACE_TAIL = 4;
-	public static final int TRACE_MAX = TRACE_HEAD + TRACE_TAIL + 6;
+    public static final int TRACE_HEAD = 8;
+    public static final int TRACE_TAIL = 4;
+    public static final int TRACE_MAX = TRACE_HEAD + TRACE_TAIL + 6;
 
     protected RubyException(Ruby runtime, RubyClass rubyClass) {
         this(runtime, rubyClass, null);
@@ -92,12 +94,12 @@ public class RubyException extends RubyObject {
                               MarshalStream marshalStream) throws IOException {
             RubyException exc = (RubyException)obj;
             
-            Map iVars = new HashMap(exc.getInstanceVariables());
-            
-            iVars.put("mesg", exc.message == null ? runtime.getNil() : exc.message);
-            iVars.put("bt", exc.getBacktrace());
-            
-            marshalStream.dumpInstanceVars(iVars);
+            marshalStream.registerLinkTarget(exc);
+            List<Variable<IRubyObject>> attrs = exc.getVariableList();
+            attrs.add(new VariableEntry<IRubyObject>(
+                    "mesg", exc.message == null ? runtime.getNil() : exc.message));
+            attrs.add(new VariableEntry<IRubyObject>("bt", exc.getBacktrace()));
+            marshalStream.dumpVariables(attrs);
         }
 
         public Object unmarshalFrom(Ruby runtime, RubyClass type,
@@ -105,10 +107,10 @@ public class RubyException extends RubyObject {
             RubyException exc = (RubyException)type.allocate();
             
             unmarshalStream.registerLinkTarget(exc);
-            unmarshalStream.defaultInstanceVarsUnmarshal(exc);
+            unmarshalStream.defaultVariablesUnmarshal(exc);
             
-            exc.message = exc.removeInstanceVariable("mesg");
-            exc.set_backtrace(exc.removeInstanceVariable("bt"));
+            exc.message = exc.removeInternalVariable("mesg");
+            exc.set_backtrace(exc.removeInternalVariable("bt"));
             
             return exc;
         }
@@ -116,21 +118,10 @@ public class RubyException extends RubyObject {
 
     public static RubyClass createExceptionClass(Ruby runtime) {
         RubyClass exceptionClass = runtime.defineClass("Exception", runtime.getObject(), EXCEPTION_ALLOCATOR);
+        runtime.setException(exceptionClass);
 
         exceptionClass.setMarshal(EXCEPTION_MARSHAL);
-        
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyException.class);
-        CallbackFactory classCB = runtime.callbackFactory(RubyClass.class);
-        // TODO: could this just  be an alias for new?
-        exceptionClass.getMetaClass().defineMethod("exception", classCB.getOptMethod("newInstance"));		
-        exceptionClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
-        exceptionClass.defineFastMethod("exception", callbackFactory.getFastOptMethod("exception"));
-        exceptionClass.defineFastMethod("to_s", callbackFactory.getFastMethod("to_s"));
-        exceptionClass.defineFastMethod("to_str", callbackFactory.getFastMethod("to_str"));
-        exceptionClass.defineFastMethod("message", callbackFactory.getFastMethod("to_str"));
-        exceptionClass.defineFastMethod("inspect", callbackFactory.getFastMethod("inspect"));
-        exceptionClass.defineFastMethod("backtrace", callbackFactory.getFastMethod("backtrace"));		
-        exceptionClass.defineFastMethod("set_backtrace", callbackFactory.getFastMethod("set_backtrace", RubyKernel.IRUBY_OBJECT));		
+        exceptionClass.defineAnnotatedMethods(RubyException.class);
 
         return exceptionClass;
     }
@@ -139,26 +130,74 @@ public class RubyException extends RubyObject {
         return new RubyException(runtime, excptnClass, msg);
     }
     
-    public void setBacktraceFrames(Frame[] backtraceFrames) {
+    public void setBacktraceFrames(StackTraceElement[] backtraceFrames) {
         this.backtraceFrames = backtraceFrames;
+        if (TRACE_TYPE == RAW ||
+                TRACE_TYPE == RAW_FILTERED ||
+                TRACE_TYPE == RUBY_COMPILED ||
+                TRACE_TYPE == RUBY_HYBRID) {
+            javaStackTrace = Thread.currentThread().getStackTrace();
+        }
+    }
+    
+    public static final int RAW = 0;
+    public static final int RAW_FILTERED = 1;
+    public static final int RUBY_FRAMED = 2;
+    public static final int RUBY_COMPILED = 3;
+    public static final int RUBY_HYBRID = 4;
+    
+    public static final int TRACE_TYPE;
+    
+    static {
+        String style = SafePropertyAccessor.getProperty("jruby.backtrace.style", "ruby_framed").toLowerCase();
+        
+        if (style.equals("raw")) TRACE_TYPE = RAW;
+        else if (style.equals("raw_filtered")) TRACE_TYPE = RAW_FILTERED;
+        else if (style.equals("ruby_framed")) TRACE_TYPE = RUBY_FRAMED;
+        else if (style.equals("ruby_compiled")) TRACE_TYPE = RUBY_COMPILED;
+        else if (style.equals("ruby_hybrid")) TRACE_TYPE = RUBY_HYBRID;
+        else TRACE_TYPE = RUBY_FRAMED;
     }
     
     public IRubyObject getBacktrace() {
         if (backtrace == null) {
-            backtrace = backtraceFrames == null ? getRuntime().getNil() : ThreadContext.createBacktraceFromFrames(getRuntime(), backtraceFrames);
+            initBacktrace();
         }
         return backtrace;
     }
+    
+    public void initBacktrace() {
+        switch (TRACE_TYPE) {
+        case RAW:
+            backtrace = ThreadContext.createRawBacktrace(getRuntime(), javaStackTrace, false);
+            break;
+        case RAW_FILTERED:
+            backtrace = ThreadContext.createRawBacktrace(getRuntime(), javaStackTrace, true);
+            break;
+        case RUBY_FRAMED:
+            backtrace = backtraceFrames == null ? getRuntime().getNil() : ThreadContext.createBacktraceFromFrames(getRuntime(), backtraceFrames);
+            break;
+        case RUBY_COMPILED:
+            backtrace = ThreadContext.createRubyCompiledBacktrace(getRuntime(), javaStackTrace);
+            break;
+        case RUBY_HYBRID:
+            backtrace = ThreadContext.createRubyHybridBacktrace(getRuntime(), backtraceFrames, javaStackTrace, getRuntime().getDebug().isTrue());
+            break;
+        }
+    }
 
+    @JRubyMethod(optional = 2, frame = true, visibility = Visibility.PRIVATE)
     public IRubyObject initialize(IRubyObject[] args, Block block) {
-        if (Arity.checkArgumentCount(getRuntime(), args, 0, 1) == 1) message = args[0];
+        if (args.length == 1) message = args[0];
         return this;
     }
 
+    @JRubyMethod
     public IRubyObject backtrace() {
         return getBacktrace(); 
     }
 
+    @JRubyMethod(required = 1)
     public IRubyObject set_backtrace(IRubyObject obj) {
         if (obj.isNil()) {
             backtrace = null;
@@ -170,6 +209,12 @@ public class RubyException extends RubyObject {
         return backtrace();
     }
     
+    @JRubyMethod(name = "exception", optional = 1, rest = true, meta = true)
+    public static IRubyObject exception(ThreadContext context, IRubyObject recv, IRubyObject[] args, Block block) {
+        return ((RubyClass) recv).newInstance(context, args, block);
+    }
+
+    @JRubyMethod(optional = 1)
     public RubyException exception(IRubyObject[] args) {
         switch (args.length) {
             case 0 :
@@ -178,7 +223,7 @@ public class RubyException extends RubyObject {
                 if(args[0] == this) {
                     return this;
                 }
-                RubyException ret = (RubyException)rbClone(Block.NULL_BLOCK);
+                RubyException ret = (RubyException)rbClone();
                 ret.initialize(args, Block.NULL_BLOCK); // This looks wrong, but it's the way MRI does it.
                 return ret;
             default :
@@ -186,53 +231,65 @@ public class RubyException extends RubyObject {
         }
     }
 
-    public IRubyObject to_s() {
-        if (message.isNil()) return getRuntime().newString(getMetaClass().getName());
+    @JRubyMethod(name = "to_s")
+    public IRubyObject to_s(ThreadContext context) {
+        if (message.isNil()) return context.getRuntime().newString(getMetaClass().getName());
         message.setTaint(isTaint());
         return message;
     }
 
-    public IRubyObject to_str() {
-        return callMethod(getRuntime().getCurrentContext(), MethodIndex.TO_S, "to_s");
+    @JRubyMethod(name = "message")
+    public IRubyObject message(ThreadContext context) {
+        return callMethod(context, "to_s");
     }
 
     /** inspects an object and return a kind of debug information
      * 
      *@return A RubyString containing the debug information.
      */
-    public IRubyObject inspect() {
+    @JRubyMethod(name = "inspect")
+    public IRubyObject inspect(ThreadContext context) {
         RubyModule rubyClass = getMetaClass();
-        RubyString exception = RubyString.objAsString(this);
+        RubyString exception = RubyString.objAsString(context, this);
 
-        if (exception.getValue().length() == 0) return getRuntime().newString(rubyClass.getName());
-        StringBuffer sb = new StringBuffer("#<");
-        sb.append(rubyClass.getName()).append(": ").append(exception.getValue()).append(">");
+        if (exception.getByteList().realSize == 0) return getRuntime().newString(rubyClass.getName());
+        StringBuilder sb = new StringBuilder("#<");
+        sb.append(rubyClass.getName()).append(": ").append(exception.getByteList()).append(">");
         return getRuntime().newString(sb.toString());
     }
 
-	public void printBacktrace(PrintStream errorStream) {
-	    IRubyObject backtrace = callMethod(getRuntime().getCurrentContext(), "backtrace");
-	    if (!backtrace.isNil() && backtrace instanceof RubyArray) {
-    		IRubyObject[] elements = ((RubyArray)backtrace.convertToArray()).toJavaArray();
-	
-    		for (int i = 1; i < elements.length; i++) {
-    		    IRubyObject stackTraceLine = elements[i];
-    			if (stackTraceLine instanceof RubyString) {
-    		        printStackTraceLine(errorStream, stackTraceLine);
-    		    }
-	
-    		    if (i == RubyException.TRACE_HEAD && elements.length > RubyException.TRACE_MAX) {
-    		        int hiddenLevels = elements.length - RubyException.TRACE_HEAD - RubyException.TRACE_TAIL;
-    				errorStream.print("\t ... " + hiddenLevels + " levels...\n");
-    		        i = elements.length - RubyException.TRACE_TAIL;
-    		    }
-    		}
-		}
-	}
+    protected void copySpecialInstanceVariables(IRubyObject clone) {
+        RubyException exception = (RubyException)clone;
+        exception.backtraceFrames = backtraceFrames;
+        exception.javaStackTrace = javaStackTrace;
+        exception.backtrace = backtrace;
+        exception.message = message;
+    }
 
-	private void printStackTraceLine(PrintStream errorStream, IRubyObject stackTraceLine) {
-		errorStream.print("\tfrom " + stackTraceLine + '\n');
-	}
+    public void printBacktrace(PrintStream errorStream) {
+        IRubyObject backtrace = callMethod(getRuntime().getCurrentContext(), "backtrace");
+        boolean debug = getRuntime().getDebug().isTrue();
+        if (!backtrace.isNil() && backtrace instanceof RubyArray) {
+            IRubyObject[] elements = backtrace.convertToArray().toJavaArray();
+
+            for (int i = 1; i < elements.length; i++) {
+                IRubyObject stackTraceLine = elements[i];
+                    if (stackTraceLine instanceof RubyString) {
+                    printStackTraceLine(errorStream, stackTraceLine);
+                }
+
+                if (!debug && i == RubyException.TRACE_HEAD && elements.length > RubyException.TRACE_MAX) {
+                    int hiddenLevels = elements.length - RubyException.TRACE_HEAD - RubyException.TRACE_TAIL;
+                            errorStream.print("\t ... " + hiddenLevels + " levels...\n");
+                    i = elements.length - RubyException.TRACE_TAIL;
+                }
+            }
+        }
+    }
+
+    private void printStackTraceLine(PrintStream errorStream, IRubyObject stackTraceLine) {
+            errorStream.print("\tfrom " + stackTraceLine + '\n');
+    }
 	
     private boolean isArrayOfStrings(IRubyObject backtrace) {
         if (!(backtrace instanceof RubyArray)) return false; 
@@ -245,12 +302,4 @@ public class RubyException extends RubyObject {
             
         return true;
     }
-
-    protected IRubyObject doClone() {
-        IRubyObject newObject = new RubyException(getRuntime(),getMetaClass().getRealClass());
-        if (newObject.getType() != getMetaClass().getRealClass()) {
-            throw getRuntime().newTypeError("wrong instance allocation");
-        }
-        return newObject;
-    }
 }
