@@ -1,4 +1,5 @@
-/***** BEGIN LICENSE BLOCK *****
+/*
+ ***** BEGIN LICENSE BLOCK *****
  * Version: CPL 1.0/GPL 2.0/LGPL 2.1
  *
  * The contents of this file are subject to the Common Public
@@ -35,34 +36,47 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby;
 
-import java.text.DecimalFormat;
+import static org.jruby.util.Numeric.f_expt;
+import static org.jruby.util.Numeric.f_mul;
+import static org.jruby.util.Numeric.f_to_i;
+import static org.jruby.util.Numeric.frexp;
+import static org.jruby.util.Numeric.ldexp;
 
-import org.jruby.runtime.CallbackFactory;
+import java.text.DecimalFormat;
+import java.text.DecimalFormatSymbols;
+import java.util.Locale;
+
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
 import org.jruby.runtime.ClassIndex;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
+import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
 
 /**
- *
- * @author  jpetersen
+  * A representation of a float object
  */
+@JRubyClass(name="Float", parent="Numeric", include="Precision")
 public class RubyFloat extends RubyNumeric {
 
     public static RubyClass createFloatClass(Ruby runtime) {
-        RubyClass floatc = runtime.defineClass("Float", runtime.getClass("Numeric"),
-                ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
+        RubyClass floatc = runtime.defineClass("Float", runtime.getNumeric(), ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
+        runtime.setFloat(floatc);
         floatc.index = ClassIndex.FLOAT;
-        
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyFloat.class);
-        floatc.getSingletonClass().undefineMethod("allocate");
+        floatc.kindOf = new RubyModule.KindOf() {
+            public boolean isKindOf(IRubyObject obj, RubyModule type) {
+                return obj instanceof RubyFloat;
+            }
+        };        
+
         floatc.getSingletonClass().undefineMethod("new");
 
-        floatc.getMetaClass().defineFastMethod("induced_from", callbackFactory.getFastSingletonMethod(
-                "induced_from", RubyKernel.IRUBY_OBJECT));
-        floatc.includeModule(runtime.getModule("Precision"));
+        if (runtime.getInstanceConfig().getCompatVersion() == CompatVersion.RUBY1_8) {
+            floatc.includeModule(runtime.getPrecision());
+        }
 
         // Java Doubles are 64 bit long:            
         floatc.defineConstant("ROUNDS", RubyFixnum.newFixnum(runtime, 1));
@@ -74,44 +88,12 @@ public class RubyFloat extends RubyNumeric {
         // Double.MAX_EXPONENT since Java 1.6            
         floatc.defineConstant("MAX_EXP", RubyFixnum.newFixnum(runtime, 1024));
         floatc.defineConstant("MIN_10_EXP", RubyFixnum.newFixnum(runtime, -307));
-        floatc.defineConstant("MAX_10_EXP", RubyFixnum.newFixnum(runtime, -308));
+        floatc.defineConstant("MAX_10_EXP", RubyFixnum.newFixnum(runtime, 308));
         floatc.defineConstant("MIN", RubyFloat.newFloat(runtime, Double.MIN_VALUE));
         floatc.defineConstant("MAX", RubyFloat.newFloat(runtime, Double.MAX_VALUE));
         floatc.defineConstant("EPSILON", RubyFloat.newFloat(runtime, 2.2204460492503131e-16));
-
-        floatc.defineFastMethod("to_s", callbackFactory.getFastMethod("to_s"));
-        floatc.defineFastMethod("coerce", callbackFactory.getFastMethod("coerce", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("-@", callbackFactory.getFastMethod("uminus"));
-        floatc.defineFastMethod("+", callbackFactory.getFastMethod("plus", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("-", callbackFactory.getFastMethod("minus", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("*", callbackFactory.getFastMethod("mul", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("/", callbackFactory.getFastMethod("fdiv", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("%", callbackFactory.getFastMethod("mod", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("modulo", callbackFactory.getFastMethod("mod", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("divmod", callbackFactory.getFastMethod("divmod", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("**", callbackFactory.getFastMethod("pow", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("==", callbackFactory.getFastMethod("equal", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("<=>", callbackFactory.getFastMethod("cmp", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod(">", callbackFactory.getFastMethod("gt", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod(">=", callbackFactory.getFastMethod("ge", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("<", callbackFactory.getFastMethod("lt", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("<=", callbackFactory.getFastMethod("le", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("eql?", callbackFactory.getFastMethod("eql_p", RubyKernel.IRUBY_OBJECT));
-        floatc.defineFastMethod("hash", callbackFactory.getFastMethod("hash"));
-        floatc.defineFastMethod("to_f", callbackFactory.getFastMethod("to_f"));
-        floatc.defineFastMethod("abs", callbackFactory.getFastMethod("abs"));
-        floatc.defineFastMethod("zero?", callbackFactory.getFastMethod("zero_p"));
-
-        floatc.defineFastMethod("to_i", callbackFactory.getFastMethod("truncate"));
-        floatc.defineFastMethod("to_int", callbackFactory.getFastMethod("truncate"));
-        floatc.defineFastMethod("floor", callbackFactory.getFastMethod("floor"));
-        floatc.defineFastMethod("ceil", callbackFactory.getFastMethod("ceil"));
-        floatc.defineFastMethod("round", callbackFactory.getFastMethod("round"));
-        floatc.defineFastMethod("truncate", callbackFactory.getFastMethod("truncate"));
-
-        floatc.defineFastMethod("nan?", callbackFactory.getFastMethod("nan_p"));
-        floatc.defineFastMethod("infinite?", callbackFactory.getFastMethod("infinite_p"));
-        floatc.defineFastMethod("finite?", callbackFactory.getFastMethod("finite_p"));
+        
+        floatc.defineAnnotatedMethods(RubyFloat.class);
 
         return floatc;
     }
@@ -127,12 +109,18 @@ public class RubyFloat extends RubyNumeric {
     }
 
     public RubyFloat(Ruby runtime, double value) {
-        super(runtime, runtime.getClass("Float"));
+        super(runtime, runtime.getFloat());
         this.value = value;
     }
 
-    public Class getJavaClass() {
-        return Double.TYPE;
+    public Class<?> getJavaClass() {
+        // this needs to be thought out more along with the changes in RubyFixnum
+        // since "to Object" coercion will generally want to produce the same
+        // type every time
+//        if (value >= Float.MIN_VALUE && value <= Float.MAX_VALUE) {
+//            return float.class;
+//        }
+        return double.class;
     }
 
     /** Getter for property value.
@@ -171,22 +159,24 @@ public class RubyFloat extends RubyNumeric {
     /** rb_flo_induced_from
      * 
      */
-    public static IRubyObject induced_from(IRubyObject recv, IRubyObject number) {
-        if (number instanceof RubyFixnum || number instanceof RubyBignum) {
-            return number.callMethod(recv.getRuntime().getCurrentContext(), MethodIndex.TO_F, "to_f");
-    }
-        if (number instanceof RubyFloat) {
+    @JRubyMethod(name = "induced_from", meta = true, compat = CompatVersion.RUBY1_8)
+    public static IRubyObject induced_from(ThreadContext context, IRubyObject recv, IRubyObject number) {
+        if (number instanceof RubyFixnum || number instanceof RubyBignum || number instanceof RubyRational) {
+            return number.callMethod(context, "to_f");
+        } else if (number instanceof RubyFloat) {
             return number;
         }
         throw recv.getRuntime().newTypeError(
                 "failed to convert " + number.getMetaClass() + " into Float");
     }
 
-    private final static DecimalFormat FORMAT = new DecimalFormat("##############0.0##############");
+    private final static DecimalFormat FORMAT = new DecimalFormat("##############0.0##############",
+            new DecimalFormatSymbols(Locale.ENGLISH));
 
     /** flo_to_s
      * 
      */
+    @JRubyMethod(name = "to_s")
     public IRubyObject to_s() {
         if (Double.isInfinite(value)) {
             return RubyString.newString(getRuntime(), value < 0 ? "-Infinity" : "Infinity");
@@ -217,64 +207,97 @@ public class RubyFloat extends RubyNumeric {
     /** flo_coerce
      * 
      */
+    @JRubyMethod(name = "coerce", required = 1)
     public IRubyObject coerce(IRubyObject other) {
-        // MRI doesn't type check here either
-        return getRuntime().newArray(
-                newFloat(getRuntime(), ((RubyNumeric) other).getDoubleValue()), this);
-        }
+        return getRuntime().newArray(RubyKernel.new_float(this, other), this);
+    }
 
     /** flo_uminus
      * 
      */
-    public IRubyObject uminus() {
+    @JRubyMethod(name = "-@")
+    public IRubyObject op_uminus() {
         return RubyFloat.newFloat(getRuntime(), -value);
-        }
+    }
 
     /** flo_plus
      * 
      */
-    public IRubyObject plus(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "+", required = 1)
+    public IRubyObject op_plus(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             return RubyFloat.newFloat(getRuntime(), value + ((RubyNumeric) other).getDoubleValue());
+        default:
+            return coerceBin(context, "+", other);
         }
-        return coerceBin("+", other);
     }
 
     /** flo_minus
      * 
      */
-    public IRubyObject minus(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "-", required = 1)
+    public IRubyObject op_minus(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             return RubyFloat.newFloat(getRuntime(), value - ((RubyNumeric) other).getDoubleValue());
-    }
-        return coerceBin("-", other);
+        default:
+            return coerceBin(context, "-", other);
+        }
     }
 
     /** flo_mul
      * 
      */
-    public IRubyObject mul(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
-            return RubyFloat.newFloat(getRuntime(), value * ((RubyNumeric) other).getDoubleValue());
-    }
-        return coerceBin("*", other);
+    @JRubyMethod(name = "*", required = 1)
+    public IRubyObject op_mul(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
+            return RubyFloat.newFloat(
+                    getRuntime(), value * ((RubyNumeric) other).getDoubleValue());
+        default:
+            return coerceBin(context, "*", other);
+        }
     }
     
     /** flo_div
      * 
      */
-    public IRubyObject fdiv(IRubyObject other) { // don't override Numeric#div !
-    	if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "/", required = 1)
+    public IRubyObject op_fdiv(ThreadContext context, IRubyObject other) { // don't override Numeric#div !
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             return RubyFloat.newFloat(getRuntime(), value / ((RubyNumeric) other).getDoubleValue());
-    	}
-        return coerceBin("div", other);
+        default:
+            return coerceBin(context, "/", other);
+        }
+    }
+
+    /** flo_quo
+    *
+    */
+    @JRubyMethod(name = "quo", compat = CompatVersion.RUBY1_9)
+        public IRubyObject magnitude(ThreadContext context, IRubyObject other) {
+        return callMethod(context, "/", other);
     }
 
     /** flo_mod
      * 
      */
-    public IRubyObject mod(IRubyObject other) {
-    	if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = {"%", "modulo"}, required = 1)
+    public IRubyObject op_mod(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double y = ((RubyNumeric) other).getDoubleValue();
             // Modelled after c ruby implementation (java /,% not same as ruby)
             double x = value;
@@ -285,134 +308,179 @@ public class RubyFloat extends RubyNumeric {
             }
 
             return RubyFloat.newFloat(getRuntime(), mod);
-    	}
-        return coerceBin("%", other);
+        default:
+            return coerceBin(context, "%", other);
+        }
     }
 
     /** flo_divmod
      * 
      */
-    public IRubyObject divmod(IRubyObject other) {
-    	if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "divmod", required = 1)
+    public IRubyObject divmod(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double y = ((RubyNumeric) other).getDoubleValue();
             double x = value;
 
             double mod = Math.IEEEremainder(x, y);
-            double div = (x - mod) / y;
+            // MRI behavior:
+            if (Double.isNaN(mod)) {
+                throw getRuntime().newFloatDomainError("NaN");
+            }
+            double div = Math.floor(x / y);
 
             if (y * mod < 0) {
                 mod += y;
-                div -= 1.0;
-    	}
+            }
             final Ruby runtime = getRuntime();
             IRubyObject car = dbl2num(runtime, div);
             RubyFloat cdr = RubyFloat.newFloat(runtime, mod);
             return RubyArray.newArray(runtime, car, cdr);
+        default:
+            return coerceBin(context, "divmod", other);
+        }
     }
-        return coerceBin("%", other);
-    	}
     	
     /** flo_pow
      * 
      */
-    public IRubyObject pow(IRubyObject other) {
-    	if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "**", required = 1)
+    public IRubyObject op_pow(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             return RubyFloat.newFloat(getRuntime(), Math.pow(value, ((RubyNumeric) other)
                     .getDoubleValue()));
-    	}
-        return coerceBin("/", other);
+        default:
+            return coerceBin(context, "**", other);
+        }
     }
 
     /** flo_eq
      * 
      */
-    public IRubyObject equal(IRubyObject other) {
+    @JRubyMethod(name = "==", required = 1)
+    public IRubyObject op_equal(ThreadContext context, IRubyObject other) {
         if (Double.isNaN(value)) {
             return getRuntime().getFalse();
-    }
-        if (other instanceof RubyNumeric) {
+        }
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             return RubyBoolean.newBoolean(getRuntime(), value == ((RubyNumeric) other)
                     .getDoubleValue());
-    }
-        // Numeric.equal            
-        return super.equal(other);
-
+        default:
+            // Numeric.equal            
+            return super.op_num_equal(context, other);
+        }
     }
 
     /** flo_cmp
      * 
      */
-    public IRubyObject cmp(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "<=>", required = 1)
+    public IRubyObject op_cmp(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double b = ((RubyNumeric) other).getDoubleValue();
             return dbl_cmp(getRuntime(), value, b);
-    	}
-        return coerceCmp("<=>", other);
+        default:
+            return coerceCmp(context, "<=>", other);
+        }
     }
 
     /** flo_gt
      * 
      */
-    public IRubyObject gt(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = ">", required = 1)
+    public IRubyObject op_gt(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double b = ((RubyNumeric) other).getDoubleValue();
             return RubyBoolean.newBoolean(getRuntime(), !Double.isNaN(b) && value > b);
+        default:
+            return coerceRelOp(context, ">", other);
         }
-        return coerceRelOp(">", other);
     }
 
     /** flo_ge
      * 
      */
-    public IRubyObject ge(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = ">=", required = 1)
+    public IRubyObject op_ge(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double b = ((RubyNumeric) other).getDoubleValue();
             return RubyBoolean.newBoolean(getRuntime(), !Double.isNaN(b) && value >= b);
+        default:
+            return coerceRelOp(context, ">=", other);
         }
-        return coerceRelOp(">=", other);
-        }
+    }
 
     /** flo_lt
      * 
      */
-    public IRubyObject lt(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "<", required = 1)
+    public IRubyObject op_lt(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double b = ((RubyNumeric) other).getDoubleValue();
             return RubyBoolean.newBoolean(getRuntime(), !Double.isNaN(b) && value < b);
-    }
-        return coerceRelOp("<", other);
+        default:
+            return coerceRelOp(context, "<", other);
+		}
     }
 
     /** flo_le
      * 
      */
-    public IRubyObject le(IRubyObject other) {
-        if (other instanceof RubyNumeric) {
+    @JRubyMethod(name = "<=", required = 1)
+    public IRubyObject op_le(ThreadContext context, IRubyObject other) {
+        switch (other.getMetaClass().index) {
+        case ClassIndex.FIXNUM:
+        case ClassIndex.BIGNUM:
+        case ClassIndex.FLOAT:
             double b = ((RubyNumeric) other).getDoubleValue();
             return RubyBoolean.newBoolean(getRuntime(), !Double.isNaN(b) && value <= b);
+        default:
+            return coerceRelOp(context, "<=", other);
 		}
-        return coerceRelOp("<=", other);
 	}
 	
     /** flo_eql
      * 
      */
+    @JRubyMethod(name = "eql?", required = 1)
     public IRubyObject eql_p(IRubyObject other) {
         if (other instanceof RubyFloat) {
             double b = ((RubyFloat) other).value;
             if (Double.isNaN(value) || Double.isNaN(b)) {
-            return getRuntime().getFalse();
-        }
+                return getRuntime().getFalse();
+            }
             if (value == b) {
                 return getRuntime().getTrue();
-    }
             }
-            return getRuntime().getFalse();
         }
+        return getRuntime().getFalse();
+    }
 
     /** flo_hash
      * 
      */
+    @JRubyMethod(name = "hash")
     public RubyFixnum hash() {
         return getRuntime().newFixnum(hashCode());
     }
@@ -425,44 +493,73 @@ public class RubyFloat extends RubyNumeric {
     /** flo_fo 
      * 
      */
+    @JRubyMethod(name = "to_f")
     public IRubyObject to_f() {
         return this;
-        }
-        
+    }
+
     /** flo_abs
      * 
      */
-    public IRubyObject abs() {
-        if (value < 0) {
-            return RubyFloat.newFloat(getRuntime(), Math.abs(value));
+    @JRubyMethod(name = "abs")
+    public IRubyObject abs(ThreadContext context) {
+        if (Double.doubleToLongBits(value) < 0) {
+            return RubyFloat.newFloat(context.getRuntime(), Math.abs(value));
         }
         return this;
     }
-    
+
+    /** flo_abs/1.9
+     * 
+     */
+    @JRubyMethod(name = "magnitude", compat = CompatVersion.RUBY1_9)
+    public IRubyObject magnitude(ThreadContext context) {
+        return abs(context);
+    }
+
     /** flo_zero_p
      * 
      */
+    @JRubyMethod(name = "zero?")
     public IRubyObject zero_p() {
         return RubyBoolean.newBoolean(getRuntime(), value == 0.0);
-        }
+    }
 
     /** flo_truncate
      * 
      */
+    @JRubyMethod(name = {"truncate", "to_i", "to_int"})
     public IRubyObject truncate() {
         double f = value;
-        if (f > 0.0) {
-            f = Math.floor(f);
-            }
-        if (f > 0.0) {
-            f = Math.ceil(f);
-        }
+        if (f > 0.0) f = Math.floor(f);
+        if (f < 0.0) f = Math.ceil(f);
+
         return dbl2num(getRuntime(), f);
-        }
-        
-    /** loor
+    }
+
+    /** float_to_r, float_decode
      * 
      */
+    static final int DBL_MANT_DIG = 53;
+    static final int FLT_RADIX = 2;
+    @JRubyMethod(name = "to_r", compat = CompatVersion.RUBY1_9)
+    public IRubyObject to_r(ThreadContext context) {
+        long[]exp = new long[1]; 
+        double f = frexp(value, exp);
+        f = ldexp(f, DBL_MANT_DIG);
+        long n = exp[0] - DBL_MANT_DIG;
+
+        Ruby runtime = context.getRuntime();
+
+        IRubyObject rf = RubyNumeric.dbl2num(runtime, f);
+        IRubyObject rn = RubyFixnum.newFixnum(runtime, n);
+        return f_mul(context, rf, f_expt(context, RubyFixnum.newFixnum(runtime, FLT_RADIX), rn));
+    }
+
+    /** floor
+     * 
+     */
+    @JRubyMethod(name = "floor")
     public IRubyObject floor() {
         return dbl2num(getRuntime(), Math.floor(value));
     }
@@ -470,27 +567,30 @@ public class RubyFloat extends RubyNumeric {
     /** flo_ceil
      * 
      */
+    @JRubyMethod(name = "ceil")
     public IRubyObject ceil() {
         return dbl2num(getRuntime(), Math.ceil(value));
-        }
+    }
 
     /** flo_round
      * 
      */
+    @JRubyMethod(name = "round")
     public IRubyObject round() {
         double f = value;
         if (f > 0.0) {
             f = Math.floor(f + 0.5);
-            }
+        }
         if (f < 0.0) {
             f = Math.ceil(f - 0.5);
         }
         return dbl2num(getRuntime(), f);
-        }
+    }
         
     /** flo_is_nan_p
      * 
      */
+    @JRubyMethod(name = "nan?")
     public IRubyObject nan_p() {
         return RubyBoolean.newBoolean(getRuntime(), Double.isNaN(value));
     }
@@ -498,16 +598,18 @@ public class RubyFloat extends RubyNumeric {
     /** flo_is_infinite_p
      * 
      */
+    @JRubyMethod(name = "infinite?")
     public IRubyObject infinite_p() {
         if (Double.isInfinite(value)) {
             return RubyFixnum.newFixnum(getRuntime(), value < 0 ? -1 : 1);
         }
         return getRuntime().getNil();
-            }
+    }
             
     /** flo_is_finite_p
      * 
      */
+    @JRubyMethod(name = "finite?")
     public IRubyObject finite_p() {
         if (Double.isInfinite(value) || Double.isNaN(value)) {
             return getRuntime().getFalse();
@@ -516,6 +618,8 @@ public class RubyFloat extends RubyNumeric {
     }
 
     public static void marshalTo(RubyFloat aFloat, MarshalStream output) throws java.io.IOException {
+        output.registerLinkTarget(aFloat);
+
         String strValue = aFloat.toString();
     
         if (Double.isInfinite(aFloat.value)) {
@@ -531,5 +635,4 @@ public class RubyFloat extends RubyNumeric {
         input.registerLinkTarget(result);
         return result;
     }
-    
 }
