@@ -35,15 +35,20 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby;
 
+import org.jruby.util.io.STDIO;
 import java.util.HashMap;
 import java.util.Map;
 
+import org.jruby.anno.JRubyMethod;
+import org.jruby.common.IRubyWarnings.ID;
 import org.jruby.environment.OSEnvironmentReaderExcepton;
 import org.jruby.environment.OSEnvironment;
 import org.jruby.internal.runtime.ValueAccessor;
+import org.jruby.javasupport.JavaUtil;
+import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Constants;
 import org.jruby.runtime.GlobalVariable;
-import org.jruby.runtime.MethodIndex;
+import org.jruby.runtime.IAccessor;
 import org.jruby.runtime.ReadonlyGlobalVariable;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
@@ -59,21 +64,27 @@ public class RubyGlobal {
      * Obligate string-keyed and string-valued hash, used for ENV and ENV_JAVA
      * 
      */
-    private static class StringOnlyRubyHash extends RubyHash {
+    public static class StringOnlyRubyHash extends RubyHash {
         
         public StringOnlyRubyHash(Ruby runtime, Map valueMap, IRubyObject defaultValue) {
             super(runtime, valueMap, defaultValue);
         }
 
-        public IRubyObject aref(IRubyObject key) {
-            if (!key.respondsTo("to_str")) {
-                throw getRuntime().newTypeError("can't convert " + key.getMetaClass() + " into String");
-            }
-
-            return super.aref(key.callMethod(getRuntime().getCurrentContext(), MethodIndex.TO_STR, "to_str", IRubyObject.NULL_ARRAY));
+        @Override
+        public RubyHash to_hash() {
+            Ruby runtime = getRuntime();
+            RubyHash hash = RubyHash.newHash(runtime);
+            hash.replace(runtime.getCurrentContext(), this);
+            return hash;
         }
 
-        public IRubyObject aset(IRubyObject key, IRubyObject value) {
+        @Override
+        public IRubyObject op_aref(ThreadContext context, IRubyObject key) {
+            return super.op_aref(context, key.convertToString());
+        }
+
+        @Override
+        public IRubyObject op_aset(ThreadContext context, IRubyObject key, IRubyObject value) {
             if (!key.respondsTo("to_str")) {
                 throw getRuntime().newTypeError("can't convert " + key.getMetaClass() + " into String");
             }
@@ -82,42 +93,69 @@ public class RubyGlobal {
             }
 
             if (value.isNil()) {
-                return super.delete(key, org.jruby.runtime.Block.NULL_BLOCK);
+                return super.delete(context, key, org.jruby.runtime.Block.NULL_BLOCK);
             }
             
-            ThreadContext context = getRuntime().getCurrentContext();
             //return super.aset(getRuntime().newString("sadfasdF"), getRuntime().newString("sadfasdF"));
-            return super.aset(key.callMethod(context, MethodIndex.TO_STR, "to_str", IRubyObject.NULL_ARRAY),
-                    value.isNil() ? getRuntime().getNil() : value.callMethod(context, MethodIndex.TO_STR, "to_str", IRubyObject.NULL_ARRAY));
+            return super.op_aset(context, RuntimeHelpers.invoke(context, key, "to_str"),
+                    value.isNil() ? getRuntime().getNil() : RuntimeHelpers.invoke(context, value, "to_str"));
         }
         
+        @JRubyMethod
+        @Override
         public IRubyObject to_s(){
             return getRuntime().newString("ENV");
-    }
+        }
     }
     
-    public static void createGlobals(Ruby runtime) {
+    public static void createGlobals(ThreadContext context, Ruby runtime) {
+        runtime.defineGlobalConstant("TOPLEVEL_BINDING", runtime.newBinding());
+        
+        runtime.defineGlobalConstant("TRUE", runtime.getTrue());
+        runtime.defineGlobalConstant("FALSE", runtime.getFalse());
+        runtime.defineGlobalConstant("NIL", runtime.getNil());
+        
+        // define ARGV and $* for this runtime
+        RubyArray argvArray = runtime.newArray();
+        String[] argv = runtime.getInstanceConfig().getArgv();
+        for (int i = 0; i < argv.length; i++) {
+            argvArray.append(RubyString.newString(runtime, argv[i].getBytes()));
+        }
+        runtime.defineGlobalConstant("ARGV", argvArray);
+        runtime.getGlobalVariables().defineReadonly("$*", new ValueAccessor(argvArray));
+
+        IAccessor d = new ValueAccessor(runtime.newString(
+                runtime.getInstanceConfig().displayedFileName()));
+        runtime.getGlobalVariables().define("$PROGRAM_NAME", d);
+        runtime.getGlobalVariables().define("$0", d);
 
         // Version information:
-        IRubyObject version = runtime.newString(Constants.RUBY_VERSION).freeze();
-        IRubyObject release = runtime.newString(Constants.COMPILE_DATE).freeze();
-        IRubyObject platform = runtime.newString(Constants.PLATFORM).freeze();
+        IRubyObject version = runtime.newString(Constants.RUBY_VERSION).freeze(context);
+        IRubyObject release = runtime.newString(Constants.COMPILE_DATE).freeze(context);
+        IRubyObject platform = runtime.newString(Constants.PLATFORM).freeze(context);
+        IRubyObject engine = runtime.newString(Constants.ENGINE).freeze(context);
 
         runtime.defineGlobalConstant("RUBY_VERSION", version);
+        runtime.defineGlobalConstant("RUBY_PATCHLEVEL", runtime.newString(Constants.RUBY_PATCHLEVEL).freeze(context));
         runtime.defineGlobalConstant("RUBY_RELEASE_DATE", release);
         runtime.defineGlobalConstant("RUBY_PLATFORM", platform);
+        runtime.defineGlobalConstant("RUBY_ENGINE", engine);
 
         runtime.defineGlobalConstant("VERSION", version);
         runtime.defineGlobalConstant("RELEASE_DATE", release);
         runtime.defineGlobalConstant("PLATFORM", platform);
         
-        IRubyObject jrubyVersion = runtime.newString(Constants.VERSION).freeze();
+        IRubyObject jrubyVersion = runtime.newString(Constants.VERSION).freeze(context);
         runtime.defineGlobalConstant("JRUBY_VERSION", jrubyVersion);
 		
         GlobalVariable kcodeGV = new KCodeGlobalVariable(runtime, "$KCODE", runtime.newString("NONE"));
         runtime.defineVariable(kcodeGV);
         runtime.defineVariable(new GlobalVariable.Copy(runtime, "$-K", kcodeGV));
-        runtime.defineVariable(new StringGlobalVariable(runtime, "$/", runtime.newString("\n").freeze()));
+        IRubyObject defaultRS = runtime.newString(runtime.getInstanceConfig().getRecordSeparator()).freeze(context);
+        GlobalVariable rs = new StringGlobalVariable(runtime, "$/", defaultRS);
+        runtime.defineVariable(rs);
+        runtime.setRecordSeparatorVar(rs);
+        runtime.getGlobalVariables().setDefaultSeparator(defaultRS);
         runtime.defineVariable(new StringGlobalVariable(runtime, "$\\", runtime.getNil()));
         runtime.defineVariable(new StringGlobalVariable(runtime, "$,", runtime.getNil()));
 
@@ -126,23 +164,45 @@ public class RubyGlobal {
         runtime.defineVariable(new LastExitStatusVariable(runtime, "$?"));
 
         runtime.defineVariable(new ErrorInfoGlobalVariable(runtime, "$!", runtime.getNil()));
-        runtime.defineVariable(new GlobalVariable(runtime, "$=", runtime.getFalse()));
+        runtime.defineVariable(new NonEffectiveGlobalVariable(runtime, "$=", runtime.getFalse()));
+
+        if(runtime.getInstanceConfig().getInputFieldSeparator() == null) {
+            runtime.defineVariable(new GlobalVariable(runtime, "$;", runtime.getNil()));
+        } else {
+            runtime.defineVariable(new GlobalVariable(runtime, "$;", RubyRegexp.newRegexp(runtime, runtime.getInstanceConfig().getInputFieldSeparator(), 0)));
+        }
+        
+        Boolean verbose = runtime.getInstanceConfig().getVerbose();
+        IRubyObject verboseValue = null;
+        if (verbose == null) {
+            verboseValue = runtime.getNil();
+        } else if(verbose == Boolean.TRUE) {
+            verboseValue = runtime.getTrue();
+        } else {
+            verboseValue = runtime.getFalse();
+        }
+        runtime.defineVariable(new VerboseGlobalVariable(runtime, "$VERBOSE", verboseValue));
+        
+        IRubyObject debug = runtime.newBoolean(runtime.getInstanceConfig().isDebug());
+        runtime.defineVariable(new DebugGlobalVariable(runtime, "$DEBUG", debug));
+        runtime.defineVariable(new DebugGlobalVariable(runtime, "$-d", debug));
 
         runtime.defineVariable(new SafeGlobalVariable(runtime, "$SAFE"));
 
         runtime.defineVariable(new BacktraceGlobalVariable(runtime, "$@"));
 
-        IRubyObject stdin = RubyIO.fdOpen(runtime, RubyIO.STDIN);
-        IRubyObject stdout = RubyIO.fdOpen(runtime, RubyIO.STDOUT);
-        IRubyObject stderr = RubyIO.fdOpen(runtime, RubyIO.STDERR);
+        IRubyObject stdin = new RubyIO(runtime, STDIO.IN);
+        IRubyObject stdout = new RubyIO(runtime, STDIO.OUT);
+        IRubyObject stderr = new RubyIO(runtime, STDIO.ERR);
 
         runtime.defineVariable(new InputGlobalVariable(runtime, "$stdin", stdin));
 
         runtime.defineVariable(new OutputGlobalVariable(runtime, "$stdout", stdout));
+        runtime.getGlobalVariables().alias("$>", "$stdout");
+        runtime.getGlobalVariables().alias("$defout", "$stdout");
+
         runtime.defineVariable(new OutputGlobalVariable(runtime, "$stderr", stderr));
-        runtime.defineVariable(new OutputGlobalVariable(runtime, "$>", stdout));
-        runtime.defineVariable(new OutputGlobalVariable(runtime, "$defout", stdout));
-        runtime.defineVariable(new OutputGlobalVariable(runtime, "$deferr", stderr));
+        runtime.getGlobalVariables().alias("$deferr", "$stderr");
 
         runtime.defineGlobalConstant("STDIN", stdin);
         runtime.defineGlobalConstant("STDOUT", stdout);
@@ -159,7 +219,12 @@ public class RubyGlobal {
         runtime.defineVariable(new PreMatchGlobalVariable(runtime, "$`"));
         runtime.defineVariable(new PostMatchGlobalVariable(runtime, "$'"));
         runtime.defineVariable(new LastMatchGlobalVariable(runtime, "$+"));
-        runtime.defineVariable(new LastMatchInfoGlobalVariable(runtime, "$~"));
+        runtime.defineVariable(new BackRefGlobalVariable(runtime, "$~"));
+
+        // On platforms without a c-library accessable through JNA, getpid will return hashCode 
+        // as $$ used to. Using $$ to kill processes could take down many runtimes, but by basing
+        // $$ on getpid() where available, we have the same semantics as MRI.
+        runtime.getGlobalVariables().defineReadonly("$$", new ValueAccessor(runtime.newFixnum(runtime.getPosix().getpid())));
 
         // after defn of $stderr as the call may produce warnings
         defineGlobalEnvConstants(runtime);
@@ -168,9 +233,18 @@ public class RubyGlobal {
         if (runtime.getGlobalVariables().get("$*").isNil()) {
             runtime.getGlobalVariables().defineReadonly("$*", new ValueAccessor(runtime.newArray()));
         }
+        
+        runtime.getGlobalVariables().defineReadonly("$-p", 
+                new ValueAccessor(runtime.getInstanceConfig().isAssumePrinting() ? runtime.getTrue() : runtime.getNil()));
+        runtime.getGlobalVariables().defineReadonly("$-n", 
+                new ValueAccessor(runtime.getInstanceConfig().isAssumeLoop() ? runtime.getTrue() : runtime.getNil()));
+        runtime.getGlobalVariables().defineReadonly("$-a", 
+                new ValueAccessor(runtime.getInstanceConfig().isSplit() ? runtime.getTrue() : runtime.getNil()));
+        runtime.getGlobalVariables().defineReadonly("$-l", 
+                new ValueAccessor(runtime.getInstanceConfig().isProcessLineEnds() ? runtime.getTrue() : runtime.getNil()));
 
         // ARGF, $< object
-        new RubyArgsFile(runtime).initArgsFile();
+        RubyArgsFile.initArgsFile(runtime);
     }
 
     private static void defineGlobalEnvConstants(Ruby runtime) {
@@ -181,7 +255,7 @@ public class RubyGlobal {
     		environmentVariableMap = environment.getEnvironmentVariableMap(runtime);
     	} catch (OSEnvironmentReaderExcepton e) {
     		// If the environment variables are not accessible shouldn't terminate 
-    		runtime.getWarnings().warn(e.getMessage());
+    		runtime.getWarnings().warn(ID.MISCELLANEOUS, e.getMessage());
     	}
 		
     	if (environmentVariableMap == null) {
@@ -191,8 +265,7 @@ public class RubyGlobal {
 
         StringOnlyRubyHash h1 = new StringOnlyRubyHash(runtime,
                                                        environmentVariableMap, runtime.getNil());
-        org.jruby.runtime.CallbackFactory cf = org.jruby.runtime.CallbackFactory.createFactory(runtime, StringOnlyRubyHash.class);
-        h1.getSingletonClass().defineFastMethod("to_s", cf.getFastMethod("to_s"));
+        h1.getSingletonClass().defineAnnotatedMethods(StringOnlyRubyHash.class);
         runtime.defineGlobalConstant("ENV", h1);
 
         // Define System.getProperties() in ENV_JAVA
@@ -202,15 +275,36 @@ public class RubyGlobal {
         
     }
 
+    private static class NonEffectiveGlobalVariable extends GlobalVariable {
+        public NonEffectiveGlobalVariable(Ruby runtime, String name, IRubyObject value) {
+            super(runtime, name, value);
+        }
+
+        @Override
+        public IRubyObject set(IRubyObject value) {
+            runtime.getWarnings().warn(ID.INEFFECTIVE_GLOBAL, "warning: variable " + name + " is no longer effective; ignored", name);
+            return value;
+        }
+
+        @Override
+        public IRubyObject get() {
+            runtime.getWarnings().warn(ID.INEFFECTIVE_GLOBAL, "warning: variable " + name + " is no longer effective", name);
+            return runtime.getFalse();
+        }
+    }
+
     private static class LastExitStatusVariable extends GlobalVariable {
         public LastExitStatusVariable(Ruby runtime, String name) {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
-            return runtime.getCurrentContext().getLastExitStatus();
+            IRubyObject lastExitStatus = runtime.getCurrentContext().getLastExitStatus();
+            return lastExitStatus == null ? runtime.getNil() : lastExitStatus;
         }
         
+        @Override
         public IRubyObject set(IRubyObject lastExitStatus) {
             runtime.getCurrentContext().setLastExitStatus(lastExitStatus);
             
@@ -223,8 +317,9 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
-            return RubyRegexp.last_match(runtime.getCurrentContext().getBackref());
+            return RubyRegexp.last_match(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
     }
 
@@ -233,8 +328,9 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
-            return RubyRegexp.match_pre(runtime.getCurrentContext().getBackref());
+            return RubyRegexp.match_pre(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
     }
 
@@ -243,8 +339,9 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
-            return RubyRegexp.match_post(runtime.getCurrentContext().getBackref());
+            return RubyRegexp.match_post(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
     }
 
@@ -253,18 +350,26 @@ public class RubyGlobal {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
-            return RubyRegexp.match_last(runtime.getCurrentContext().getBackref());
+            return RubyRegexp.match_last(runtime.getCurrentContext().getCurrentFrame().getBackRef());
         }
     }
 
-    private static class LastMatchInfoGlobalVariable extends GlobalVariable {
-        public LastMatchInfoGlobalVariable(Ruby runtime, String name) {
+    private static class BackRefGlobalVariable extends GlobalVariable {
+        public BackRefGlobalVariable(Ruby runtime, String name) {
             super(runtime, name, runtime.getNil());
         }
         
+        @Override
         public IRubyObject get() {
-            return runtime.getCurrentContext().getBackref();
+            return RuntimeHelpers.getBackref(runtime, runtime.getCurrentContext());
+        }
+
+        @Override
+        public IRubyObject set(IRubyObject value) {
+            RuntimeHelpers.setBackref(runtime, runtime.getCurrentContext(), value);
+            return value;
         }
     }
 
@@ -275,8 +380,9 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
-            ((RubyArgsFile) runtime.getGlobalVariables().get("$<")).setCurrentLineNumber(RubyNumeric.fix2int(value));
+            RubyArgsFile.setCurrentLineNumber(runtime.getGlobalVariables().get("$<"),RubyNumeric.fix2int(value));
             return super.set(value);
         }
     }
@@ -287,14 +393,18 @@ public class RubyGlobal {
             set(value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
-            if (!value.isNil() && ! value.isKindOf(runtime.getClass("Exception"))) {
+            if (!value.isNil() &&
+                    !runtime.getException().isInstance(value) &&
+                    !(JavaUtil.isJavaObject(value) && JavaUtil.unwrapJavaObject(value) instanceof Exception)) {
                 throw runtime.newTypeError("assigning non-exception to $!");
             }
             
             return runtime.getCurrentContext().setErrorInfo(value);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.getCurrentContext().getErrorInfo();
         }
@@ -306,6 +416,7 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (!value.isNil() && ! (value instanceof RubyString)) {
                 throw runtime.newTypeError("value of " + name() + " must be a String");
@@ -319,15 +430,14 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.getKCode().kcode(runtime);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
-            if (!value.isNil() && ! (value instanceof RubyString)) {
-                throw runtime.newTypeError("value of " + name() + " must be a String");
-            }
-            runtime.setKCode(KCode.create(runtime, value.toString()));
+            runtime.setKCode(KCode.create(runtime, value.convertToString().toString()));
             return value;
         }
     }
@@ -337,19 +447,68 @@ public class RubyGlobal {
             super(runtime, name, null);
         }
 
+        @Override
         public IRubyObject get() {
             return runtime.newFixnum(runtime.getSafeLevel());
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
-            int level = RubyNumeric.fix2int(value);
-            if (level < runtime.getSafeLevel()) {
-            	throw runtime.newSecurityError("tried to downgrade safe level from " + 
-            			runtime.getSafeLevel() + " to " + level);
-            }
-            runtime.setSafeLevel(level);
+//            int level = RubyNumeric.fix2int(value);
+//            if (level < runtime.getSafeLevel()) {
+//            	throw runtime.newSecurityError("tried to downgrade safe level from " + 
+//            			runtime.getSafeLevel() + " to " + level);
+//            }
+//            runtime.setSafeLevel(level);
             // thread.setSafeLevel(level);
-            return value;
+            runtime.getWarnings().warn(ID.SAFE_NOT_SUPPORTED, "SAFE levels are not supported in JRuby");
+            return RubyFixnum.newFixnum(runtime, runtime.getSafeLevel());
+        }
+    }
+
+    private static class VerboseGlobalVariable extends GlobalVariable {
+        public VerboseGlobalVariable(Ruby runtime, String name, IRubyObject initialValue) {
+            super(runtime, name, initialValue);
+            set(initialValue);
+        }
+        
+        @Override
+        public IRubyObject get() {
+            return runtime.getVerbose();
+        }
+
+        @Override
+        public IRubyObject set(IRubyObject newValue) {
+            if (newValue.isNil()) {
+                runtime.setVerbose(newValue);
+            } else {
+                runtime.setVerbose(runtime.newBoolean(newValue.isTrue()));
+            }
+
+            return newValue;
+        }
+    }
+
+    private static class DebugGlobalVariable extends GlobalVariable {
+        public DebugGlobalVariable(Ruby runtime, String name, IRubyObject initialValue) {
+            super(runtime, name, initialValue);
+            set(initialValue);
+        }
+
+        @Override
+        public IRubyObject get() {
+            return runtime.getDebug();
+        }
+
+        @Override
+        public IRubyObject set(IRubyObject newValue) {
+            if (newValue.isNil()) {
+                runtime.setDebug(newValue);
+            } else {
+                runtime.setDebug(runtime.newBoolean(newValue.isTrue()));
+            }
+
+            return newValue;
         }
     }
 
@@ -358,6 +517,7 @@ public class RubyGlobal {
             super(runtime, name, null);
         }
 
+        @Override
         public IRubyObject get() {
             IRubyObject errorInfo = runtime.getGlobalVariables().get("$!");
             IRubyObject backtrace = errorInfo.isNil() ? runtime.getNil() : errorInfo.callMethod(errorInfo.getRuntime().getCurrentContext(), "backtrace");
@@ -368,6 +528,7 @@ public class RubyGlobal {
             return backtrace;
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (runtime.getGlobalVariables().get("$!").isNil()) {
                 throw runtime.newArgumentError("$! not set.");
@@ -382,12 +543,14 @@ public class RubyGlobal {
             super(runtime, name, null);
         }
 
+        @Override
         public IRubyObject get() {
-            return runtime.getCurrentContext().getLastline();
+            return RuntimeHelpers.getLastLine(runtime, runtime.getCurrentContext());
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
-            runtime.getCurrentContext().setLastline(value);
+            RuntimeHelpers.setLastLine(runtime, runtime.getCurrentContext(), value);
             return value;
         }
     }
@@ -397,13 +560,12 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (value == get()) {
                 return value;
             }
-            if (value instanceof RubyIO) {
-                ((RubyIO) value).checkReadable();
-            }
+            
             return super.set(value);
         }
     }
@@ -413,26 +575,28 @@ public class RubyGlobal {
             super(runtime, name, value);
         }
 
+        @Override
         public IRubyObject set(IRubyObject value) {
             if (value == get()) {
                 return value;
             }
             if (value instanceof RubyIO) {
-                ((RubyIO) value).checkWriteable();
+                RubyIO io = (RubyIO)value;
+                
+                // HACK: in order to have stdout/err act like ttys and flush always,
+                // we set anything assigned to stdout/stderr to sync
+                io.getHandler().setSync(true);
             }
-            if (! value.respondsTo("write")) {
+
+            if (!value.respondsTo("write")) {
                 throw runtime.newTypeError(name() + " must have write method, " +
                                     value.getType().getName() + " given");
             }
-            
-            if ("$stdout".equals(name())) {
-                runtime.defineVariable(new OutputGlobalVariable(runtime, "$>", value));
-            }
 
             return super.set(value);
         }
     }
-    
+
     private static class LoadPath extends ReadonlyGlobalVariable {
         public LoadPath(Ruby runtime, String name) {
             super(runtime, name, null);
@@ -441,6 +605,7 @@ public class RubyGlobal {
         /**
          * @see org.jruby.runtime.GlobalVariable#get()
          */
+        @Override
         public IRubyObject get() {
             return runtime.getLoadService().getLoadPath();
         }
@@ -454,6 +619,7 @@ public class RubyGlobal {
         /**
          * @see org.jruby.runtime.GlobalVariable#get()
          */
+        @Override
         public IRubyObject get() {
             return runtime.getLoadService().getLoadedFeatures();
         }
