@@ -39,15 +39,20 @@ import java.nio.charset.IllegalCharsetNameException;
 import java.nio.charset.MalformedInputException;
 import java.nio.charset.UnmappableCharacterException;
 import java.nio.charset.UnsupportedCharsetException;
+
+import org.jruby.anno.JRubyMethod;
+import org.jruby.anno.JRubyModule;
+import org.jruby.anno.JRubyClass;
 import org.jruby.runtime.Arity;
 
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.ObjectAllocator;
+import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 
 import org.jruby.util.ByteList;
 
+@JRubyClass(name="Iconv")
 public class RubyIconv extends RubyObject {
     //static private final String TRANSLIT = "//translit";
     static private final String IGNORE = "//ignore";
@@ -59,37 +64,39 @@ public class RubyIconv extends RubyObject {
         super(runtime, type);
     }
     
-    private static ObjectAllocator ICONV_ALLOCATOR = new ObjectAllocator() {
+    private static final ObjectAllocator ICONV_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
             return new RubyIconv(runtime, klass);
         }
     };
 
+    @JRubyModule(name="Iconv::Failure")
+    public static class Failure {}
+    @JRubyClass(name="Iconv::IllegalSequence", parent="ArgumentError", include="Iconv::Failure")
+    public static class IllegalSequence {}
+    @JRubyClass(name="Iconv::InvalidCharacter", parent="ArgumentError", include="Iconv::Failure")
+    public static class InvalidCharacter {}
+    @JRubyClass(name="Iconv::InvalidEncoding", parent="ArgumentError", include="Iconv::Failure")
+    public static class InvalidEncoding {}
+    @JRubyClass(name="Iconv::OutOfRange", parent="ArgumentError", include="Iconv::Failure")
+    public static class OutOfRange {}
+    @JRubyClass(name="Iconv::BrokenLibrary", parent="ArgumentError", include="Iconv::Failure")
+    public static class BrokenLibrary {}
+
     public static void createIconv(Ruby runtime) {
         RubyClass iconvClass = runtime.defineClass("Iconv", runtime.getObject(), ICONV_ALLOCATOR);
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyIconv.class);
-
-        iconvClass.getMetaClass().defineFastMethod("iconv", callbackFactory.getOptSingletonMethod("iconv"));
-        iconvClass.getMetaClass().defineFastMethod("conv", callbackFactory.getOptSingletonMethod("conv"));
-        iconvClass.getMetaClass().defineMethod("open", callbackFactory.getSingletonMethod("open", RubyKernel.IRUBY_OBJECT, RubyKernel.IRUBY_OBJECT));
         
-        iconvClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
-        iconvClass.defineFastMethod("iconv", callbackFactory.getFastOptMethod("iconv"));
-        iconvClass.defineFastMethod("close", callbackFactory.getFastMethod("close"));
+        iconvClass.defineAnnotatedMethods(RubyIconv.class);
 
         RubyModule failure = iconvClass.defineModuleUnder("Failure");
-        CallbackFactory failureCallbackFactory = runtime.callbackFactory(RubyFailure.class);
-        RubyClass argumentError = runtime.getClass("ArgumentError");
+        RubyClass argumentError = runtime.getArgumentError();
 
         String[] iconvErrors = {"IllegalSequence", "InvalidCharacter", "InvalidEncoding", 
                 "OutOfRange", "BrokenLibrary"};
         
         for (int i = 0; i < iconvErrors.length; i++) {
             RubyClass subClass = iconvClass.defineClassUnder(iconvErrors[i], argumentError, RubyFailure.ICONV_FAILURE_ALLOCATOR);
-            subClass.defineMethod("initialize", failureCallbackFactory.getOptMethod("initialize"));
-            subClass.defineFastMethod("success", failureCallbackFactory.getFastMethod("success"));
-            subClass.defineFastMethod("failed", failureCallbackFactory.getFastMethod("failed"));
-            subClass.defineFastMethod("inspect", failureCallbackFactory.getFastMethod("inspect"));
+            subClass.defineAnnotatedMethods(RubyFailure.class);
             subClass.includeModule(failure);
         }    
     }
@@ -102,7 +109,7 @@ public class RubyIconv extends RubyObject {
             return new RubyFailure(runtime, excptnClass, msg);
         }
 
-        protected static ObjectAllocator ICONV_FAILURE_ALLOCATOR = new ObjectAllocator() {
+        protected static final ObjectAllocator ICONV_FAILURE_ALLOCATOR = new ObjectAllocator() {
             public IRubyObject allocate(Ruby runtime, RubyClass klass) {
                 return new RubyFailure(runtime, klass);
             }
@@ -116,8 +123,8 @@ public class RubyIconv extends RubyObject {
             super(runtime, rubyClass, message);
         }
 
+        @JRubyMethod(name = "initialize", required = 1, optional = 2, frame = true)
         public IRubyObject initialize(IRubyObject[] args, Block block) {
-            Arity.checkArgumentCount(getRuntime(), args, 1, 3);
             super.initialize(args, block);
             success = args.length >= 2 ? args[1] : getRuntime().getNil();
             failed = args.length == 3 ? args[2] : getRuntime().getNil();
@@ -125,17 +132,20 @@ public class RubyIconv extends RubyObject {
             return this;
         }
 
+        @JRubyMethod(name = "success")
         public IRubyObject success() {
             return success;
         }
 
+        @JRubyMethod(name = "failed")
         public IRubyObject failed() {
             return failed;
         }
 
+        @JRubyMethod(name = "inspect")
         public IRubyObject inspect() {
             RubyModule rubyClass = getMetaClass();
-            StringBuffer buffer = new StringBuffer("#<");
+            StringBuilder buffer = new StringBuilder("#<");
             buffer.append(rubyClass.getName()).append(": ").append(success.inspect().toString());
             buffer.append(", ").append(failed.inspect().toString()).append(">");
 
@@ -158,16 +168,17 @@ public class RubyIconv extends RubyObject {
         return encoding.toLowerCase().indexOf(IGNORE) != -1 ? true : false;
     }
 
-    public static IRubyObject open(IRubyObject recv, IRubyObject to, IRubyObject from, Block block) {
-        Ruby runtime = recv.getRuntime();
-        RubyIconv iconv =
-            (RubyIconv) runtime.getClass("Iconv").newInstance(
-                    new IRubyObject[] { to, from }, Block.NULL_BLOCK);
+    @JRubyMethod(name = "open", required = 2, frame = true, meta = true)
+    public static IRubyObject open(ThreadContext context, IRubyObject recv, IRubyObject to, IRubyObject from, Block block) {
+        Ruby runtime = context.getRuntime();
+
+        RubyIconv iconv = newIconv(context, recv, to, from);
+
         if (!block.isGiven()) return iconv;
 
         IRubyObject result = runtime.getNil();
         try {
-            result = block.yield(recv.getRuntime().getCurrentContext(), iconv);
+            result = block.yield(context, iconv);
         } finally {
             iconv.close();
         }
@@ -175,18 +186,26 @@ public class RubyIconv extends RubyObject {
         return result;
     }
     
-    public IRubyObject initialize(IRubyObject[] args, Block unusedBlock) {
-        Arity.checkArgumentCount(getRuntime(), args, 2, 2);
+    private static RubyIconv newIconv(ThreadContext context, IRubyObject recv,
+            IRubyObject to, IRubyObject from) {
+        RubyClass klazz = (RubyClass)recv;
+
+        return (RubyIconv) klazz.newInstance(
+                context, new IRubyObject[] {to, from}, Block.NULL_BLOCK);
+    }
+
+    @JRubyMethod(name = "initialize", required = 2, frame = true)
+    public IRubyObject initialize(IRubyObject arg1, IRubyObject arg2, Block unusedBlock) {
         Ruby runtime = getRuntime();
-        if (!args[0].respondsTo("to_str")) {
-            throw runtime.newTypeError("can't convert " + args[0].getMetaClass() + " into String");
+        if (!arg1.respondsTo("to_str")) {
+            throw runtime.newTypeError("can't convert " + arg1.getMetaClass() + " into String");
         }
-        if (!args[1].respondsTo("to_str")) {
-            throw runtime.newTypeError("can't convert " + args[1].getMetaClass() + " into String");
+        if (!arg2.respondsTo("to_str")) {
+            throw runtime.newTypeError("can't convert " + arg2.getMetaClass() + " into String");
         }
 
-        String to = args[0].convertToString().toString();
-        String from = args[1].convertToString().toString();
+        String to = arg1.convertToString().toString();
+        String from = arg2.convertToString().toString();
 
         try {
 
@@ -196,9 +215,9 @@ public class RubyIconv extends RubyObject {
             if (!isIgnore(from)) fromEncoding.onUnmappableCharacter(CodingErrorAction.REPORT);
             if (!isIgnore(to)) toEncoding.onUnmappableCharacter(CodingErrorAction.REPORT);
         } catch (IllegalCharsetNameException e) {
-            throw runtime.newArgumentError("invalid encoding");
+            throw runtime.newInvalidEncoding("invalid encoding");
         } catch (UnsupportedCharsetException e) {
-            throw runtime.newArgumentError("invalid encoding");
+            throw runtime.newInvalidEncoding("invalid encoding");
         } catch (Exception e) {
             throw runtime.newSystemCallError(e.toString());
         }
@@ -206,41 +225,89 @@ public class RubyIconv extends RubyObject {
         return this;
     }
 
+    @JRubyMethod(name = "close")
     public IRubyObject close() {
         toEncoding = null;
         fromEncoding = null;
-        return getRuntime().newString("");
+        return RubyString.newEmptyString(getRuntime());
     }
 
-    public IRubyObject iconv(IRubyObject[] args) {
-        Ruby runtime = getRuntime();
-        args = Arity.scanArgs(runtime, args, 1, 2);
-        int start = 0;
-        int length = -1;
+    @JRubyMethod
+    public IRubyObject iconv(IRubyObject str) {
+        return iconv(str, 0, -1);
+    }
 
-        if (args[0].isNil()) {
+    @JRubyMethod
+    public IRubyObject iconv(IRubyObject str, IRubyObject startArg) {
+        int start = 0;
+        if (!startArg.isNil()) start = RubyNumeric.fix2int(startArg);
+        return iconv(str, start, -1);
+    }
+
+    @JRubyMethod
+    public IRubyObject iconv(IRubyObject str, IRubyObject startArg, IRubyObject endArg) {
+        int start = 0;
+        int end = -1;
+
+        if (!startArg.isNil()) start = RubyNumeric.fix2int(startArg);
+        if (!endArg.isNil()) end = RubyNumeric.fix2int(endArg);
+
+        return iconv(str, start, end);
+    }
+    
+    private IRubyObject iconv(IRubyObject str, int start, int end) {
+        if (str.isNil()) {
             fromEncoding.reset();
             toEncoding.reset();
-            return runtime.newString("");
+            return RubyString.newEmptyString(getRuntime());
         }
-        if (!args[0].respondsTo("to_str")) {
-            throw runtime.newTypeError("can't convert " + args[0].getMetaClass() + " into String");
+
+        return _iconv(str.convertToString(), start, end);
+    }
+
+    /**
+     * Variable-arity version for compatibility. Not bound to Ruby.
+     * @deprecated Use the versions with one, two or three arguments.
+     */
+    public IRubyObject iconv(IRubyObject[] args) {
+        switch (args.length) {
+        case 1:
+            return iconv(args[0]);
+        case 2:
+            return iconv(args[0], args[1]);
+        case 3:
+            return iconv(args[0], args[1], args[2]);
+        default:
+            Arity.raiseArgumentError(getRuntime(), args.length, 1, 2);
+            return null; // not reached
         }
-        if (!args[1].isNil()) start = RubyNumeric.fix2int(args[1]);
-        if (!args[2].isNil()) length = RubyNumeric.fix2int(args[2]);
-        
-        IRubyObject result = _iconv(args[0].convertToString(), start, length);
-        return result;
     }
 
     // FIXME: We are assuming that original string will be raw bytes.  If -Ku is provided
     // this will not be true, but that is ok for now.  Deal with that when someone needs it.
-    private IRubyObject _iconv(RubyString str, int start, int length) {
+    private IRubyObject _iconv(RubyString str, int start, int end) {
+        if (fromEncoding == null) {
+            throw getRuntime().newArgumentError("closed iconv");
+        }
+        
         ByteList bytes = str.getByteList();
         
-        if (length < 0) length = bytes.length() - start;
+        // treat start and end as start...end for end >= 0, start..end for end < 0
+        if (start < 0) {
+            start += bytes.length();
+        }
         
-        ByteBuffer buf = ByteBuffer.wrap(bytes.unsafeBytes(), start, length);
+        if (end < 0) {
+            end += 1 + bytes.length();
+        } else if (end > bytes.length()) {
+            end = bytes.length();
+        }
+        
+        if (start < 0 || end < start) { // invalid ranges result in an empty string
+            return RubyString.newEmptyString(getRuntime());
+        }
+        
+        ByteBuffer buf = ByteBuffer.wrap(bytes.unsafeBytes(), bytes.begin() + start, end - start);
         
         try {
             CharBuffer cbuf = fromEncoding.decode(buf);
@@ -256,23 +323,44 @@ public class RubyIconv extends RubyObject {
         return getRuntime().newString(new ByteList(arr, 0, buf.limit()));
     }
 
-    public static IRubyObject iconv(IRubyObject recv, IRubyObject[] args, Block unusedBlock) {
-        return convertWithArgs(recv, args, "iconv");
+    @JRubyMethod(name = "iconv", required = 2, rest = true, meta = true)
+    public static IRubyObject iconv(ThreadContext context, IRubyObject recv, IRubyObject[] args, Block unusedBlock) {
+        return convertWithArgs(context, recv, args, "iconv");
     }
     
-    public static IRubyObject conv(IRubyObject recv, IRubyObject[] args, Block unusedBlock) {
-        return convertWithArgs(recv, args, "conv").join(recv.getRuntime().newString(""));
+    @JRubyMethod(name = "conv", required = 3, rest = true, meta = true)
+    public static IRubyObject conv(ThreadContext context, IRubyObject recv, IRubyObject[] args, Block unusedBlock) {
+        return convertWithArgs(context, recv, args, "conv").join(context, RubyString.newEmptyString(recv.getRuntime()));
     }
-    
-    public static RubyArray convertWithArgs(IRubyObject recv, IRubyObject[] args, String function) {
-        Arity.checkArgumentCount(recv.getRuntime(), args, 3, -1);
 
-        String fromEncoding = args[1].convertToString().toString();
-        String toEncoding = args[0].convertToString().toString();
-        RubyArray array = recv.getRuntime().newArray();
-        
-        for (int i = 2; i < args.length; i++) {
-            array.append(convert2(fromEncoding, toEncoding, args[i].convertToString()));
+    @JRubyMethod(name = "charset_map", meta= true)
+    public static IRubyObject charset_map_get(IRubyObject recv) {
+        return recv.getRuntime().getCharsetMap();
+    }
+
+    private static String mapCharset(ThreadContext context, IRubyObject val) {
+        RubyHash charset = val.getRuntime().getCharsetMap();
+        if (charset.size() > 0) {
+            RubyString key = val.callMethod(context, "downcase").convertToString();
+            IRubyObject tryVal = charset.fastARef(key);
+            if (tryVal != null) val = tryVal;
+        }
+
+        return val.convertToString().toString();
+    }
+
+    public static RubyArray convertWithArgs(ThreadContext context, IRubyObject recv, IRubyObject[] args, String function) {
+        assert args.length >= 2;
+
+        RubyArray array = context.getRuntime().newArray(args.length - 2);
+        RubyIconv iconv = newIconv(context, recv, args[0], args[1]);
+
+        try {
+            for (int i = 2; i < args.length; i++) {
+                array.append(iconv.iconv(args[i]));
+            }
+        } finally {
+            iconv.close();
         }
 
         return array;
@@ -289,35 +377,4 @@ public class RubyIconv extends RubyObject {
         return RubyString.newString(original.getRuntime(), string);
     }
     */
-
-    // FIXME: We are assuming that original string will be raw bytes.  If -Ku is provided
-    // this will not be true, but that is ok for now.  Deal with that when someone needs it.
-    private static IRubyObject convert2(String fromEncoding, String toEncoding, RubyString original) {
-        // Don't bother to convert if it is already in toEncoding
-        if (fromEncoding.equals(toEncoding)) return original;
-        
-        try {
-            // Get all bytes from string and pretend they are not encoded in any way.
-            ByteList bytes = original.getByteList();
-            ByteBuffer buf = ByteBuffer.wrap(bytes.unsafeBytes(), bytes.begin(), bytes.length());
-
-            CharsetDecoder decoder = Charset.forName(getCharset(fromEncoding)).newDecoder();
-            
-            if (!isIgnore(fromEncoding)) decoder.onUnmappableCharacter(CodingErrorAction.REPORT);
-
-            CharBuffer cbuf = decoder.decode(buf);
-            CharsetEncoder encoder = Charset.forName(getCharset(toEncoding)).newEncoder();
-            
-            if (!isIgnore(toEncoding)) encoder.onUnmappableCharacter(CodingErrorAction.REPORT);
-
-            buf = encoder.encode(cbuf);
-            byte[] arr = buf.array();
-            return RubyString.newString(original.getRuntime(), new ByteList(arr,0,buf.limit()));
-        } catch (UnsupportedCharsetException e) {
-            throw original.getRuntime().newInvalidEncoding("invalid encoding");
-        } catch (UnmappableCharacterException e) {
-        } catch (CharacterCodingException e) {
-        }
-        return original.getRuntime().getNil();
-    }
 }
