@@ -34,18 +34,27 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby;
 
-import java.math.BigInteger;
-import java.util.List;
+import static org.jruby.RubyEnumerator.enumeratorize;
+import static org.jruby.util.Numeric.f_abs;
+import static org.jruby.util.Numeric.f_arg;
+import static org.jruby.util.Numeric.f_mul;
+import static org.jruby.util.Numeric.f_negative_p;
 
+import java.math.BigInteger;
+
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
 import org.jruby.exceptions.RaiseException;
+import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
+import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
 import org.jruby.util.Convert;
+import org.jruby.util.Numeric;
 
 /**
  * Base class for all numerical types in ruby.
@@ -53,54 +62,54 @@ import org.jruby.util.Convert;
 // TODO: Numeric.new works in Ruby and it does here too.  However trying to use
 //   that instance in a numeric operation should generate an ArgumentError. Doing
 //   this seems so pathological I do not see the need to fix this now.
+@JRubyClass(name="Numeric", include="Comparable")
 public class RubyNumeric extends RubyObject {
     
     public static RubyClass createNumericClass(Ruby runtime) {
         RubyClass numeric = runtime.defineClass("Numeric", runtime.getObject(), NUMERIC_ALLOCATOR);
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyNumeric.class);
-        numeric.defineFastMethod("singleton_method_added", callbackFactory.getFastMethod("sadded",
-                RubyKernel.IRUBY_OBJECT));
+        runtime.setNumeric(numeric);
 
-        numeric.includeModule(runtime.getModule("Comparable"));
+        numeric.kindOf = new RubyModule.KindOf() {
+            @Override
+            public boolean isKindOf(IRubyObject obj, RubyModule type) {
+                return obj instanceof RubyNumeric;
+            }
+        };
 
-        numeric.defineFastMethod("initialize_copy", callbackFactory.getFastMethod("init_copy", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("coerce", callbackFactory.getFastMethod("coerce", RubyKernel.IRUBY_OBJECT));
-
-        numeric.defineFastMethod("+@", callbackFactory.getFastMethod("uplus"));
-        numeric.defineFastMethod("-@", callbackFactory.getFastMethod("uminus"));
-        numeric.defineFastMethod("<=>", callbackFactory.getFastMethod("cmp", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("quo", callbackFactory.getFastMethod("quo", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("eql?", callbackFactory.getFastMethod("eql_p", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("div", callbackFactory.getFastMethod("div", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("divmod", callbackFactory.getFastMethod("divmod", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("modulo", callbackFactory.getFastMethod("modulo", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("remainder", callbackFactory.getFastMethod("remainder", RubyKernel.IRUBY_OBJECT));
-        numeric.defineFastMethod("abs", callbackFactory.getFastMethod("abs"));
-        numeric.defineFastMethod("to_int", callbackFactory.getFastMethod("to_int"));
-        numeric.defineFastMethod("integer?", callbackFactory.getFastMethod("int_p"));
-        numeric.defineFastMethod("zero?", callbackFactory.getFastMethod("zero_p"));
-        numeric.defineFastMethod("nonzero?", callbackFactory.getFastMethod("nonzero_p"));
-        numeric.defineFastMethod("floor", callbackFactory.getFastMethod("floor"));
-        numeric.defineFastMethod("ceil", callbackFactory.getFastMethod("ceil"));
-        numeric.defineFastMethod("round", callbackFactory.getFastMethod("round"));
-        numeric.defineFastMethod("truncate", callbackFactory.getFastMethod("truncate"));
-        numeric.defineMethod("step", callbackFactory.getOptMethod("step"));
+        numeric.includeModule(runtime.getComparable());
+        numeric.defineAnnotatedMethods(RubyNumeric.class);
 
         return numeric;
     }
 
-    protected static ObjectAllocator NUMERIC_ALLOCATOR = new ObjectAllocator() {
+    protected static final ObjectAllocator NUMERIC_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klass) {
             return new RubyNumeric(runtime, klass);
         }
     };
 
     public static double DBL_EPSILON=2.2204460492503131e-16;
+
+    private static IRubyObject convertToNum(double val, Ruby runtime) {
+
+        if (val >= (double) RubyFixnum.MAX || val < (double) RubyFixnum.MIN) {
+            return RubyBignum.newBignum(runtime, val);
+        }
+        return RubyFixnum.newFixnum(runtime, (long) val);
+    }
     
     public RubyNumeric(Ruby runtime, RubyClass metaClass) {
         super(runtime, metaClass);
     }
 
+    public RubyNumeric(Ruby runtime, RubyClass metaClass, boolean useObjectSpace) {
+        super(runtime, metaClass, useObjectSpace);
+    }    
+
+    public RubyNumeric(Ruby runtime, RubyClass metaClass, boolean useObjectSpace, boolean canBeTainted) {
+        super(runtime, metaClass, useObjectSpace, canBeTainted);
+    }    
+    
     // The implementations of these are all bonus (see TODO above)  I was going
     // to throw an error from these, but it appears to be the wrong place to
     // do it.
@@ -113,7 +122,7 @@ public class RubyNumeric extends RubyObject {
     }
     
     public static RubyNumeric newNumeric(Ruby runtime) {
-    	return new RubyNumeric(runtime, runtime.getClass("Numeric"));
+    	return new RubyNumeric(runtime, runtime.getNumeric());
     }
 
     /*  ================
@@ -129,49 +138,77 @@ public class RubyNumeric extends RubyObject {
 
         checkInt(arg, num);
         return (int)num;
-        }
+    }
     
     /** check_int
      * 
      */
     public static void checkInt(IRubyObject arg, long num){
-        String s;
         if (num < Integer.MIN_VALUE) {
-            s = "small";
+            tooSmall(arg, num);
         } else if (num > Integer.MAX_VALUE) {
-            s = "big";
+            tooBig(arg, num);
         } else {
             return;
         }
-        throw arg.getRuntime().newRangeError("integer " + num + " too " + s + " to convert to `int'");
+    }
+    
+    private static void tooSmall(IRubyObject arg, long num) {
+        throw arg.getRuntime().newRangeError("integer " + num + " too small to convert to `int'");
+    }
+    
+    private static void tooBig(IRubyObject arg, long num) {
+        throw arg.getRuntime().newRangeError("integer " + num + " too big to convert to `int'");
+    }
+
+    /**
+     * NUM2CHR
+     */
+    public static byte num2chr(IRubyObject arg) {
+        if (arg instanceof RubyString) {
+            String value = ((RubyString) arg).toString();
+
+            if (value != null && value.length() > 0) return (byte) value.charAt(0);
+        } 
+
+        return (byte) num2int(arg);
     }
 
-    // TODO: Find all consumers and convert to correct conversion protocol <- done
     /** rb_num2long and FIX2LONG (numeric.c)
      * 
      */
     public static long num2long(IRubyObject arg) {
         if (arg instanceof RubyFixnum) {
             return ((RubyFixnum) arg).getLongValue();
+        } else {
+            return other2long(arg);
         }
-        if (arg.isNil()) {
-            throw arg.getRuntime().newTypeError("no implicit conversion from nil to integer");
     }
 
-        if (arg instanceof RubyFloat) {
-            double aFloat = ((RubyFloat) arg).getDoubleValue();
-            if (aFloat <= (double) Long.MAX_VALUE && aFloat >= (double) Long.MIN_VALUE) {
-                return (long) aFloat;
-            } else {
-                // TODO: number formatting here, MRI uses "%-.10g", 1.4 API is a must?
-                throw arg.getRuntime().newTypeError("float " + aFloat + "out of range of integer");
-            }
+    private static long other2long(IRubyObject arg) throws RaiseException {
+        if (arg.isNil()) {
+            throw arg.getRuntime().newTypeError("no implicit conversion from nil to integer");
+        } else if (arg instanceof RubyFloat) {
+            return float2long((RubyFloat)arg);
         } else if (arg instanceof RubyBignum) {
             return RubyBignum.big2long((RubyBignum) arg);
         }
         return arg.convertToInteger().getLongValue();
     }
+    
+    private static long float2long(RubyFloat flt) {
+        double aFloat = flt.getDoubleValue();
+        if (aFloat <= (double) Long.MAX_VALUE && aFloat >= (double) Long.MIN_VALUE) {
+            return (long) aFloat;
+        } else {
+            // TODO: number formatting here, MRI uses "%-.10g", 1.4 API is a must?
+            throw flt.getRuntime().newRangeError("float " + aFloat + " out of range of integer");
+        }
+    }
 
+    /** rb_dbl2big + LONG2FIX at once (numeric.c)
+     * 
+     */
     /** rb_dbl2big + LONG2FIX at once (numeric.c)
      * 
      */
@@ -182,11 +219,7 @@ public class RubyNumeric extends RubyObject {
         if (Double.isNaN(val)) {
             throw runtime.newFloatDomainError("NaN");
         }
-
-        if (val > (double) RubyFixnum.MAX || val < (double) RubyFixnum.MIN) {
-            return RubyBignum.newBignum(runtime, val);
-        }
-        return RubyFixnum.newFixnum(runtime, (long) val);
+        return convertToNum(val,runtime);
     }
 
     /** rb_num2dbl and NUM2DBL
@@ -200,7 +233,7 @@ public class RubyNumeric extends RubyObject {
         } else if (arg == arg.getRuntime().getNil()) {
             throw arg.getRuntime().newTypeError("no implicit conversion to float from nil");
         }
-        return arg.convertToFloat().getDoubleValue();
+        return RubyKernel.new_float(arg, arg).getDoubleValue();
     }
 
     /** rb_dbl_cmp (numeric.c)
@@ -220,15 +253,15 @@ public class RubyNumeric extends RubyObject {
     }
 
     public static long fix2long(IRubyObject arg) {
-            return ((RubyFixnum) arg).getLongValue();
-        }
+        return ((RubyFixnum) arg).getLongValue();
+    }
 
     public static int fix2int(IRubyObject arg) {
         long num = arg instanceof RubyFixnum ? fix2long(arg) : num2long(arg);
 
         checkInt(arg, num);
         return (int) num;
-        }
+    }
 
     public static RubyInteger str2inum(Ruby runtime, RubyString str, int base) {
         return str2inum(runtime,str,base,false);
@@ -285,28 +318,30 @@ public class RubyNumeric extends RubyObject {
     public static RubyInteger str2inum(Ruby runtime, RubyString str, int base, boolean strict) {
         if (base != 0 && (base < 2 || base > 36)) {
             throw runtime.newArgumentError("illegal radix " + base);
-            }
+        }
         ByteList bytes = str.getByteList();
         try {
-            return runtime.newFixnum(Convert.byteListToLong(bytes,base,strict));
-
+            return runtime.newFixnum(Convert.byteListToLong(bytes, base, strict));
         } catch (InvalidIntegerException e) {
-            if (strict) {
-                throw runtime.newArgumentError("invalid value for Integer: "
-                        + str.callMethod(runtime.getCurrentContext(), "inspect").toString());
-            }
-            return RubyFixnum.zero(runtime);
+            return str2inumIIE(strict, runtime, str);
         } catch (NumberTooLargeException e) {
+            return str2inumNTLE(strict, runtime, str, bytes, base);
+        }
+    }
+
+    private static RubyInteger str2inumIIE(boolean strict, Ruby runtime, RubyString str) throws RaiseException {
+        if (strict) {
+            throw runtime.newArgumentError("invalid value for Integer: " + str.callMethod(runtime.getCurrentContext(), "inspect").toString());
+        }
+        return RubyFixnum.zero(runtime);
+    }
+    
+    private static RubyInteger str2inumNTLE(boolean strict, Ruby runtime, RubyString str, ByteList bytes, int base) {
         try {
-                BigInteger bi = Convert.byteListToBigInteger(bytes,base,strict);
-                return new RubyBignum(runtime,bi);
-            } catch (InvalidIntegerException e2) {
-                if(strict) {
-                    throw runtime.newArgumentError("invalid value for Integer: "
-                            + str.callMethod(runtime.getCurrentContext(), "inspect").toString());
-                }
-                return RubyFixnum.zero(runtime);
-            }
+            BigInteger bi = Convert.byteListToBigInteger(bytes, base, strict);
+            return new RubyBignum(runtime, bi);
+        } catch (InvalidIntegerException e2) {
+            return str2inumIIE(strict, runtime, str);
         }
     }
 
@@ -350,11 +385,11 @@ public class RubyNumeric extends RubyObject {
      *
      */
     
-    protected IRubyObject[] getCoerced(IRubyObject other, boolean error) {
+    protected IRubyObject[] getCoerced(ThreadContext context, IRubyObject other, boolean error) {
         IRubyObject result;
         
         try {
-            result = other.callMethod(getRuntime().getCurrentContext(), "coerce", this);
+            result = other.callMethod(context, "coerce", this);
         } catch (RaiseException e) {
             if (error) {
                 throw getRuntime().newTypeError(
@@ -371,20 +406,20 @@ public class RubyNumeric extends RubyObject {
         return ((RubyArray)result).toJavaArray();
     }
 
-    protected IRubyObject callCoerced(String method, IRubyObject other, boolean err) {
-        IRubyObject[] args = getCoerced(other, err);
+    protected IRubyObject callCoerced(ThreadContext context, String method, IRubyObject other, boolean err) {
+        IRubyObject[] args = getCoerced(context, other, err);
         if(args == null) {
             return getRuntime().getNil();
         }
-        return args[0].callMethod(getRuntime().getCurrentContext(), method, args[1]);
+        return args[0].callMethod(context, method, args[1]);
     }
 
-    protected IRubyObject callCoerced(String method, IRubyObject other) {
-        IRubyObject[] args = getCoerced(other, false);
+    protected IRubyObject callCoerced(ThreadContext context, String method, IRubyObject other) {
+        IRubyObject[] args = getCoerced(context, other, false);
         if(args == null) {
             return getRuntime().getNil();
         }
-        return args[0].callMethod(getRuntime().getCurrentContext(), method, args[1]);
+        return args[0].callMethod(context, method, args[1]);
     }
     
     // beneath are rewritten coercions that reflect MRI logic, the aboves are used only by RubyBigDecimal
@@ -392,65 +427,69 @@ public class RubyNumeric extends RubyObject {
     /** coerce_body
      *
      */
-    protected final IRubyObject coerceBody(IRubyObject other) {
-        return other.callMethod(getRuntime().getCurrentContext(), "coerce", this);
-        } 
+    protected final IRubyObject coerceBody(ThreadContext context, IRubyObject other) {
+        return other.callMethod(context, "coerce", this);
+    }
 
     /** do_coerce
      * 
      */
-    protected final RubyArray doCoerce(IRubyObject other, boolean err) {
+    protected final RubyArray doCoerce(ThreadContext context, IRubyObject other, boolean err) {
         IRubyObject result;
         try {
-            result = coerceBody(other);
+            result = coerceBody(context, other);
         } catch (RaiseException e) {
             if (err) {
                 throw getRuntime().newTypeError(
                         other.getMetaClass().getName() + " can't be coerced into " + getMetaClass().getName());
-    }
+            }
             return null;
         }
     
         if (!(result instanceof RubyArray) || ((RubyArray) result).getLength() != 2) {
             throw getRuntime().newTypeError("coerce must return [x, y]");
-    }
+        }
         return (RubyArray) result;
     }
 
     /** rb_num_coerce_bin
      *  coercion taking two arguments
      */
-    protected final IRubyObject coerceBin(String method, IRubyObject other) {
-        RubyArray ary = doCoerce(other, true);
-        return (ary.eltInternal(0)).callMethod(getRuntime().getCurrentContext(), method, ary.eltInternal(1));
+    protected final IRubyObject coerceBin(ThreadContext context, String method, IRubyObject other) {
+        RubyArray ary = doCoerce(context, other, true);
+        return (ary.eltInternal(0)).callMethod(context, method, ary.eltInternal(1));
     }
     
     /** rb_num_coerce_cmp
      *  coercion used for comparisons
      */
-    protected final IRubyObject coerceCmp(String method, IRubyObject other) {
-        RubyArray ary = doCoerce(other, false);
+    protected final IRubyObject coerceCmp(ThreadContext context, String method, IRubyObject other) {
+        RubyArray ary = doCoerce(context, other, false);
         if (ary == null) {
             return getRuntime().getNil(); // MRI does it!
         } 
-        return (ary.eltInternal(0)).callMethod(getRuntime().getCurrentContext(), method, ary.eltInternal(1));
+        return (ary.eltInternal(0)).callMethod(context, method, ary.eltInternal(1));
     }
         
     /** rb_num_coerce_relop
      *  coercion used for relative operators
      */
-    protected final IRubyObject coerceRelOp(String method, IRubyObject other) {
-        RubyArray ary = doCoerce(other, false);
-        if (ary != null) {
-            IRubyObject result = (ary.eltInternal(0)).callMethod(getRuntime().getCurrentContext(), method,ary.eltInternal(1));
-            if (!result.isNil()) {
-                return result;
-    }
+    protected final IRubyObject coerceRelOp(ThreadContext context, String method, IRubyObject other) {
+        RubyArray ary = doCoerce(context, other, false);
+        if (ary == null) {
+            return RubyComparable.cmperr(this, other);
         }
+
+        return unwrapCoerced(context, method, other, ary);
+    }
     
-        RubyComparable.cmperr(this, other); // MRI also does it that way       
-        return null; // not reachd as in MRI
-        } 
+    private final IRubyObject unwrapCoerced(ThreadContext context, String method, IRubyObject other, RubyArray ary) {
+        IRubyObject result = (ary.eltInternal(0)).callMethod(context, method, ary.eltInternal(1));
+        if (result.isNil()) {
+            return RubyComparable.cmperr(this, other);
+        }
+        return result;
+    }
         
     public RubyNumeric asNumeric() {
         return this;
@@ -464,48 +503,55 @@ public class RubyNumeric extends RubyObject {
     /** num_sadded
      *
      */
+    @JRubyMethod(name = "singleton_method_added")
     public IRubyObject sadded(IRubyObject name) {
         throw getRuntime().newTypeError("can't define singleton method " + name + " for " + getType().getName());
-        } 
+    } 
         
     /** num_init_copy
      *
      */
-    public IRubyObject init_copy(IRubyObject arg) {
+    @Override
+    @JRubyMethod(name = "initialize_copy", visibility = Visibility.PRIVATE)
+    public IRubyObject initialize_copy(IRubyObject arg) {
         throw getRuntime().newTypeError("can't copy " + getType().getName());
     }
     
     /** num_coerce
      *
      */
+    @JRubyMethod(name = "coerce")
     public IRubyObject coerce(IRubyObject other) {
-        if (getMetaClass() == other.getMetaClass()) {
-            return getRuntime().newArray(other, this);
-        } 
-        
-        return getRuntime().newArray(other.convertToFloat(), convertToFloat());
+        if (getClass() == other.getClass()) return getRuntime().newArray(other, this);
+
+        IRubyObject cdr = RubyKernel.new_float(this, this);
+        IRubyObject car = RubyKernel.new_float(this, other);
+
+        return getRuntime().newArray(car, cdr);
     }
 
     /** num_uplus
      *
      */
-    public IRubyObject uplus() {
+    @JRubyMethod(name = "+@")
+    public IRubyObject op_uplus() {
         return this;
     }
 
     /** num_uminus
      *
      */
-    public IRubyObject uminus() {
-        RubyFixnum zero = RubyFixnum.zero(getRuntime());
-        List list = zero.doCoerce(this, true);
-        return ((RubyObject) list.get(0)).callMethod(getRuntime().getCurrentContext(), MethodIndex.OP_MINUS, "-", (RubyObject) list.get(1));
+    @JRubyMethod(name = "-@")
+    public IRubyObject op_uminus(ThreadContext context) {
+        RubyArray ary = RubyFixnum.zero(context.getRuntime()).doCoerce(context, this, true);
+        return ary.eltInternal(0).callMethod(context, "-", ary.eltInternal(1));
     }
     
     /** num_cmp
      *
      */
-    public IRubyObject cmp(IRubyObject other) {
+    @JRubyMethod(name = "<=>")
+    public IRubyObject op_cmp(IRubyObject other) {
         if (this == other) { // won't hurt fixnums
             return RubyFixnum.zero(getRuntime());
         }
@@ -515,102 +561,135 @@ public class RubyNumeric extends RubyObject {
     /** num_eql
      *
      */
-    public IRubyObject eql_p(IRubyObject other) {
-        if (getMetaClass() != other.getMetaClass()) {
-            return getRuntime().getFalse();
-                }
-        return super.equal(other);
-            }
-            
+    @JRubyMethod(name = "eql?")
+    public IRubyObject eql_p(ThreadContext context, IRubyObject other) {
+        if (getClass() != other.getClass()) return getRuntime().getFalse();
+        return equalInternal(context, this, other) ? getRuntime().getTrue() : getRuntime().getFalse();
+    }
+
     /** num_quo
      *
      */
-    public IRubyObject quo(IRubyObject other) {
-        return callMethod(getRuntime().getCurrentContext(), "/", other);
-    	}
+    @JRubyMethod(name = "quo")
+    public IRubyObject quo(ThreadContext context, IRubyObject other) {
+        return callMethod(context, "/", other);
+    }
+    
+    /** num_quo
+    *
+    */
+    @JRubyMethod(name = "quo", compat = CompatVersion.RUBY1_9)
+    public IRubyObject quo_19(ThreadContext context, IRubyObject other) {
+        return RubyRational.newRationalRaw(context.getRuntime(), this).callMethod(context, "/", other);
+    }
 
     /** num_div
      * 
      */
-    public IRubyObject div(IRubyObject other) {
-        return callMethod(getRuntime().getCurrentContext(), "/", other).convertToFloat().floor();
+    @JRubyMethod(name = "div")
+    public IRubyObject div(ThreadContext context, IRubyObject other) {
+        return callMethod(context, "/", other).convertToFloat().floor();
     }
 
     /** num_divmod
      * 
      */
-    public IRubyObject divmod(IRubyObject other) {
-        IRubyObject cdr = callMethod(getRuntime().getCurrentContext(), "%", other);
-        IRubyObject car = div(other);
-        return RubyArray.newArray(getRuntime(), car, cdr);
-
+    @JRubyMethod(name = "divmod")
+    public IRubyObject divmod(ThreadContext context, IRubyObject other) {
+        return RubyArray.newArray(getRuntime(), div(context, other), modulo(context, other));
+    }
+    
+    /** num_fdiv (1.9) */
+    @JRubyMethod(name = "fdiv", compat = CompatVersion.RUBY1_9)
+    public IRubyObject fdiv(ThreadContext context, IRubyObject other) {
+        return RuntimeHelpers.invoke(context, this.convertToFloat(), "/", other);
     }
 
     /** num_modulo
      *
      */
-    public IRubyObject modulo(IRubyObject other) {
-        return callMethod(getRuntime().getCurrentContext(), "%", other);
-    	}
-    	
+    @JRubyMethod(name = "modulo")
+    public IRubyObject modulo(ThreadContext context, IRubyObject other) {
+        return callMethod(context, "%", other);
+    }
+
     /** num_remainder
      *
      */
-    public IRubyObject remainder(IRubyObject y) {
-            ThreadContext context = getRuntime().getCurrentContext();
-        IRubyObject z = callMethod(context, "%", y);
+    @JRubyMethod(name = "remainder")
+    public IRubyObject remainder(ThreadContext context, IRubyObject dividend) {
+        IRubyObject z = callMethod(context, "%", dividend);
         IRubyObject x = this;
         RubyFixnum zero = RubyFixnum.zero(getRuntime());
 
-        if (!(((RubyNumeric)z).equal(zero).isTrue())
-            && ((x.callMethod(context, MethodIndex.OP_LT, "<", zero)).isTrue()
-                && (y.callMethod(context, MethodIndex.OP_GT, ">", zero)).isTrue()) 
-            || ((x.callMethod(context, MethodIndex.OP_GT, ">", zero)).isTrue()
-                && (y.callMethod(context, MethodIndex.OP_LT, "<", zero)).isTrue())) {
-            return z.callMethod(context, MethodIndex.OP_MINUS, "-",y);
+        if (!equalInternal(context, z, zero) &&
+                ((x.callMethod(context, MethodIndex.OP_LT, "<", zero).isTrue() &&
+                dividend.callMethod(context, MethodIndex.OP_GT, ">", zero).isTrue()) ||
+                (x.callMethod(context, MethodIndex.OP_GT, ">", zero).isTrue() &&
+                dividend.callMethod(context, MethodIndex.OP_LT, "<", zero).isTrue()))) {
+            return z.callMethod(context, MethodIndex.OP_MINUS, "-", dividend);
+        } else {
+            return z;
         }
-
-        return z;
-            }
+    }
 
     /** num_abs
      *
      */
-    public IRubyObject abs() {
-        ThreadContext context = getRuntime().getCurrentContext();
+    @JRubyMethod(name = "abs")
+    public IRubyObject abs(ThreadContext context) {
         if (callMethod(context, MethodIndex.OP_LT, "<", RubyFixnum.zero(getRuntime())).isTrue()) {
-            return (RubyNumeric) callMethod(context, "-@");
-    	}
+            return callMethod(context, "-@");
+        }
         return this;
     }
-    	
+
+    /** num_abs/1.9
+     * 
+     */
+    @JRubyMethod(name = "magnitude", compat = CompatVersion.RUBY1_9)
+    public IRubyObject magnitude(ThreadContext context) {
+        return abs(context);
+    }
+
     /** num_to_int
      * 
      */
-    public IRubyObject to_int() {
-        return callMethod(getRuntime().getCurrentContext(), MethodIndex.TO_I, "to_i", IRubyObject.NULL_ARRAY);
+    @JRubyMethod(name = "to_int")
+    public IRubyObject to_int(ThreadContext context) {
+        return RuntimeHelpers.invoke(context, this, "to_i");
+    }
+
+    /** num_real_p
+    *
+    */
+    @JRubyMethod(name = "real?", compat = CompatVersion.RUBY1_9)
+    public IRubyObject scalar_p() {
+        return getRuntime().getTrue();
     }
 
     /** num_int_p
      *
      */
-    public IRubyObject int_p() {
+    @JRubyMethod(name = "integer?")
+    public IRubyObject integer_p() {
         return getRuntime().getFalse();
     }
     
     /** num_zero_p
      *
      */
-    public IRubyObject zero_p() {
-        // Will always return a boolean
-        return equal(RubyFixnum.zero(getRuntime())).isTrue() ? getRuntime().getTrue() : getRuntime().getFalse();
+    @JRubyMethod(name = "zero?")
+    public IRubyObject zero_p(ThreadContext context) {
+        return equalInternal(context, this, RubyFixnum.zero(getRuntime())) ? getRuntime().getTrue() : getRuntime().getFalse();
     }
     
     /** num_nonzero_p
      *
      */
-    public IRubyObject nonzero_p() {
-        if (callMethod(getRuntime().getCurrentContext(), "zero?").isTrue()) {
+    @JRubyMethod(name = "nonzero?")
+    public IRubyObject nonzero_p(ThreadContext context) {
+        if (callMethod(context, "zero?").isTrue()) {
             return getRuntime().getNil();
         }
         return this;
@@ -619,13 +698,15 @@ public class RubyNumeric extends RubyObject {
     /** num_floor
      *
      */
+    @JRubyMethod(name = "floor")
     public IRubyObject floor() {
         return convertToFloat().floor();
-        }
+    }
         
     /** num_ceil
      *
      */
+    @JRubyMethod(name = "ceil")
     public IRubyObject ceil() {
         return convertToFloat().ceil();
     }
@@ -633,32 +714,41 @@ public class RubyNumeric extends RubyObject {
     /** num_round
      *
      */
+    @JRubyMethod(name = "round")
     public IRubyObject round() {
         return convertToFloat().round();
-        }
+    }
 
     /** num_truncate
      *
      */
+    @JRubyMethod(name = "truncate")
     public IRubyObject truncate() {
         return convertToFloat().truncate();
     }
     
-    public IRubyObject step(IRubyObject[] args, Block block) {
-        IRubyObject to;
-        IRubyObject step;
-        
-        if(args.length == 1){ 
-            to = args[0];
-            step = RubyFixnum.one(getRuntime());
-        } else if (args.length == 2) {
-            to = args[0];
-            step = args[1];
-        }else{
-            throw getRuntime().newTypeError("wrong number of arguments");
+    @JRubyMethod(name = "step", required = 1, optional = 1, frame = true)
+    public IRubyObject step(ThreadContext context, IRubyObject[] args, Block block) {
+        switch (args.length) {
+        case 0: throw context.getRuntime().newArgumentError(0, 1);
+        case 1: return step(context, args[0], block);
+        case 2: return step(context, args[0], args[1], block);
+        default: throw context.getRuntime().newArgumentError(args.length, 2);
         }
-        
-        ThreadContext context = getRuntime().getCurrentContext();
+    }
+    
+    @JRubyMethod(name = "step", frame = true)
+    public IRubyObject step(ThreadContext context, IRubyObject arg0, Block block) {
+        return step(context, arg0, RubyFixnum.one(context.getRuntime()), block);
+    }
+
+    @JRubyMethod(name = "step", frame = true, compat = CompatVersion.RUBY1_9)
+    public IRubyObject step19(ThreadContext context, IRubyObject arg0, Block block) {
+        return block.isGiven() ? step(context, arg0, block) : enumeratorize(context.getRuntime(), this, "step", arg0);
+    }
+    
+    @JRubyMethod(name = "step", frame = true)
+    public IRubyObject step(ThreadContext context, IRubyObject to, IRubyObject step, Block block) {
         if (this instanceof RubyFixnum && to instanceof RubyFixnum && step instanceof RubyFixnum) {
             long value = getLongValue();
             long end = ((RubyFixnum) to).getLongValue();
@@ -707,7 +797,7 @@ public class RubyNumeric extends RubyObject {
             } else {
                 cmp = MethodIndex.OP_LT;
             }
-            cmpString = MethodIndex.NAMES[cmp];
+            cmpString = MethodIndex.NAMES.get(cmp);
 
             while (true) {
                 if (i.callMethod(context, cmp, cmpString, to).isTrue()) {
@@ -719,25 +809,101 @@ public class RubyNumeric extends RubyObject {
         }
         return this;
     }
-    
-    //    /** num_equal
-    //     *
-    //     */
-    //    public RubyBoolean veryEqual(IRubyObject other) {
-    //        IRubyObject truth = super.equal(other);
-    //
-    //        return truth == getRuntime().getNil() ? getRuntime().getFalse() : (RubyBoolean) truth;
-    //    }
-    //
-    /** num_equal
-     *
-     */
-    public IRubyObject equal(IRubyObject other) {
-        if (this == other) { // it won't hurt fixnums
-            return getRuntime().getTrue();
+
+    @JRubyMethod(name = "step", frame = true, compat = CompatVersion.RUBY1_9)
+    public IRubyObject step19(ThreadContext context, IRubyObject to, IRubyObject step, Block block) {
+        return block.isGiven() ? step(context, to, step, block) : enumeratorize(context.getRuntime(), this, "step", new IRubyObject[] {to, step});
     }
 
-        return other.callMethod(getRuntime().getCurrentContext(), MethodIndex.EQUALEQUAL, "==", this);
+    /** num_equal, doesn't override RubyObject.op_equal
+     *
+     */
+    protected final IRubyObject op_num_equal(ThreadContext context, IRubyObject other) {
+        // it won't hurt fixnums
+        if (this == other)  return getRuntime().getTrue();
+
+        return other.callMethod(context, MethodIndex.EQUALEQUAL, "==", this);
+    }
+
+    /** num_numerator
+     * 
+     */
+    @JRubyMethod(name = "numerator", compat = CompatVersion.RUBY1_9)
+    public IRubyObject numerator(ThreadContext context) {
+        return RubyRational.newRationalConvert(context, this).callMethod(context, "numerator");
+    }
+    
+    /** num_denominator
+     * 
+     */
+    @JRubyMethod(name = "denominator", compat = CompatVersion.RUBY1_9)
+    public IRubyObject denominator(ThreadContext context) {
+        return RubyRational.newRationalConvert(context, this).callMethod(context, "denominator");
+    }
+
+    /** numeric_to_c
+     * 
+     */
+    @JRubyMethod(name = "to_c", compat = CompatVersion.RUBY1_9)
+    public IRubyObject to_c(ThreadContext context) {
+        return RubyComplex.newComplexCanonicalize(context, this);
+    }
+
+    /** numeric_real
+     * 
+     */
+    @JRubyMethod(name = "real", compat = CompatVersion.RUBY1_9)
+    public IRubyObject real(ThreadContext context) {
+        return this;
+    }
+
+    /** numeric_image
+     * 
+     */
+    @JRubyMethod(name = {"imaginary", "imag"}, compat = CompatVersion.RUBY1_9)
+    public IRubyObject image(ThreadContext context) {
+        return RubyFixnum.zero(context.getRuntime());
+    }
+
+    /** numeric_abs2
+     * 
+     */
+    @JRubyMethod(name = "abs2", compat = CompatVersion.RUBY1_9)
+    public IRubyObject abs2(ThreadContext context) {
+        return f_mul(context, this, this);
+    }
+
+    /** numeric_arg
+     * 
+     */
+    @JRubyMethod(name = {"arg", "angle", "phase"}, compat = CompatVersion.RUBY1_9)
+    public IRubyObject arg(ThreadContext context) {
+        if (!f_negative_p(context, this)) return RubyFixnum.zero(context.getRuntime());
+        return context.getRuntime().getMath().fastFetchConstant("PI");
+    }    
+
+    /** numeric_rect
+     * 
+     */
+    @JRubyMethod(name = {"rectangular", "rect"}, compat = CompatVersion.RUBY1_9)
+    public IRubyObject rect(ThreadContext context) {
+        return context.getRuntime().newArray(this, RubyFixnum.zero(context.getRuntime()));
+    }    
+
+    /** numeric_polar
+     * 
+     */
+    @JRubyMethod(name = "polar", compat = CompatVersion.RUBY1_9)
+    public IRubyObject polar(ThreadContext context) {
+        return context.getRuntime().newArray(f_abs(context, this), f_arg(context, this));
+    }    
+
+    /** numeric_real
+     * 
+     */
+    @JRubyMethod(name = {"conjugate", "conj"}, compat = CompatVersion.RUBY1_9)
+    public IRubyObject conjugate(ThreadContext context) {
+        return this;
     }
 
     public static class InvalidIntegerException extends NumberFormatException {
@@ -749,6 +915,9 @@ public class RubyNumeric extends RubyObject {
         public InvalidIntegerException(String message) {
             super(message);
         }
+        public Throwable fillInStackTrace() {
+            return this;
+        }
     }
     
     public static class NumberTooLargeException extends NumberFormatException {
@@ -759,7 +928,8 @@ public class RubyNumeric extends RubyObject {
         public NumberTooLargeException(String message) {
             super(message);
         }
-        
+        public Throwable fillInStackTrace() {
+            return this;
+        }
     }
-    
 }
