@@ -32,24 +32,34 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby;
 
+import static org.jruby.RubyEnumerator.enumeratorize;
+
 import java.util.List;
 
+import org.jruby.anno.JRubyMethod;
+import org.jruby.anno.JRubyClass;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
+import org.jruby.runtime.Frame;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
+import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
+import org.jruby.util.ByteList;
 import org.jruby.util.IdUtil;
+import org.jruby.common.IRubyWarnings.ID;
 import org.jruby.exceptions.RaiseException;
+import org.jruby.internal.runtime.methods.CallConfiguration;
+import org.jruby.internal.runtime.methods.DynamicMethod;
 import org.jruby.runtime.ClassIndex;
 
 /**
  * @author  jpetersen
  */
+@JRubyClass(name="Struct")
 public class RubyStruct extends RubyObject {
     private IRubyObject[] values;
 
@@ -60,55 +70,39 @@ public class RubyStruct extends RubyObject {
      */
     public RubyStruct(Ruby runtime, RubyClass rubyClass) {
         super(runtime, rubyClass);
+        
+        int size = RubyNumeric.fix2int(getInternalVariable((RubyClass)rubyClass, "__size__"));
+
+        values = new IRubyObject[size];
+
+        for (int i = 0; i < size; i++) {
+            values[i] = getRuntime().getNil();
+        }
     }
 
     public static RubyClass createStructClass(Ruby runtime) {
         // TODO: NOT_ALLOCATABLE_ALLOCATOR may be ok here, but it's unclear how Structs
         // work with marshalling. Confirm behavior and ensure we're doing this correctly. JRUBY-415
         RubyClass structClass = runtime.defineClass("Struct", runtime.getObject(), ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
+        runtime.setStructClass(structClass);
         structClass.index = ClassIndex.STRUCT;
-        
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyStruct.class);
-        structClass.includeModule(runtime.getModule("Enumerable"));
-
-        structClass.getMetaClass().defineMethod("new", callbackFactory.getOptSingletonMethod("newInstance"));
-
-        structClass.defineMethod("initialize", callbackFactory.getOptMethod("initialize"));
-        structClass.defineFastMethod("initialize_copy", callbackFactory.getFastMethod("initialize_copy", RubyKernel.IRUBY_OBJECT));
-        structClass.defineMethod("clone", callbackFactory.getMethod("rbClone"));
-
-        structClass.defineFastMethod("==", callbackFactory.getFastMethod("equal", RubyKernel.IRUBY_OBJECT));
-        structClass.defineFastMethod("eql?", callbackFactory.getFastMethod("eql_p", RubyKernel.IRUBY_OBJECT));
-
-        structClass.defineFastMethod("to_s", callbackFactory.getFastMethod("to_s"));
-        structClass.defineFastMethod("inspect", callbackFactory.getFastMethod("inspect"));
-        structClass.defineFastMethod("to_a", callbackFactory.getFastMethod("to_a"));
-        structClass.defineFastMethod("values", callbackFactory.getFastMethod("to_a"));
-        structClass.defineFastMethod("size", callbackFactory.getFastMethod("size"));
-        structClass.defineFastMethod("length", callbackFactory.getFastMethod("size"));
-        structClass.defineFastMethod("hash", callbackFactory.getFastMethod("hash"));
-
-        structClass.defineMethod("each", callbackFactory.getMethod("each"));
-        structClass.defineMethod("each_pair", callbackFactory.getMethod("each_pair"));
-        structClass.defineFastMethod("[]", callbackFactory.getFastMethod("aref", RubyKernel.IRUBY_OBJECT));
-        structClass.defineFastMethod("[]=", callbackFactory.getFastMethod("aset", RubyKernel.IRUBY_OBJECT, RubyKernel.IRUBY_OBJECT));
-        structClass.defineFastMethod("values_at", callbackFactory.getFastOptMethod("values_at"));
-
-        structClass.defineFastMethod("members", callbackFactory.getFastMethod("members"));
+        structClass.includeModule(runtime.getEnumerable());
+        structClass.defineAnnotatedMethods(RubyStruct.class);
 
         return structClass;
     }
     
+    @Override
     public int getNativeTypeIndex() {
         return ClassIndex.STRUCT;
     }
     
-    private static IRubyObject getInstanceVariable(RubyClass type, String name) {
-        RubyClass structClass = type.getRuntime().getClass("Struct");
+    private static IRubyObject getInternalVariable(RubyClass type, String internedName) {
+        RubyClass structClass = type.getRuntime().getStructClass();
+        IRubyObject variable;
 
         while (type != null && type != structClass) {
-            IRubyObject variable = type.getInstanceVariable(name);
-            if (variable != null) {
+            if ((variable = type.fastGetInternalVariable(internedName)) != null) {
                 return variable;
             }
 
@@ -130,28 +124,28 @@ public class RubyStruct extends RubyObject {
         }
     }
     
-    public RubyFixnum hash() {
+    @JRubyMethod
+    public RubyFixnum hash(ThreadContext context) {
         Ruby runtime = getRuntime();
-        ThreadContext context = runtime.getCurrentContext();
         int h = getMetaClass().getRealClass().hashCode();
 
         for (int i = 0; i < values.length; i++) {
             h = (h << 1) | (h < 0 ? 1 : 0);
-            h ^= RubyNumeric.num2long(values[i].callMethod(context, MethodIndex.HASH, "hash"));
+            h ^= RubyNumeric.num2long(values[i].callMethod(context, "hash"));
         }
         
         return runtime.newFixnum(h);
     }
 
     private IRubyObject setByName(String name, IRubyObject value) {
-        RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
+        RubyArray member = (RubyArray) getInternalVariable(classOf(), "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
         modify();
 
         for (int i = 0,k=member.getLength(); i < k; i++) {
-            if (member.eltInternal(i).asSymbol().equals(name)) {
+            if (member.eltInternal(i).asJavaString().equals(name)) {
                 return values[i] = value;
             }
         }
@@ -160,12 +154,12 @@ public class RubyStruct extends RubyObject {
     }
 
     private IRubyObject getByName(String name) {
-        RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
+        RubyArray member = (RubyArray) getInternalVariable(classOf(), "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
         for (int i = 0,k=member.getLength(); i < k; i++) {
-            if (member.eltInternal(i).asSymbol().equals(name)) {
+            if (member.eltInternal(i).asJavaString().equals(name)) {
                 return values[i];
             }
         }
@@ -180,62 +174,119 @@ public class RubyStruct extends RubyObject {
      * MRI: rb_struct_s_def / make_struct
      *
      */
+    @JRubyMethod(name = "new", required = 1, rest = true, frame = true, meta = true)
     public static RubyClass newInstance(IRubyObject recv, IRubyObject[] args, Block block) {
         String name = null;
+        boolean nilName = false;
         Ruby runtime = recv.getRuntime();
-        Arity.checkArgumentCount(runtime, args, 1, -1);
 
-        if (args.length > 0 && args[0] instanceof RubyString) {
-            name = args[0].toString();
+        if (args.length > 0) {
+            IRubyObject firstArgAsString = args[0].checkStringType();
+            if (!firstArgAsString.isNil()) {
+                name = ((RubyString)firstArgAsString).getByteList().toString();
+            } else if (args[0].isNil()) {
+                nilName = true;
+            }
         }
 
         RubyArray member = runtime.newArray();
 
-        for (int i = name == null ? 0 : 1; i < args.length; i++) {
-            member.append(RubySymbol.newSymbol(runtime, args[i].asSymbol()));
+        for (int i = (name == null && !nilName) ? 0 : 1; i < args.length; i++) {
+            member.append(runtime.newSymbol(args[i].asJavaString()));
         }
 
         RubyClass newStruct;
         RubyClass superClass = (RubyClass)recv;
 
-        if (name == null) {
-            newStruct = new RubyClass(superClass, STRUCT_INSTANCE_ALLOCATOR);
+        if (name == null || nilName) {
+            newStruct = RubyClass.newClass(runtime, superClass); 
+            newStruct.setAllocator(STRUCT_INSTANCE_ALLOCATOR);
+            newStruct.makeMetaClass(superClass.getMetaClass());
+            newStruct.inherit(superClass);
         } else {
             if (!IdUtil.isConstant(name)) {
                 throw runtime.newNameError("identifier " + name + " needs to be constant", name);
             }
 
             IRubyObject type = superClass.getConstantAt(name);
-
             if (type != null) {
-                runtime.getWarnings().warn(runtime.getCurrentContext().getFramePosition(), "redefining constant Struct::" + name);
+                ThreadContext context = runtime.getCurrentContext();
+                Frame frame = context.getCurrentFrame();
+                runtime.getWarnings().warn(ID.STRUCT_CONSTANT_REDEFINED, frame.getFile(), frame.getLine(), "redefining constant Struct::" + name, name);
+                superClass.remove_const(context, runtime.newString(name));
             }
-            newStruct = superClass.newSubClass(name, STRUCT_INSTANCE_ALLOCATOR, superClass.getCRef(), false);
+            newStruct = superClass.defineClassUnder(name, superClass, STRUCT_INSTANCE_ALLOCATOR);
         }
 
         newStruct.index = ClassIndex.STRUCT;
         
-        newStruct.setInstanceVariable("__size__", member.length());
-        newStruct.setInstanceVariable("__member__", member);
+        newStruct.fastSetInternalVariable("__size__", member.length());
+        newStruct.fastSetInternalVariable("__member__", member);
 
-        CallbackFactory callbackFactory = runtime.callbackFactory(RubyStruct.class);
-        newStruct.getSingletonClass().defineMethod("new", callbackFactory.getOptSingletonMethod("newStruct"));
-        newStruct.getSingletonClass().defineMethod("[]", callbackFactory.getOptSingletonMethod("newStruct"));
-        newStruct.getSingletonClass().defineMethod("members", callbackFactory.getSingletonMethod("members"));
+        newStruct.getSingletonClass().defineAnnotatedMethods(StructMethods.class);
 
         // define access methods.
-        for (int i = name == null ? 0 : 1; i < args.length; i++) {
-            String memberName = args[i].asSymbol();
-            newStruct.defineMethod(memberName, callbackFactory.getMethod("get"));
-            newStruct.defineMethod(memberName + "=", callbackFactory.getMethod("set", RubyKernel.IRUBY_OBJECT));
+        for (int i = (name == null && !nilName) ? 0 : 1; i < args.length; i++) {
+            final String memberName = args[i].asJavaString();
+            // if we are storing a name as well, index is one too high for values
+            final int index = (name == null && !nilName) ? i : i - 1;
+            newStruct.addMethod(memberName, new DynamicMethod(newStruct, Visibility.PUBLIC, CallConfiguration.NO_FRAME_NO_SCOPE) {
+                @Override
+                public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
+                    Arity.checkArgumentCount(self.getRuntime(), args, 0, 0);
+                    return ((RubyStruct)self).get(index);
+                }
+
+                @Override
+                public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name) {
+                    return ((RubyStruct)self).get(index);
+                }
+
+                @Override
+                public DynamicMethod dup() {
+                    return this;
+                }
+            });
+            newStruct.addMethod(memberName + "=", new DynamicMethod(newStruct, Visibility.PUBLIC, CallConfiguration.NO_FRAME_NO_SCOPE) {
+                @Override
+                public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
+                    Arity.checkArgumentCount(self.getRuntime(), args, 1, 1);
+                    return ((RubyStruct)self).set(args[0], index);
+                }
+
+                @Override
+                public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg) {
+                    return ((RubyStruct)self).set(arg, index);
+                }
+
+                @Override
+                public DynamicMethod dup() {
+                    return this;
+                }
+            });
         }
         
         if (block.isGiven()) {
+            // Struct bodies should be public by default, so set block visibility to public. JRUBY-1185.
+            block.getBinding().setVisibility(Visibility.PUBLIC);
             block.yield(runtime.getCurrentContext(), null, newStruct, newStruct, false);
         }
 
         return newStruct;
     }
+    
+    // For binding purposes on the newly created struct types
+    public static class StructMethods {
+        @JRubyMethod(name = {"new", "[]"}, rest = true, frame = true)
+        public static IRubyObject newStruct(IRubyObject recv, IRubyObject[] args, Block block) {
+            return RubyStruct.newStruct(recv, args, block);
+        }
+        
+        @JRubyMethod
+        public static IRubyObject members(IRubyObject recv, Block block) {
+            return RubyStruct.members(recv, block);
+        }
+    }
 
     /** Create new Structure.
      *
@@ -245,19 +296,16 @@ public class RubyStruct extends RubyObject {
     public static RubyStruct newStruct(IRubyObject recv, IRubyObject[] args, Block block) {
         RubyStruct struct = new RubyStruct(recv.getRuntime(), (RubyClass) recv);
 
-        int size = RubyNumeric.fix2int(getInstanceVariable((RubyClass) recv, "__size__"));
-
-        struct.values = new IRubyObject[size];
-
         struct.callInit(args, block);
 
         return struct;
     }
 
+    @JRubyMethod(rest = true, frame = true, visibility = Visibility.PRIVATE)
     public IRubyObject initialize(IRubyObject[] args, Block unusedBlock) {
         modify();
 
-        int size = RubyNumeric.fix2int(getInstanceVariable(getMetaClass(), "__size__"));
+        int size = RubyNumeric.fix2int(getInternalVariable(getMetaClass(), "__size__"));
 
         if (args.length > size) {
             throw getRuntime().newArgumentError("struct size differs (" + args.length +" for " + size + ")");
@@ -267,152 +315,145 @@ public class RubyStruct extends RubyObject {
             values[i] = args[i];
         }
 
-        for (int i = args.length; i < size; i++) {
-            values[i] = getRuntime().getNil();
-        }
-
         return getRuntime().getNil();
     }
     
     public static RubyArray members(IRubyObject recv, Block block) {
-        RubyArray member = (RubyArray) getInstanceVariable((RubyClass) recv, "__member__");
+        RubyArray member = (RubyArray) getInternalVariable((RubyClass) recv, "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
         RubyArray result = recv.getRuntime().newArray(member.getLength());
         for (int i = 0,k=member.getLength(); i < k; i++) {
-            result.append(recv.getRuntime().newString(member.eltInternal(i).asSymbol()));
+            result.append(recv.getRuntime().newString(member.eltInternal(i).asJavaString()));
         }
 
         return result;
     }
 
+    @JRubyMethod
     public RubyArray members() {
         return members(classOf(), Block.NULL_BLOCK);
     }
-
-    public IRubyObject set(IRubyObject value, Block block) {
-        String name = getRuntime().getCurrentContext().getFrameName();
-        if (name.endsWith("=")) {
-            name = name.substring(0, name.length() - 1);
+    
+    @JRubyMethod
+    public RubyArray select(ThreadContext context, Block block) {
+        RubyArray array = RubyArray.newArray(context.getRuntime());
+        
+        for (int i = 0; i < values.length; i++) {
+            if (block.yield(context, values[i]).isTrue()) {
+                array.append(values[i]);
+            }
         }
+        
+        return array;
+    }
 
-        RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
+    public IRubyObject set(IRubyObject value, int index) {
+        RubyArray member = (RubyArray) getInternalVariable(classOf(), "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
         modify();
 
-        for (int i = 0,k=member.getLength(); i < k; i++) {
-            if (member.eltInternal(i).asSymbol().equals(name)) {
-                return values[i] = value;
-            }
-        }
-
-        throw notStructMemberError(name);
+        return values[index] = value;
     }
 
     private RaiseException notStructMemberError(String name) {
         return getRuntime().newNameError(name + " is not struct member", name);
     }
 
-    public IRubyObject get(Block block) {
-        String name = getRuntime().getCurrentContext().getFrameName();
-
-        RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
+    public IRubyObject get(int index) {
+        RubyArray member = (RubyArray) getInternalVariable(classOf(), "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
-        for (int i = 0,k=member.getLength(); i < k; i++) {
-            if (member.eltInternal(i).asSymbol().equals(name)) {
-                return values[i];
-            }
-        }
-
-        throw notStructMemberError(name);
+        return values[index];
     }
 
-    public IRubyObject rbClone(Block block) {
-        RubyStruct clone = new RubyStruct(getRuntime(), getMetaClass());
-
-        clone.values = new IRubyObject[values.length];
-        System.arraycopy(values, 0, clone.values, 0, values.length);
-
-        clone.setFrozen(this.isFrozen());
-        clone.setTaint(this.isTaint());
-
-        return clone;
+    @Override
+    public void copySpecialInstanceVariables(IRubyObject clone) {
+        RubyStruct struct = (RubyStruct)clone;
+        struct.values = new IRubyObject[values.length];
+        System.arraycopy(values, 0, struct.values, 0, values.length);
     }
 
-    public IRubyObject equal(IRubyObject other) {
+    @JRubyMethod(name = "==", required = 1)
+    public IRubyObject op_equal(ThreadContext context, IRubyObject other) {
         if (this == other) return getRuntime().getTrue();
         if (!(other instanceof RubyStruct)) return getRuntime().getFalse();
         if (getMetaClass().getRealClass() != other.getMetaClass().getRealClass()) return getRuntime().getFalse();
         
         Ruby runtime = getRuntime();
-        ThreadContext context = runtime.getCurrentContext();
         RubyStruct otherStruct = (RubyStruct)other;
-            for (int i = 0; i < values.length; i++) {
-            if (!values[i].equalInternal(context, otherStruct.values[i]).isTrue()) {
-                return runtime.getFalse();
-                }
-            }
-        return runtime.getTrue();
+        for (int i = 0; i < values.length; i++) {
+            if (!equalInternal(context, values[i], otherStruct.values[i])) return runtime.getFalse();
         }
+        return runtime.getTrue();
+    }
     
-    public IRubyObject eql_p(IRubyObject other) {
+    @JRubyMethod(name = "eql?", required = 1)
+    public IRubyObject eql_p(ThreadContext context, IRubyObject other) {
         if (this == other) return getRuntime().getTrue();
         if (!(other instanceof RubyStruct)) return getRuntime().getFalse();
         if (getMetaClass() != other.getMetaClass()) return getRuntime().getFalse();
         
         Ruby runtime = getRuntime();
-        ThreadContext context = runtime.getCurrentContext();
         RubyStruct otherStruct = (RubyStruct)other;
         for (int i = 0; i < values.length; i++) {
-            if (!values[i].eqlInternal(context, otherStruct.values[i])) {
-                return runtime.getFalse();
-    }
+            if (!eqlInternal(context, values[i], otherStruct.values[i])) return runtime.getFalse();
         }
         return runtime.getTrue();        
     }
 
-    public IRubyObject to_s() {
-        return inspect();
-    }
-
-    public IRubyObject inspect() {
-        RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
+    /** inspect_struct
+    *
+    */
+    private IRubyObject inspectStruct(final ThreadContext context) {    
+        RubyArray member = (RubyArray) getInternalVariable(classOf(), "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
-        StringBuffer sb = new StringBuffer(100);
-
-        sb.append("#<struct ").append(getMetaClass().getRealClass().getName()).append(' ');
+        ByteList buffer = new ByteList("#<struct ".getBytes());
+        buffer.append(getMetaClass().getRealClass().getRealClass().getName().getBytes());
+        buffer.append(' ');
 
         for (int i = 0,k=member.getLength(); i < k; i++) {
-            if (i > 0) {
-                sb.append(", ");
-            }
-
-            sb.append(member.eltInternal(i).asSymbol()).append("=");
-            sb.append(values[i].callMethod(getRuntime().getCurrentContext(), "inspect"));
+            if (i > 0) buffer.append(',').append(' ');
+            // FIXME: MRI has special case for constants here 
+            buffer.append(RubyString.objAsString(context, member.eltInternal(i)).getByteList());
+            buffer.append('=');
+            buffer.append(inspect(context, values[i]).getByteList());
         }
 
-        sb.append('>');
-
-        return getRuntime().newString(sb.toString()); // OBJ_INFECT
+        buffer.append('>');
+        return getRuntime().newString(buffer); // OBJ_INFECT        
     }
 
+    @JRubyMethod(name = {"inspect", "to_s"})
+    public IRubyObject inspect(ThreadContext context) {
+        if (getRuntime().isInspecting(this)) return getRuntime().newString("#<struct " + getMetaClass().getRealClass().getName() + ":...>");
+
+        try {
+            getRuntime().registerInspecting(this);
+            return inspectStruct(context);
+        } finally {
+            getRuntime().unregisterInspecting(this);
+        }
+    }
+
+    @JRubyMethod(name = {"to_a", "values"})
     public RubyArray to_a() {
         return getRuntime().newArray(values);
     }
 
+    @JRubyMethod(name = {"size", "length"} )
     public RubyFixnum size() {
         return getRuntime().newFixnum(values.length);
     }
 
-    public IRubyObject each(Block block) {
-        ThreadContext context = getRuntime().getCurrentContext();
+    @JRubyMethod(name = "each", backtrace = true)
+    public IRubyObject each(ThreadContext context, Block block) {
         for (int i = 0; i < values.length; i++) {
             block.yield(context, values[i]);
         }
@@ -420,12 +461,17 @@ public class RubyStruct extends RubyObject {
         return this;
     }
 
-    public IRubyObject each_pair(Block block) {
-        RubyArray member = (RubyArray) getInstanceVariable(classOf(), "__member__");
+    @JRubyMethod(name = "each", frame = true, compat = CompatVersion.RUBY1_9)
+    public IRubyObject each19(final ThreadContext context, final Block block) {
+        return block.isGiven() ? each(context, block) : enumeratorize(context.getRuntime(), this, "each");
+    }
+
+    @JRubyMethod(name = "each_pair", frame = true)
+    public IRubyObject each_pair(ThreadContext context, Block block) {
+        RubyArray member = (RubyArray) getInternalVariable(classOf(), "__member__");
 
         assert !member.isNil() : "uninitialized struct";
 
-        ThreadContext context = getRuntime().getCurrentContext();
         for (int i = 0; i < values.length; i++) {
             block.yield(context, getRuntime().newArrayNoCopy(new IRubyObject[]{member.eltInternal(i), values[i]}));
         }
@@ -433,9 +479,15 @@ public class RubyStruct extends RubyObject {
         return this;
     }
 
+    @JRubyMethod(name = "each_pair", frame = true, compat = CompatVersion.RUBY1_9)
+    public IRubyObject each_pair19(final ThreadContext context, final Block block) {
+        return block.isGiven() ? each_pair(context, block) : enumeratorize(context.getRuntime(), this, "each_pair");
+    }
+
+    @JRubyMethod(name = "[]", required = 1)
     public IRubyObject aref(IRubyObject key) {
         if (key instanceof RubyString || key instanceof RubySymbol) {
-            return getByName(key.asSymbol());
+            return getByName(key.asJavaString());
         }
 
         int idx = RubyNumeric.fix2int(key);
@@ -451,9 +503,10 @@ public class RubyStruct extends RubyObject {
         return values[idx];
     }
 
+    @JRubyMethod(name = "[]=", required = 2)
     public IRubyObject aset(IRubyObject key, IRubyObject value) {
         if (key instanceof RubyString || key instanceof RubySymbol) {
-            return setByName(key.asSymbol(), value);
+            return setByName(key.asJavaString(), value);
         }
 
         int idx = RubyNumeric.fix2int(key);
@@ -473,6 +526,7 @@ public class RubyStruct extends RubyObject {
     // FIXME: This is copied code from RubyArray.  Both RE, Struct, and Array should share one impl
     // This is also hacky since I construct ruby objects to access ruby arrays through aref instead
     // of something lower.
+    @JRubyMethod(rest = true)
     public IRubyObject values_at(IRubyObject[] args) {
         long olen = values.length;
         RubyArray result = getRuntime().newArray(args.length);
@@ -503,9 +557,10 @@ public class RubyStruct extends RubyObject {
     }
 
     public static void marshalTo(RubyStruct struct, MarshalStream output) throws java.io.IOException {
+        output.registerLinkTarget(struct);
         output.dumpDefaultObjectHeader('S', struct.getMetaClass());
 
-        List members = ((RubyArray) getInstanceVariable(struct.classOf(), "__member__")).getList();
+        List members = ((RubyArray) getInternalVariable(struct.classOf(), "__member__")).getList();
         output.writeInt(members.size());
 
         for (int i = 0; i < members.size(); i++) {
@@ -519,9 +574,9 @@ public class RubyStruct extends RubyObject {
         Ruby runtime = input.getRuntime();
 
         RubySymbol className = (RubySymbol) input.unmarshalObject();
-        RubyClass rbClass = pathToClass(runtime, className.asSymbol());
+        RubyClass rbClass = pathToClass(runtime, className.asJavaString());
         if (rbClass == null) {
-            throw runtime.newNameError("uninitialized constant " + className, className.asSymbol());
+            throw runtime.newNameError("uninitialized constant " + className, className.asJavaString());
         }
 
         RubyArray mem = members(rbClass, Block.NULL_BLOCK);
@@ -559,6 +614,8 @@ public class RubyStruct extends RubyObject {
         }
     };
     
+    @Override
+    @JRubyMethod(required = 1)
     public IRubyObject initialize_copy(IRubyObject arg) {
         if (this == arg) return this;
         RubyStruct original = (RubyStruct) arg;
