@@ -40,12 +40,16 @@ import java.io.PrintWriter;
 import java.io.StringWriter;
 
 import org.jruby.*;
+import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Block;
+import org.jruby.runtime.EventHook;
+import org.jruby.runtime.RubyEvent;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 
 public class RaiseException extends JumpException {
+    public static final boolean DEBUG = false;
     private static final long serialVersionUID = -7612079169559973951L;
     
     private RubyException exception;
@@ -55,15 +59,26 @@ public class RaiseException extends JumpException {
     }
 
     public RaiseException(Ruby runtime, RubyClass excptnClass, String msg, boolean nativeException) {
-        super(msg, JumpType.RaiseJump);
+        super(msg);
         if (msg == null) {
             msg = "No message available";
         }
-        setException((RubyException) excptnClass.callMethod(runtime.getCurrentContext(), "new", new IRubyObject[] {excptnClass.getRuntime().newString(msg)}, Block.NULL_BLOCK), nativeException);
+        if (DEBUG) {
+            Thread.dumpStack();
+        }
+        setException((RubyException)RuntimeHelpers.invoke(
+                runtime.getCurrentContext(),
+                excptnClass,
+                "new",
+                RubyString.newUnicodeString(excptnClass.getRuntime(), msg)),
+                nativeException);
     }
 
     public RaiseException(RubyException exception, boolean isNativeException) {
-        super(JumpType.RaiseJump);
+        super();
+        if (DEBUG) {
+            Thread.dumpStack();
+        }
         setException(exception, isNativeException);
     }
 
@@ -73,7 +88,7 @@ public class RaiseException extends JumpException {
     }
 
     private static String buildMessage(Throwable exception) {
-        StringBuffer sb = new StringBuffer();
+        StringBuilder sb = new StringBuilder();
         StringWriter stackTrace = new StringWriter();
         exception.printStackTrace(new PrintWriter(stackTrace));
     
@@ -85,7 +100,7 @@ public class RaiseException extends JumpException {
     }
 
     public RaiseException(Throwable cause, NativeException nativeException) {
-        super(buildMessage(cause), cause, JumpType.RaiseJump);
+        super(buildMessage(cause), cause);
         setException(nativeException, false);
     }
 
@@ -109,9 +124,14 @@ public class RaiseException extends JumpException {
             runtime.getGlobalVariables().set("$!", newException);
         }
 
-        if (runtime.getTraceFunction() != null) {
-            runtime.callTraceFunction(context, "return", context.getPosition(),
-                    RubyBinding.newBinding(runtime), context.getFrameName(), context.getFrameKlazz());
+        if (runtime.hasEventHooks()) {
+            runtime.callEventHooks(
+                    context,
+                    RubyEvent.RAISE,
+                    context.getFile(),
+                    context.getLine(),
+                    context.getFrameName(),
+                    context.getFrameKlazz());
         }
 
         this.exception = newException;
@@ -122,18 +142,19 @@ public class RaiseException extends JumpException {
 
         runtime.setStackTraces(runtime.getStackTraces() + 1);
 
-        if (newException.callMethod(context, "backtrace").isNil() && context.getSourceFile() != null) {
-            IRubyObject backtrace = context.createBacktrace(0, nativeException);
-            newException.callMethod(context, "set_backtrace", backtrace);
+        StackTraceElement[] stackTrace = context.createBacktrace2(0, nativeException);
+        newException.setBacktraceFrames(stackTrace);
+
+        // JRUBY-2673: if wrapping a NativeException, use the actual Java exception's trace as our Java trace
+        if (newException instanceof NativeException) {
+            setStackTrace(((NativeException)newException).getCause().getStackTrace());
+        } else {
+            setStackTrace(stackTrace);
         }
 
         runtime.setStackTraces(runtime.getStackTraces() - 1);
     }
 
-    public Throwable fillInStackTrace() {
-        return originalFillInStackTrace();
-    }
-    
     public void printStackTrace() {
         printStackTrace(System.err);
     }
@@ -150,10 +171,10 @@ public class RaiseException extends JumpException {
         IRubyObject backtrace = exception.backtrace();
         Ruby runtime = backtrace.getRuntime();
         if (runtime.getNil() != backtrace) {
-            String firstLine = backtrace.callMethod(runtime.getCurrentContext(), "first").callMethod(runtime.getCurrentContext(), MethodIndex.TO_S, "to_s").toString();
+            String firstLine = backtrace.callMethod(runtime.getCurrentContext(), "first").convertToString().toString();
             ps.print(firstLine + ": ");
         }
-        ps.println(exception.message + " (" + exception.getMetaClass().toString() + ")");
+        ps.println(exception.message.convertToString() + " (" + exception.getMetaClass().toString() + ")");
         exception.printBacktrace(ps);
         ps.println("\t...internal jruby stack elided...");
         for (int i = externalIndex; i < trace.length; i++) {
