@@ -30,59 +30,82 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.internal.runtime;
 
-import edu.emory.mathcs.backport.java.util.concurrent.locks.ReentrantLock;
-import java.lang.ref.WeakReference;
+import java.lang.ref.SoftReference;
+import java.util.concurrent.locks.ReentrantLock;
 
 import java.util.ArrayList;
 import java.util.Collections;
 import java.util.Iterator;
 import java.util.List;
-import java.util.Set;
+import java.util.Map;
 
+import java.util.WeakHashMap;
 import org.jruby.Ruby;
 import org.jruby.RubyThread;
 import org.jruby.runtime.ThreadContext;
-import org.jruby.util.collections.WeakHashSet;
 
 public class ThreadService {
     private Ruby runtime;
     private ThreadContext mainContext;
-    private ThreadLocal localContext;
+    private ThreadLocal<SoftReference<ThreadContext>> localContext;
     private ThreadGroup rubyThreadGroup;
-    private Set rubyThreadList;
-    private Thread mainThread;
+    private Map<Thread, RubyThread> rubyThreadMap;
     
     private ReentrantLock criticalLock = new ReentrantLock();
 
     public ThreadService(Ruby runtime) {
         this.runtime = runtime;
         this.mainContext = ThreadContext.newContext(runtime);
-        this.localContext = new ThreadLocal();
+        this.localContext = new ThreadLocal<SoftReference<ThreadContext>>();
         this.rubyThreadGroup = new ThreadGroup("Ruby Threads#" + runtime.hashCode());
-        this.rubyThreadList = Collections.synchronizedSet(new WeakHashSet());
+        this.rubyThreadMap = Collections.synchronizedMap(new WeakHashMap<Thread, RubyThread>());
         
         // Must be called from main thread (it is currently, but this bothers me)
-        mainThread = Thread.currentThread();
-        localContext.set(new WeakReference(mainContext));
-        rubyThreadList.add(mainThread);
+        localContext.set(new SoftReference<ThreadContext>(mainContext));
     }
 
     public void disposeCurrentThread() {
         localContext.set(null);
+        rubyThreadMap.remove(Thread.currentThread());
     }
 
+    /**
+     * In order to provide an appropriate execution context for a given thread,
+     * we store ThreadContext instances in a threadlocal. This method is a utility
+     * to get at that threadlocal context from anywhere in the program it may
+     * not be immediately available. This method should be used sparingly, and
+     * if it is possible to pass ThreadContext on the argument list, it is
+     * preferable.
+     * 
+     * <b>Description of behavior</b>
+     * 
+     * The threadlocal does not actually contain the ThreadContext directly;
+     * instead, it contains a SoftReference that holds the ThreadContext. This
+     * is to allow new threads to enter the system and execute Ruby code with
+     * a valid context, but still allow that context to garbage collect if the
+     * thread stays alive much longer. We use SoftReference here because
+     * WeakReference is collected too quickly, resulting in very expensive
+     * ThreadContext churn (and this originally lead to JRUBY-2261's leak of
+     * adopted RubyThread instances).
+     * 
+     * @return The ThreadContext instance for the current thread, or a new one
+     * if none has previously been created or the old ThreadContext has been
+     * collected.
+     */
     public ThreadContext getCurrentContext() {
-        WeakReference wr = null;
+        SoftReference sr = null;
         ThreadContext context = null;
         
         while (context == null) {
-            // loop until a context is available, to clean up weakrefs that might get collected
-            if ((wr = (WeakReference)localContext.get()) == null) {
-                wr = adoptCurrentThread();
-                context = (ThreadContext)wr.get();
+            // loop until a context is available, to clean up softrefs that might have been collected
+            if ((sr = (SoftReference)localContext.get()) == null) {
+                sr = adoptCurrentThread();
+                context = (ThreadContext)sr.get();
             } else {
-                context = (ThreadContext)wr.get();
+                context = (ThreadContext)sr.get();
             }
+            
+            // context is null, wipe out the SoftReference (this could be done with a reference queue)
             if (context == null) {
                 localContext.set(null);
             }
@@ -91,35 +114,35 @@ public class ThreadService {
         return context;
     }
     
-    private WeakReference adoptCurrentThread() {
+    private SoftReference adoptCurrentThread() {
         Thread current = Thread.currentThread();
         
-        RubyThread.adopt(runtime.getClass("Thread"), current);
+        RubyThread.adopt(runtime.getThread(), current);
         
-        return (WeakReference) localContext.get();
+        return (SoftReference) localContext.get();
     }
 
     public RubyThread getMainThread() {
         return mainContext.getThread();
     }
 
-    public void setMainThread(RubyThread thread) {
-        mainContext.setThread(thread);
+    public void setMainThread(Thread thread, RubyThread rubyThread) {
+        mainContext.setThread(rubyThread);
+        rubyThreadMap.put(thread, rubyThread);
     }
     
     public synchronized RubyThread[] getActiveRubyThreads() {
     	// all threads in ruby thread group plus main thread
 
-        synchronized(rubyThreadList) {
-            List rtList = new ArrayList(rubyThreadList.size());
+        synchronized(rubyThreadMap) {
+            List<RubyThread> rtList = new ArrayList<RubyThread>(rubyThreadMap.size());
         
-            for (Iterator iter = rubyThreadList.iterator(); iter.hasNext();) {
-                Thread t = (Thread)iter.next();
+            for (Map.Entry<Thread, RubyThread> entry : rubyThreadMap.entrySet()) {
+                Thread t = entry.getKey();
             
                 if (!t.isAlive()) continue;
             
-                RubyThread rt = getRubyThreadFromThread(t);
-                rtList.add(rt);
+                rtList.add(entry.getValue());
             }
         
             RubyThread[] rubyThreads = new RubyThread[rtList.size()];
@@ -133,32 +156,28 @@ public class ThreadService {
     	return rubyThreadGroup;
     }
 
-    public synchronized void registerNewThread(RubyThread thread) {
-        localContext.set(new WeakReference(ThreadContext.newContext(runtime)));
+    public synchronized ThreadContext registerNewThread(RubyThread thread) {
+        ThreadContext context = ThreadContext.newContext(runtime);
+        localContext.set(new SoftReference(context));
         getCurrentContext().setThread(thread);
         // This requires register to be called from within the registree thread
-        rubyThreadList.add(Thread.currentThread());
+        rubyThreadMap.put(Thread.currentThread(), thread);
+        return context;
     }
     
     public synchronized void unregisterThread(RubyThread thread) {
-        rubyThreadList.remove(Thread.currentThread());
+        rubyThreadMap.remove(Thread.currentThread());
         getCurrentContext().setThread(null);
         localContext.set(null);
     }
     
     private RubyThread getRubyThreadFromThread(Thread activeThread) {
-        RubyThread rubyThread;
-        if (activeThread instanceof RubyNativeThread) {
-            RubyNativeThread rubyNativeThread = (RubyNativeThread)activeThread;
-            rubyThread = rubyNativeThread.getRubyThread();
-        } else {
-            // main thread
-            rubyThread = mainContext.getThread();
-        }
+        RubyThread rubyThread = rubyThreadMap.get(activeThread);
+        
         return rubyThread;
     }
     
-    public synchronized void setCritical(boolean critical) {
+    public void setCritical(boolean critical) {
         if (criticalLock.isHeldByCurrentThread()) {
             if (critical) {
                 // do nothing
@@ -174,7 +193,7 @@ public class ThreadService {
         }
     }
     
-    public synchronized boolean getCritical() {
+    public boolean getCritical() {
         return criticalLock.isHeldByCurrentThread();
     }
     
