@@ -36,52 +36,52 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.javasupport;
 
+import java.lang.annotation.Annotation;
 import java.lang.reflect.AccessibleObject;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
+import java.lang.reflect.Type;
 
 import org.jruby.Ruby;
 import org.jruby.RubyBoolean;
 import org.jruby.RubyClass;
 import org.jruby.RubyModule;
 import org.jruby.RubyString;
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
 import org.jruby.javasupport.proxy.InternalJavaProxy;
 import org.jruby.javasupport.proxy.JavaProxyClass;
 import org.jruby.javasupport.proxy.JavaProxyMethod;
-import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.builtin.IRubyObject;
 
+@JRubyClass(name="Java::JavaMethod")
 public class JavaMethod extends JavaCallable {
     private final Method method;
-    private final Class[] parameterTypes;
+    private final Class<?>[] parameterTypes;
+    private final JavaUtil.JavaConverter returnConverter;
+
+    public Object getValue() {
+        return method;
+    }
 
     public static RubyClass createJavaMethodClass(Ruby runtime, RubyModule javaModule) {
         // TODO: NOT_ALLOCATABLE_ALLOCATOR is probably ok here, since we don't intend for people to monkey with
         // this type and it can't be marshalled. Confirm. JRUBY-415
         RubyClass result = 
             javaModule.defineClassUnder("JavaMethod", runtime.getObject(), ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
-        CallbackFactory callbackFactory = runtime.callbackFactory(JavaMethod.class);
 
         JavaAccessibleObject.registerRubyMethods(runtime, result);
+        JavaCallable.registerRubyMethods(runtime, result);
         
-        result.defineFastMethod("name", callbackFactory.getFastMethod("name"));
-        result.defineFastMethod("arity", callbackFactory.getFastMethod("arity"));
-        result.defineFastMethod("public?", callbackFactory.getFastMethod("public_p"));
-        result.defineFastMethod("final?", callbackFactory.getFastMethod("final_p"));
-        result.defineFastMethod("static?", callbackFactory.getFastMethod("static_p"));
-        result.defineFastMethod("invoke", callbackFactory.getFastOptMethod("invoke"));
-        result.defineFastMethod("invoke_static", callbackFactory.getFastOptMethod("invoke_static"));
-        result.defineFastMethod("argument_types", callbackFactory.getFastMethod("argument_types"));
-        result.defineFastMethod("inspect", callbackFactory.getFastMethod("inspect"));
-        result.defineFastMethod("return_type", callbackFactory.getFastMethod("return_type"));
+        result.defineAnnotatedMethods(JavaMethod.class);
 
         return result;
     }
 
     public JavaMethod(Ruby runtime, Method method) {
-        super(runtime, (RubyClass) runtime.getModule("Java").getClass("JavaMethod"));
+        super(runtime, runtime.getJavaSupport().getJavaMethodClass());
         this.method = method;
         this.parameterTypes = method.getParameterTypes();
 
@@ -91,15 +91,17 @@ public class JavaMethod extends JavaCallable {
         if (Modifier.isPublic(method.getModifiers()) &&
             Modifier.isPublic(method.getClass().getModifiers()) &&
             !Modifier.isPublic(method.getDeclaringClass().getModifiers())) {
-            accesibleObject().setAccessible(true);
+            accessibleObject().setAccessible(true);
         }
+        
+        returnConverter = JavaUtil.getJavaConverter(method.getReturnType());
     }
 
     public static JavaMethod create(Ruby runtime, Method method) {
         return new JavaMethod(runtime, method);
     }
 
-    public static JavaMethod create(Ruby runtime, Class javaClass, String methodName, Class[] argumentTypes) {
+    public static JavaMethod create(Ruby runtime, Class<?> javaClass, String methodName, Class<?>[] argumentTypes) {
         try {
             Method method = javaClass.getMethod(methodName, argumentTypes);
             return create(runtime, method);
@@ -109,44 +111,95 @@ public class JavaMethod extends JavaCallable {
         }
     }
 
-    public static JavaMethod createDeclared(Ruby runtime, Class javaClass, String methodName, Class[] argumentTypes) {
+    public static JavaMethod createDeclared(Ruby runtime, Class<?> javaClass, String methodName, Class<?>[] argumentTypes) {
         try {
-            Method method = javaClass.getDeclaredMethod(methodName, argumentTypes);
-            return create(runtime, method);
+            return create(runtime, javaClass.getDeclaredMethod(methodName, argumentTypes));
         } catch (NoSuchMethodException e) {
             throw runtime.newNameError("undefined method '" + methodName + "' for class '" + javaClass.getName() + "'",
                     methodName);
         }
     }
 
+    public static JavaMethod getMatchingDeclaredMethod(Ruby runtime, Class<?> javaClass, String methodName, Class<?>[] argumentTypes) {
+        // FIXME: do we really want 'declared' methods?  includes private/protected, and does _not_
+        // include superclass methods.  also, the getDeclared calls may throw SecurityException if
+        // we're running under a restrictive security policy.
+        try {
+            return create(runtime, javaClass.getDeclaredMethod(methodName, argumentTypes));
+        } catch (NoSuchMethodException e) {
+            // search through all declared methods to find a closest match
+            MethodSearch: for (Method method : javaClass.getDeclaredMethods()) {
+                if (method.getName().equals(methodName)) {
+                    Class<?>[] targetTypes = method.getParameterTypes();
+                
+                    // for zero args case we can stop searching
+                    if (targetTypes.length == 0 && argumentTypes.length == 0) {
+                        return create(runtime, method);
+                    }
+                    
+                    TypeScan: for (int i = 0; i < argumentTypes.length; i++) {
+                        if (i >= targetTypes.length) continue MethodSearch;
+
+                        if (targetTypes[i].isAssignableFrom(argumentTypes[i])) {
+                            continue TypeScan;
+                        } else {
+                            continue MethodSearch;
+                        }
+                    }
+
+                    // if we get here, we found a matching method, use it
+                    // TODO: choose narrowest method by continuing to search
+                    return create(runtime, method);
+                }
+            }
+        }
+        // no matching method found
+        return null;
+    }
+    
+    public boolean equals(Object other) {
+        return other instanceof JavaMethod &&
+            this.method == ((JavaMethod)other).method;
+    }
+    
+    public int hashCode() {
+        return method.hashCode();
+    }
+
+    @JRubyMethod
     public RubyString name() {
         return getRuntime().newString(method.getName());
     }
 
-    protected int getArity() {
+    public int getArity() {
         return parameterTypes.length;
     }
 
+    @JRubyMethod(name = "public?")
     public RubyBoolean public_p() {
         return getRuntime().newBoolean(Modifier.isPublic(method.getModifiers()));
     }
 
+    @JRubyMethod(name = "final?")
     public RubyBoolean final_p() {
         return getRuntime().newBoolean(Modifier.isFinal(method.getModifiers()));
     }
 
+    @JRubyMethod(rest = true)
     public IRubyObject invoke(IRubyObject[] args) {
         if (args.length != 1 + getArity()) {
             throw getRuntime().newArgumentError(args.length, 1 + getArity());
         }
 
-        IRubyObject invokee = args[0];
-        if (! (invokee instanceof JavaObject)) {
-            throw getRuntime().newTypeError("invokee not a java object");
-        }
-        Object javaInvokee = ((JavaObject) invokee).getValue();
         Object[] arguments = new Object[args.length - 1];
-        convertArguments(arguments, args, 1);
+        convertArguments(getRuntime(), arguments, args, 1);
+
+        IRubyObject invokee = args[0];
+        if(invokee.isNil()) {
+            return invokeWithExceptionHandling(method, null, arguments);
+        }
+
+        Object javaInvokee = JavaUtil.unwrapJavaObject(getRuntime(), invokee, "invokee not a java object").getValue();
 
         if (! method.getDeclaringClass().isInstance(javaInvokee)) {
             throw getRuntime().newTypeError("invokee not instance of method's class (" +
@@ -172,18 +225,62 @@ public class JavaMethod extends JavaCallable {
         return invokeWithExceptionHandling(method, javaInvokee, arguments);
     }
 
+    public IRubyObject invoke(IRubyObject self, Object[] args) {
+        if (args.length != getArity()) {
+            throw getRuntime().newArgumentError(args.length, getArity());
+        }
+
+        if (! (self instanceof JavaObject)) {
+            throw getRuntime().newTypeError("invokee not a java object");
+        }
+        Object javaInvokee = ((JavaObject) self).getValue();
+
+        if (! method.getDeclaringClass().isInstance(javaInvokee)) {
+            throw getRuntime().newTypeError("invokee not instance of method's class (" +
+                                              "got" + javaInvokee.getClass().getName() + " wanted " +
+                                              method.getDeclaringClass().getName() + ")");
+        }
+        
+        //
+        // this test really means, that this is a ruby-defined subclass of a java class
+        //
+        if (javaInvokee instanceof InternalJavaProxy &&
+                // don't bother to check if final method, it won't
+                // be there (not generated, can't be!)
+                !Modifier.isFinal(method.getModifiers())) {
+            JavaProxyClass jpc = ((InternalJavaProxy) javaInvokee)
+                    .___getProxyClass();
+            JavaProxyMethod jpm;
+            if ((jpm = jpc.getMethod(method.getName(), parameterTypes)) != null &&
+                    jpm.hasSuperImplementation()) {
+                return invokeWithExceptionHandling(jpm.getSuperMethod(), javaInvokee, args);
+            }
+        }
+        return invokeWithExceptionHandling(method, javaInvokee, args);
+    }
+
+    @JRubyMethod(rest = true)
     public IRubyObject invoke_static(IRubyObject[] args) {
         if (args.length != getArity()) {
             throw getRuntime().newArgumentError(args.length, getArity());
         }
         Object[] arguments = new Object[args.length];
         System.arraycopy(args, 0, arguments, 0, arguments.length);
-        convertArguments(arguments, args, 0);
+        convertArguments(getRuntime(), arguments, args, 0);
         return invokeWithExceptionHandling(method, null, arguments);
     }
 
+    public IRubyObject invoke_static(Object[] args) {
+        if (args.length != getArity()) {
+            throw getRuntime().newArgumentError(args.length, getArity());
+        }
+
+        return invokeWithExceptionHandling(method, null, args);
+    }
+
+    @JRubyMethod
     public IRubyObject return_type() {
-        Class klass = method.getReturnType();
+        Class<?> klass = method.getReturnType();
         
         if (klass.equals(void.class)) {
             return getRuntime().getNil();
@@ -191,12 +288,17 @@ public class JavaMethod extends JavaCallable {
         return JavaClass.get(getRuntime(), klass);
     }
 
+    @JRubyMethod
+    public IRubyObject type_parameters() {
+        return Java.getInstance(getRuntime(), method.getTypeParameters());
+    }
+
     private IRubyObject invokeWithExceptionHandling(Method method, Object javaInvokee, Object[] arguments) {
         try {
             Object result = method.invoke(javaInvokee, arguments);
-            return JavaObject.wrap(getRuntime(), result);
+            return returnConverter.convert(getRuntime(), result);
         } catch (IllegalArgumentException iae) {
-            throw getRuntime().newTypeError("expected " + argument_types().inspect() + "; got: "
+            throw getRuntime().newTypeError("for method " + method.getName() + " expected " + argument_types().inspect() + "; got: "
                         + dumpArgTypes(arguments)
                         + "; error: " + iae.getMessage());
         } catch (IllegalAccessException iae) {
@@ -209,7 +311,7 @@ public class JavaMethod extends JavaCallable {
     }
 
     private String dumpArgTypes(Object[] arguments) {
-        StringBuffer str = new StringBuffer("[");
+        StringBuilder str = new StringBuilder("[");
         for (int i = 0; i < arguments.length; i++) {
             if (i > 0) {
                 str.append(",");
@@ -224,17 +326,37 @@ public class JavaMethod extends JavaCallable {
         return str.toString();
     }
 
-    private void convertArguments(Object[] arguments, Object[] args, int from) {
-        Class[] parameterTypes = parameterTypes();
-        for (int i = 0; i < arguments.length; i++) {
-            arguments[i] = JavaUtil.convertArgument(args[i+from], parameterTypes[i]);
+    private void convertArguments(Ruby runtime, Object[] arguments, Object[] args, int from) {
+        Class<?>[] types = parameterTypes;
+        for (int i = arguments.length; --i >= 0; ) {
+            arguments[i] = JavaUtil.convertArgument(runtime, args[i+from], types[i]);
         }
     }
 
-    protected Class[] parameterTypes() {
+    public Class<?>[] getParameterTypes() {
         return parameterTypes;
     }
 
+    public Class<?>[] getExceptionTypes() {
+        return method.getExceptionTypes();
+    }
+
+    public Type[] getGenericParameterTypes() {
+        return method.getGenericParameterTypes();
+    }
+
+    public Type[] getGenericExceptionTypes() {
+        return method.getGenericExceptionTypes();
+    }
+    
+    public Annotation[][] getParameterAnnotations() {
+        return method.getParameterAnnotations();
+    }
+
+    public boolean isVarArgs() {
+        return method.isVarArgs();
+    }
+
     protected String nameOnInspection() {
         return "#<" + getType().toString() + "/" + method.getName() + "(";
     }
@@ -242,16 +364,24 @@ public class JavaMethod extends JavaCallable {
     public RubyBoolean static_p() {
         return getRuntime().newBoolean(isStatic());
     }
+    
+    public RubyBoolean bridge_p() {
+        return getRuntime().newBoolean(method.isBridge());
+    }
 
     private boolean isStatic() {
         return Modifier.isStatic(method.getModifiers());
     }
 
-    protected int getModifiers() {
+    public int getModifiers() {
         return method.getModifiers();
     }
 
-    protected AccessibleObject accesibleObject() {
+    public String toGenericString() {
+        return method.toGenericString();
+    }
+
+    protected AccessibleObject accessibleObject() {
         return method;
     }
 }
