@@ -1,4 +1,5 @@
-/***** BEGIN LICENSE BLOCK *****
+/*
+ ***** BEGIN LICENSE BLOCK *****
  * Version: CPL 1.0/GPL 2.0/LGPL 2.1
  *
  * The contents of this file are subject to the Common Public
@@ -12,6 +13,7 @@
  * rights and limitations under the License.
  *
  * Copyright (C) 2006 Kresten Krab Thorup <krab@gnu.org>
+ * Copyright (C) 2007 William N Dortch <bill.dortch@gmail.com>
  * 
  * Alternatively, the contents of this file may be used under the terms of
  * either of the GNU General Public License Version 2 or later (the "GPL"),
@@ -39,7 +41,6 @@ import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.HashMap;
 import java.util.HashSet;
-import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
@@ -52,11 +53,13 @@ import org.jruby.RubyModule;
 import org.jruby.RubyObject;
 import org.jruby.RubyNil;
 import org.jruby.RubyString;
+import org.jruby.anno.JRubyClass;
+import org.jruby.anno.JRubyMethod;
 import org.jruby.exceptions.RaiseException;
+import org.jruby.internal.runtime.methods.DynamicMethod;
 import org.jruby.javasupport.JavaClass;
 import org.jruby.javasupport.JavaObject;
 import org.jruby.javasupport.JavaUtil;
-import org.jruby.runtime.CallbackFactory;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.builtin.IRubyObject;
 
@@ -77,17 +80,28 @@ import org.jruby.runtime.builtin.IRubyObject;
  * 
  */
 public class JavaProxyClass extends JavaProxyReflectionObject {
-    static ThreadLocal runtimeTLS = new ThreadLocal();
+    static ThreadLocal<Ruby> runtimeTLS = new ThreadLocal<Ruby>();
     private final Class proxyClass;
-    private ArrayList methods = new ArrayList();
-    private HashMap methodMap = new HashMap();
+    private final ArrayList<JavaProxyMethod> methods = new ArrayList<JavaProxyMethod>();
+    private final HashMap<String, List<JavaProxyMethod>> methodMap = new HashMap<String, List<JavaProxyMethod>>();
+    private final RubyArray constructors;
 
     /* package scope */
     JavaProxyClass(Class proxyClass) {
         super(getThreadLocalRuntime(), 
-                (RubyClass) getThreadLocalRuntime().getModule("Java").getClass("JavaProxyClass"));
+                (RubyClass) getThreadLocalRuntime().fastGetModule("Java").fastGetClass("JavaProxyClass"));
         
         this.proxyClass = proxyClass;
+        this.constructors = buildRubyArray(getConstructors());
+    }
+
+    public boolean equals(Object other) {
+        return other instanceof JavaProxyClass &&
+            this.proxyClass == ((JavaProxyClass)other).proxyClass;
+    }
+    
+    public int hashCode() {
+        return proxyClass.hashCode();
     }
 
     public Object getValue() {
@@ -95,17 +109,17 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
     }
 
     private static Ruby getThreadLocalRuntime() {
-        return (Ruby) runtimeTLS.get();
+        return runtimeTLS.get();
     }
 
     public static JavaProxyClass getProxyClass(Ruby runtime, Class superClass,
             Class[] interfaces, Set names) throws InvocationTargetException {
-        Object save = runtimeTLS.get();
+        Ruby save = runtimeTLS.get();
         runtimeTLS.set(runtime);
         try {
-            ClassLoader loader = runtime.getJavaSupport().getJavaClassLoader();
+            ClassLoader loader = runtime.getJRubyClassLoader();
 
-            return JavaProxyClassFactory.newProxyClass(loader, null, superClass, interfaces, names);
+            return JavaProxyClassFactory.newProxyClass(runtime, loader, null, superClass, interfaces, names);
         } finally {
             runtimeTLS.set(save);
         }
@@ -166,11 +180,11 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
     }
 
     public JavaProxyMethod[] getMethods() {
-        return (JavaProxyMethod[]) methods.toArray(new JavaProxyMethod[methods.size()]);
+        return methods.toArray(new JavaProxyMethod[methods.size()]);
     }
 
     public JavaProxyMethod getMethod(String name, Class[] parameterTypes) {
-        List methods = (List)methodMap.get(name);
+        List<JavaProxyMethod> methods = methodMap.get(name);
         if (methods != null) {
             for (int i = methods.size(); --i >= 0; ) {
                 ProxyMethodImpl jpm = (ProxyMethodImpl) methods.get(i);
@@ -184,7 +198,13 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
     Class getProxyClass() {
         return proxyClass;
     }
+    
+    @Override
+    public Class getJavaClass() {
+        return proxyClass;
+    }
 
+    @JRubyClass(name="JavaProxy::JavaProxyMethod")
     public static class ProxyMethodImpl extends JavaProxyReflectionObject
             implements JavaProxyMethod {
         private final Method m;
@@ -198,14 +218,23 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
 
         public ProxyMethodImpl(Ruby runtime, JavaProxyClass clazz, Method m,
                 Method sm) {
-            super(runtime, runtime.getModule("Java")
-                    .getClass("JavaProxyMethod"));
+            super(runtime, runtime.getJavaSupport().getJavaModule()
+                    .fastGetClass("JavaProxyMethod"));
             this.m = m;
             this.parameterTypes = m.getParameterTypes();
             this.sm = sm;
             this.clazz = clazz;
         }
 
+        public boolean equals(Object other) {
+            return other instanceof ProxyMethodImpl &&
+                this.m == ((ProxyMethodImpl)other).m;
+        }
+        
+        public int hashCode() {
+            return m.hashCode();
+        }
+
         public Method getMethod() {
             return m;
         }
@@ -222,13 +251,17 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
             return m.getName();
         }
 
-        public Class[] getExceptionTypes() {
+        public Class<?>[] getExceptionTypes() {
             return m.getExceptionTypes();
         }
 
-        public Class[] getParameterTypes() {
+        public Class<?>[] getParameterTypes() {
             return parameterTypes;
         }
+        
+        public boolean isVarArgs() {
+            return m.isVarArgs();
+        }
 
         public Object getState() {
             return state;
@@ -281,18 +314,9 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
             RubyClass result = javaProxyModule.defineClassUnder("JavaProxyMethod", 
                     runtime.getObject(), ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
 
-            CallbackFactory callbackFactory = 
-                runtime.callbackFactory(JavaProxyClass.ProxyMethodImpl.class);
-
             JavaProxyReflectionObject.registerRubyMethods(runtime, result);
 
-            result.defineFastMethod("argument_types", callbackFactory.getFastMethod("argument_types"));
-            result.defineFastMethod("declaring_class", callbackFactory.getFastMethod("getDeclaringClass"));
-            result.defineFastMethod("super?", callbackFactory.getFastMethod("super_p"));
-            result.defineFastMethod("arity", callbackFactory.getFastMethod("arity"));
-            result.defineFastMethod("name", callbackFactory.getFastMethod("name"));
-            result.defineFastMethod("inspect", callbackFactory.getFastMethod("inspect"));
-            result.defineFastMethod("invoke", callbackFactory.getFastOptMethod("do_invoke"));
+            result.defineAnnotatedMethods(ProxyMethodImpl.class);
 
             return result;
         }
@@ -301,18 +325,22 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
             return getRuntime().newString(getName());
         }
 
+        @JRubyMethod(name = "declaring_class")
         public JavaProxyClass getDeclaringClass() {
             return clazz;
         }
 
+        @JRubyMethod
         public RubyArray argument_types() {
             return buildRubyArray(getParameterTypes());
         }
 
+        @JRubyMethod(name = "super?")
         public IRubyObject super_p() {
             return hasSuperImplementation() ? getRuntime().getTrue() : getRuntime().getFalse();
         }
 
+        @JRubyMethod
         public RubyFixnum arity() {
             return getRuntime().newFixnum(getArity());
         }
@@ -321,8 +349,9 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
             return getDeclaringClass().nameOnInspection() + "/" + getName();
         }
 
+        @JRubyMethod
         public IRubyObject inspect() {
-            StringBuffer result = new StringBuffer();
+            StringBuilder result = new StringBuilder();
             result.append(nameOnInspection());
             result.append("(");
             Class[] parameterTypes = getParameterTypes();
@@ -336,6 +365,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
             return getRuntime().newString(result.toString());
         }
 
+        @JRubyMethod(name = "invoke", rest = true)
         public IRubyObject do_invoke(IRubyObject[] nargs) {
             if (nargs.length != 1 + getArity()) {
                 throw getRuntime().newArgumentError(nargs.length, 1 + getArity());
@@ -364,7 +394,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
                 throw getRuntime().newTypeError("illegal access on '" + sm.getName() + "': " + 
                         iae.getMessage());
             } catch (InvocationTargetException ite) {
-                ite.getTargetException().printStackTrace();
+                if (getRuntime().getDebug().isTrue()) ite.getTargetException().printStackTrace();
                 getRuntime().getJavaSupport().handleNativeException(ite.getTargetException());
                 // This point is only reached if there was an exception handler
                 // installed.
@@ -390,9 +420,9 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
 
             JavaProxyMethod jpm = new ProxyMethodImpl(getRuntime(), this, m, sm);
             methods.add(jpm);
-            List methodsWithName = (List)methodMap.get(name);
+            List<JavaProxyMethod> methodsWithName = methodMap.get(name);
             if (methodsWithName == null) {
-                methodsWithName = new ArrayList(2);
+                methodsWithName = new ArrayList<JavaProxyMethod>(2);
                 methodMap.put(name,methodsWithName);
             }
             methodsWithName.add(jpm);
@@ -409,7 +439,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
 
     private static Class[] parse(final ClassLoader loader, String desc)
             throws ClassNotFoundException {
-        List al = new ArrayList();
+        List<Class> al = new ArrayList<Class>();
         int idx = 1;
         while (desc.charAt(idx) != ')') {
 
@@ -427,13 +457,9 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
                 final String name = desc.substring(idx + 1, semi);
                 idx = semi;
                 try {
-                    type = (Class) AccessController
-                            .doPrivileged(new PrivilegedExceptionAction() {
-                                public Object run()
-                                        throws ClassNotFoundException {
-                                    return Class.forName(
-                                            name.replace('/', '.'), false,
-                                            loader);
+                    type = AccessController.doPrivileged(new PrivilegedExceptionAction<Class>() {
+                                public Class run() throws ClassNotFoundException {
+                                    return Class.forName(name.replace('/', '.'), false, loader);
                                 }
                             });
                 } catch (PrivilegedActionException e) {
@@ -472,24 +498,21 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
     public static RubyClass createJavaProxyClassClass(Ruby runtime, RubyModule javaModule) {
         RubyClass result = javaModule.defineClassUnder("JavaProxyClass",
                 runtime.getObject(),ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
-        CallbackFactory callbackFactory = runtime.callbackFactory(JavaProxyClass.class);
 
         JavaProxyReflectionObject.registerRubyMethods(runtime, result);
 
-        result.defineFastMethod("constructors", callbackFactory.getFastMethod("constructors"));
-        result.defineFastMethod("superclass", callbackFactory.getFastMethod("superclass"));
-        result.defineFastMethod("interfaces", callbackFactory.getFastMethod("interfaces"));
-        result.defineFastMethod("methods", callbackFactory.getFastMethod("methods"));
-
-        result.getMetaClass().defineFastMethod("get", 
-                callbackFactory.getFastSingletonMethod("get", JavaClass.class));
-        result.getMetaClass().defineFastMethod("get_with_class", 
-                callbackFactory.getFastSingletonMethod("get_with_class", RubyClass.class));
+        result.defineAnnotatedMethods(JavaProxyClass.class);
 
         return result;
     }
 
-    public static RubyObject get(IRubyObject recv, JavaClass type) {
+    @JRubyMethod(meta = true)
+    public static RubyObject get(IRubyObject recv, IRubyObject obj) {
+        if (!(obj instanceof JavaClass)) {
+            throw recv.getRuntime().newTypeError(obj, recv.getRuntime().getJavaSupport().getJavaClassClass());
+        }
+        JavaClass type = (JavaClass)obj;
+        
         try {
             return getProxyClass(recv.getRuntime(), (Class) type.getValue(), new Class[0]);
         } catch (Error e) {
@@ -503,7 +526,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         }
     }
     
-    private static final HashSet EXCLUDE_MODULES = new HashSet();
+    private static final HashSet<String> EXCLUDE_MODULES = new HashSet<String>();
     static {
         EXCLUDE_MODULES.add("Kernel");
         EXCLUDE_MODULES.add("Java");
@@ -511,7 +534,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         EXCLUDE_MODULES.add("Enumerable");
     }
 
-    private static final HashSet EXCLUDE_METHODS = new HashSet();
+    private static final HashSet<String> EXCLUDE_METHODS = new HashSet<String>();
     static {
         EXCLUDE_METHODS.add("class");
         EXCLUDE_METHODS.add("finalize");
@@ -522,9 +545,16 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         EXCLUDE_METHODS.add("__jsend!");
     }
 
-    public static RubyObject get_with_class(IRubyObject recv, RubyClass clazz) {
+    @JRubyMethod(meta = true)
+    public static RubyObject get_with_class(IRubyObject recv, IRubyObject obj) {
         Ruby runtime = recv.getRuntime();
         
+        if (!(obj instanceof RubyClass)) {
+            throw runtime.newTypeError(obj, runtime.getClassClass());
+        }
+        
+        RubyClass clazz = (RubyClass)obj;
+        
         // Let's only generate methods for those the user may actually 
         // intend to override.  That includes any defined in the current
         // class, and any ancestors that are also JavaProxyClasses (but none
@@ -534,26 +564,25 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         // TODO: may want to exclude other common mixins?
 
         JavaClass javaClass = null;
-        Set names = new HashSet(); // need names ordered for key generation later
-        List interfaceList = new ArrayList();
+        Set<String> names = new HashSet<String>(); // need names ordered for key generation later
+        List<Class<?>> interfaceList = new ArrayList<Class<?>>();
 
-        List ancestors = clazz.getAncestorList();
+        List<IRubyObject> ancestors = clazz.getAncestorList();
         boolean skipRemainingClasses = false;
-        for (Iterator iter = ancestors.iterator(); iter.hasNext(); ) {
-            RubyModule ancestor = (RubyModule)iter.next();
+        for (IRubyObject ancestorObject: ancestors) {
+            RubyModule ancestor = (RubyModule) ancestorObject;
             if (ancestor instanceof RubyClass) {
                 if (skipRemainingClasses) continue;
-                Map vars = ancestor.getInstanceVariables();
                 // we only collect methods and interfaces for 
                 // user-defined proxy classes.
-                if (!vars.containsKey("@java_proxy_class")) {
+                if (!ancestor.getInstanceVariables().fastHasInstanceVariable("@java_proxy_class")) {
                     skipRemainingClasses = true;
                     continue;
                 }
 
                 // get JavaClass if this is the new proxy class; verify it
                 // matches if this is a superclass proxy.
-                IRubyObject var = (IRubyObject)vars.get("@java_class");
+                IRubyObject var = ancestor.getInstanceVariables().fastGetInstanceVariable("@java_class");
                 if (var == null) {
                     throw runtime.newTypeError(
                             "no java_class defined for proxy (or ancestor): " + ancestor);
@@ -571,7 +600,7 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
                             " (" + var + ")");
                 }
                 // get any included interfaces
-                var = (IRubyObject)vars.get("@java_interfaces");
+                var = ancestor.getInstanceVariables().fastGetInstanceVariable("@java_interfaces");
                 if (var != null && !(var instanceof RubyNil)) {
                     if (!(var instanceof RubyArray)) {
                         throw runtime.newTypeError(
@@ -601,24 +630,25 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
                 // set this class's method names in var @__java_ovrd_methods if this
                 // is the new class; otherwise, get method names from there if this is
                 // a proxy superclass.
-                var = (IRubyObject)vars.get("@__java_ovrd_methods");
+                
+                // FIXME: shouldn't need @__java_ovrd_methods, just query locally defined methods.
+                
+                var = ancestor.getInstanceVariables().fastGetInstanceVariable("@__java_ovrd_methods");
                 if (var == null) {
                     // lock in the overridden methods for the new class, and any as-yet
                     // uninstantiated ancestor class.
-                    Map methods;
+                    Map<String, DynamicMethod> methods;
                     RubyArray methodNames;
                     synchronized(methods = ancestor.getMethods()) {
                         methodNames = RubyArray.newArrayLight(runtime,methods.size());
-                        for (Iterator meths = methods.keySet().iterator(); meths.hasNext(); ) {
-                            String methodName = (String)meths.next();
+                        for (String methodName: methods.keySet()) {
                             if (!EXCLUDE_METHODS.contains(methodName)) {
                                 names.add(methodName);
-                                methodNames.add(runtime.newString(methodName));
+                                methodNames.append(runtime.newString(methodName));
                             }
                         }
                     }
-                    // TODO: OK to just do a put here?
-                    ancestor.setInstanceVariable("@__java_ovrd_methods",methodNames);
+                    ancestor.fastSetInstanceVariable("@__java_ovrd_methods",methodNames);
                 } else {
                     if (!(var instanceof RubyArray)) {
                         throw runtime.newTypeError(
@@ -634,14 +664,13 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
                                     "invalid method name defined for proxy (or ancestor): " +
                                     ancestor + ": " + methodName);
                         }
-                        names.add(methodName.asSymbol());
+                        names.add(methodName.asJavaString());
                     }
                 }
             } else if (!EXCLUDE_MODULES.contains(ancestor.getName())) {
-                Map methods;
+                Map<String, DynamicMethod> methods;
                 synchronized(methods = ancestor.getMethods()) {
-                    for (Iterator meths = methods.keySet().iterator(); meths.hasNext(); ) {
-                        String methodName = (String)meths.next();
+                    for (String methodName: methods.keySet()) {
                         if (!EXCLUDE_METHODS.contains(methodName)) {
                             names.add(methodName);
                         }
@@ -655,9 +684,9 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         }
         
         int interfaceCount = interfaceList.size();
-        Class[] interfaces = new Class[interfaceCount];
+        Class<?>[] interfaces = new Class<?>[interfaceCount];
         for (int i = interfaceCount; --i >= 0; ) {
-            interfaces[i] = (Class)interfaceList.get(i);
+            interfaces[i] = interfaceList.get(i);
         }
        
         try {
@@ -675,26 +704,30 @@ public class JavaProxyClass extends JavaProxyReflectionObject {
         }
     }
 
+    @JRubyMethod
     public RubyObject superclass() {
         return JavaClass.get(getRuntime(), getSuperclass());
     }
 
+    @JRubyMethod
     public RubyArray methods() {
         return buildRubyArray(getMethods());
     }
 
+    @JRubyMethod
     public RubyArray interfaces() {
         return buildRubyArray(getInterfaces());
     }
 
+    @JRubyMethod
     public RubyArray constructors() {
-        return buildRubyArray(getConstructors());
+        return this.constructors;
     }
 
     public static void createJavaProxyModule(Ruby runtime) {
         // TODO Auto-generated method stub
 
-        RubyModule javaProxyModule = runtime.getModule("Java");
+        RubyModule javaProxyModule = runtime.getJavaSupport().getJavaModule();
         JavaProxyClass.createJavaProxyClassClass(runtime, javaProxyModule);
         ProxyMethodImpl.createJavaProxyMethodClass(runtime, javaProxyModule);
         JavaProxyConstructor.createJavaProxyConstructorClass(runtime, javaProxyModule);
