@@ -38,38 +38,52 @@ import org.jruby.RubyFixnum;
 import org.jruby.RubyModule;
 import org.jruby.RubyObject;
 import org.jruby.RubyProc;
+import org.jruby.anno.JRubyMethod;
 import org.jruby.exceptions.RaiseException;
+import org.jruby.internal.runtime.methods.DynamicMethod;
 import org.jruby.javasupport.Java;
 import org.jruby.javasupport.JavaObject;
 import org.jruby.javasupport.JavaUtil;
+import org.jruby.javasupport.ParameterTypes;
+import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
-import org.jruby.runtime.CallbackFactory;
+import org.jruby.runtime.CallType;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.builtin.IRubyObject;
 
-public class JavaProxyConstructor extends JavaProxyReflectionObject {
+public class JavaProxyConstructor extends JavaProxyReflectionObject implements ParameterTypes {
 
-    private final Constructor proxyConstructor;
-    private final Class[] parameterTypes;
+    private final Constructor<?> proxyConstructor;
+    private final Class<?>[] apparentParameterTypes;
 
     private final JavaProxyClass declaringProxyClass;
 
     JavaProxyConstructor(Ruby runtime, JavaProxyClass pClass,
-            Constructor constructor) {
-        super(runtime, runtime.getModule("Java").getClass(
+            Constructor<?> constructor) {
+        super(runtime, runtime.getJavaSupport().getJavaModule().fastGetClass(
                 "JavaProxyConstructor"));
         this.declaringProxyClass = pClass;
         this.proxyConstructor = constructor;
-        this.parameterTypes = proxyConstructor.getParameterTypes();
+        Class<?>[] parameterTypes = constructor.getParameterTypes();
+        int len = parameterTypes.length - 1;
+        this.apparentParameterTypes = new Class<?>[len];
+        System.arraycopy(parameterTypes, 0, apparentParameterTypes, 0, len);
     }
 
-    public Class[] getParameterTypes() {
-        Class[] result = new Class[parameterTypes.length - 1];
-        System.arraycopy(parameterTypes, 0, result, 0, result.length);
-        return result;
+    public Class<?>[] getParameterTypes() {
+        return apparentParameterTypes;
     }
 
+    public Class<?>[] getExceptionTypes() {
+        return proxyConstructor.getExceptionTypes();
+    }
+    
+    public boolean isVarArgs() {
+        return proxyConstructor.isVarArgs();
+    }
+
+    @JRubyMethod(name = "declaring_class")
     public JavaProxyClass getDeclaringClass() {
         return declaringProxyClass;
     }
@@ -77,7 +91,7 @@ public class JavaProxyConstructor extends JavaProxyReflectionObject {
     public Object newInstance(Object[] args, JavaProxyInvocationHandler handler)
             throws IllegalArgumentException, InstantiationException,
             IllegalAccessException, InvocationTargetException {
-        if (args.length + 1 != parameterTypes.length) {
+        if (args.length != apparentParameterTypes.length) {
             throw new IllegalArgumentException("wrong number of parameters");
         }
 
@@ -90,41 +104,39 @@ public class JavaProxyConstructor extends JavaProxyReflectionObject {
 
     public static RubyClass createJavaProxyConstructorClass(Ruby runtime,
             RubyModule javaProxyModule) {
-        RubyClass result = javaProxyModule.defineClassUnder(
-                                                            "JavaProxyConstructor", runtime.getObject(), ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
-
-        CallbackFactory callbackFactory = runtime
-                .callbackFactory(JavaProxyConstructor.class);
+        RubyClass result = javaProxyModule.defineClassUnder("JavaProxyConstructor",
+                runtime.getObject(), ObjectAllocator.NOT_ALLOCATABLE_ALLOCATOR);
 
         JavaProxyReflectionObject.registerRubyMethods(runtime, result);
 
-        result.defineFastMethod("argument_types", callbackFactory
-                .getFastMethod("argument_types"));
-
-        result.defineFastMethod("declaring_class", callbackFactory
-                .getFastMethod("getDeclaringClass"));
-
-        result.defineMethod("new_instance", callbackFactory
-                .getOptMethod("new_instance"));
-
-        result.defineFastMethod("arity", callbackFactory.getFastMethod("arity"));
+        result.defineAnnotatedMethods(JavaProxyConstructor.class);
 
         return result;
 
     }
 
+    @JRubyMethod
     public RubyFixnum arity() {
         return getRuntime().newFixnum(getParameterTypes().length);
     }
+    
+    public boolean equals(Object other) {
+        return other instanceof JavaProxyConstructor &&
+            this.proxyConstructor == ((JavaProxyConstructor)other).proxyConstructor;
+    }
+    
+    public int hashCode() {
+        return proxyConstructor.hashCode();
+    }
 
     protected String nameOnInspection() {
         return getDeclaringClass().nameOnInspection();
     }
 
     public IRubyObject inspect() {
-        StringBuffer result = new StringBuffer();
+        StringBuilder result = new StringBuilder();
         result.append(nameOnInspection());
-        Class[] parameterTypes = getParameterTypes();
+        Class<?>[] parameterTypes = getParameterTypes();
         for (int i = 0; i < parameterTypes.length; i++) {
             result.append(parameterTypes[i].getName());
             if (i < parameterTypes.length - 1) {
@@ -135,10 +147,98 @@ public class JavaProxyConstructor extends JavaProxyReflectionObject {
         return getRuntime().newString(result.toString());
     }
 
+    @JRubyMethod
     public RubyArray argument_types() {
         return buildRubyArray(getParameterTypes());
     }
+    
+    @JRubyMethod(frame = true, rest = true)
+    public RubyObject new_instance2(IRubyObject[] args, Block unusedBlock) {
+        Arity.checkArgumentCount(getRuntime(), args, 2, 2);
 
+        final IRubyObject self = args[0];
+        final Ruby runtime = self.getRuntime();
+        final RubyModule javaUtilities = runtime.getJavaSupport().getJavaUtilitiesModule();
+        RubyArray constructor_args = (RubyArray) args[1];
+        Class<?>[] parameterTypes = getParameterTypes();
+        int count = (int) constructor_args.length().getLongValue();
+        Object[] converted = new Object[count];
+        
+        for (int i = 0; i < count; i++) {
+            // TODO: call ruby method
+            IRubyObject ith = constructor_args.aref(getRuntime().newFixnum(i));
+            converted[i] = JavaUtil.convertArgument(getRuntime(), Java.ruby_to_java(this, ith, Block.NULL_BLOCK), parameterTypes[i]);
+        }
+
+        JavaProxyInvocationHandler handler = new JavaProxyInvocationHandler() {
+            public Object invoke(Object proxy, JavaProxyMethod m, Object[] nargs) throws Throwable {
+                String name = m.getName();
+                DynamicMethod method = self.getMetaClass().searchMethod(name);
+                int v = method.getArity().getValue();
+                IRubyObject[] newArgs = new IRubyObject[nargs.length];
+                for (int i = nargs.length; --i >= 0; ) {
+                    newArgs[i] = Java.java_to_ruby(
+                            javaUtilities,
+                            JavaObject.wrap(runtime, nargs[i]),
+                            Block.NULL_BLOCK);
+                }
+                
+                if (v < 0 || v == (newArgs.length)) {
+                    return JavaUtil.convertRubyToJava(RuntimeHelpers.invoke(runtime.getCurrentContext(), self, name, newArgs), m.getReturnType());
+                } else {
+                    RubyClass superClass = self.getMetaClass().getSuperClass();
+                    return JavaUtil.convertRubyToJava(RuntimeHelpers.invokeAs(runtime.getCurrentContext(), superClass, self, name, newArgs, Block.NULL_BLOCK), m.getReturnType());
+                }
+            }
+        };
+
+        try {
+            return JavaObject.wrap(getRuntime(), newInstance(converted, handler));
+        } catch (Exception e) {
+            RaiseException ex = getRuntime().newArgumentError(
+                    "Constructor invocation failed: " + e.getMessage());
+            ex.initCause(e);
+            throw ex;
+        }
+    }
+    
+    public JavaObject newInstance(final IRubyObject self, Object[] args) {
+        final Ruby runtime = self.getRuntime();
+        final RubyModule javaUtilities = runtime.getJavaSupport().getJavaUtilitiesModule();
+
+        JavaProxyInvocationHandler handler = new JavaProxyInvocationHandler() {
+            public Object invoke(Object proxy, JavaProxyMethod m, Object[] nargs) throws Throwable {
+                String name = m.getName();
+                DynamicMethod method = self.getMetaClass().searchMethod(name);
+                int v = method.getArity().getValue();
+                IRubyObject[] newArgs = new IRubyObject[nargs.length];
+                for (int i = nargs.length; --i >= 0; ) {
+                    newArgs[i] = Java.java_to_ruby(
+                            javaUtilities,
+                            JavaObject.wrap(runtime, nargs[i]),
+                            Block.NULL_BLOCK);
+                }
+                
+                if (v < 0 || v == (newArgs.length)) {
+                    return JavaUtil.convertRubyToJava(RuntimeHelpers.invoke(runtime.getCurrentContext(), self, name, newArgs), m.getReturnType());
+                } else {
+                    RubyClass superClass = self.getMetaClass().getSuperClass();
+                    return JavaUtil.convertRubyToJava(RuntimeHelpers.invokeAs(runtime.getCurrentContext(), superClass, self, name, newArgs, Block.NULL_BLOCK), m.getReturnType());
+                }
+            }
+        };
+
+        try {
+            return JavaObject.wrap(getRuntime(), newInstance(args, handler));
+        } catch (Exception e) {
+            RaiseException ex = getRuntime().newArgumentError(
+                    "Constructor invocation failed: " + e.getMessage());
+            ex.initCause(e);
+            throw ex;
+        }
+    }
+
+    @JRubyMethod(required = 1, optional = 1, frame = true)
     public RubyObject new_instance(IRubyObject[] args, Block block) {
         int size = Arity.checkArgumentCount(getRuntime(), args, 1, 2) - 1;
         final RubyProc proc;
@@ -153,14 +253,14 @@ public class JavaProxyConstructor extends JavaProxyReflectionObject {
         }
 
         RubyArray constructor_args = (RubyArray) args[0];
-        Class[] parameterTypes = getParameterTypes();
+        Class<?>[] parameterTypes = getParameterTypes();
 
         int count = (int) constructor_args.length().getLongValue();
         Object[] converted = new Object[count];
         for (int i = 0; i < count; i++) {
             // TODO: call ruby method
-            IRubyObject ith = constructor_args.aref(new IRubyObject[] { getRuntime().newFixnum(i) });
-            converted[i] = JavaUtil.convertArgument(Java.ruby_to_java(this, ith, Block.NULL_BLOCK), parameterTypes[i]);
+            IRubyObject ith = constructor_args.aref(getRuntime().newFixnum(i));
+            converted[i] = JavaUtil.convertArgument(getRuntime(), Java.ruby_to_java(this, ith, Block.NULL_BLOCK), parameterTypes[i]);
         }
 
         final IRubyObject recv = this;
@@ -177,7 +277,7 @@ public class JavaProxyConstructor extends JavaProxyReflectionObject {
                     rubyArgs[i + 2] = JavaUtil.convertJavaToRuby(getRuntime(),
                             nargs[i]);
                 }
-                IRubyObject call_result = proc.call(rubyArgs);
+                IRubyObject call_result = proc.call(getRuntime().getCurrentContext(), rubyArgs);
                 Object converted_result = JavaUtil.convertRubyToJava(
                         call_result, method.getReturnType());
                 return converted_result;
