@@ -1,4 +1,5 @@
-/***** BEGIN LICENSE BLOCK *****
+/*
+ ***** BEGIN LICENSE BLOCK *****
  * Version: CPL 1.0/GPL 2.0/LGPL 2.1
  *
  * The contents of this file are subject to the Common Public
@@ -30,46 +31,84 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.parser;
 
-import java.io.Reader;
-import java.io.StringReader;
+import java.io.ByteArrayInputStream;
+import java.io.InputStream;
 
 import org.jruby.Ruby;
+import org.jruby.RubyArray;
 import org.jruby.RubyFile;
+import org.jruby.RubyHash;
+import org.jruby.RubyString;
 import org.jruby.ast.Node;
+import org.jruby.common.NullWarnings;
+//import org.jruby.lexer.yacc.ByteListLexerSource;
 import org.jruby.lexer.yacc.LexerSource;
 import org.jruby.lexer.yacc.SyntaxException;
 import org.jruby.runtime.DynamicScope;
 import org.jruby.runtime.ThreadContext;
-import org.jruby.util.collections.SinglyLinkedList;
+import org.jruby.runtime.builtin.IRubyObject;
+import org.jruby.util.ByteList;
 
 /**
  * Serves as a simple facade for all the parsing magic.
  */
 public class Parser {
     private final Ruby runtime;
+    private volatile long totalTime;
+    private volatile int totalBytes;
 
     public Parser(Ruby runtime) {
         this.runtime = runtime;
     }
 
-    public Node parse(String file, String content, DynamicScope blockScope, int lineNumber, 
-            boolean extraPositionInformation) {
-        return parse(file, new StringReader(content), blockScope, lineNumber, extraPositionInformation);
-    }
-    
-    public Node parse(String file, String content, DynamicScope blockScope, int lineNumber) {
-        return parse(file, new StringReader(content), blockScope, lineNumber, false);
-    }
-    
-    public Node parse(String file, Reader content, DynamicScope blockScope, int lineNumber) {
-        return parse(file, content, blockScope, lineNumber, false);
+    public long getTotalTime() {
+        return totalTime;
     }
 
-    public Node parse(String file, Reader content, DynamicScope blockScope, int lineNumber, 
-            boolean extraPositionInformation) {
-        RubyParserConfiguration configuration = new RubyParserConfiguration(); 
-        SinglyLinkedList cref = runtime.getObject().getCRef();
-        ThreadContext tc = runtime.getCurrentContext();
+    public int getTotalBytes() {
+        return totalBytes;
+    }
+    
+    public Node parseRewriter(String file, InputStream content, 
+            ParserConfiguration configuration) throws SyntaxException {
+        long startTime = System.nanoTime();
+        DefaultRubyParser parser = RubyParserPool.getInstance().borrowParser();
+        parser.setWarnings(new NullWarnings());
+        LexerSource lexerSource = LexerSource.getSource(file, content, null, configuration);
+        
+        try {
+            return parser.parse(configuration, lexerSource).getAST();
+        } finally {
+            RubyParserPool.getInstance().returnParser(parser);
+            totalTime += System.nanoTime() - startTime;
+            totalBytes += lexerSource.getOffset();
+        }
+    }
+    
+    @SuppressWarnings("unchecked")
+    public Node parse(String file, ByteList content, DynamicScope blockScope,
+            ParserConfiguration configuration) {
+        return parse(file, new ByteArrayInputStream(content.bytes()), blockScope, configuration);
+    }
+    
+    @SuppressWarnings("unchecked")
+    public Node parse(String file, InputStream content, DynamicScope blockScope,
+            ParserConfiguration configuration) {
+        long startTime = System.nanoTime();
+        IRubyObject scriptLines = runtime.getObject().fastGetConstantAt("SCRIPT_LINES__");
+        RubyArray list = null;
+        
+        if (!configuration.isEvalParse() && scriptLines != null) {
+            if (scriptLines instanceof RubyHash) {
+                RubyString filename = runtime.newString(file);
+                ThreadContext context = runtime.getCurrentContext();
+                IRubyObject object = ((RubyHash) scriptLines).op_aref(context, filename);
+                
+                list = (RubyArray) (object instanceof RubyArray ? object : runtime.newArray()); 
+                
+                ((RubyHash) scriptLines).op_aset(context, filename, list);
+            }
+        }
 
         // We only need to pass in current scope if we are evaluating as a block (which
         // is only done for evals).  We need to pass this in so that we can appropriately scope
@@ -80,29 +119,26 @@ public class Parser {
         
         DefaultRubyParser parser = null;
         RubyParserResult result = null;
+        parser = RubyParserPool.getInstance().borrowParser();
+        parser.setWarnings(runtime.getWarnings());
+        LexerSource lexerSource = LexerSource.getSource(file, content, list, configuration);
         try {
-            parser = RubyParserPool.getInstance().borrowParser();
-            parser.setWarnings(runtime.getWarnings());
-            tc.setCRef(cref);
-            LexerSource lexerSource = 
-                LexerSource.getSource(file, content, lineNumber, extraPositionInformation);
             result = parser.parse(configuration, lexerSource);
-            if (result.isEndSeen()) {
-                org.jruby.runtime.builtin.IRubyObject verbose = runtime.getVerbose();
+            if (result.getEndOffset() >= 0) {
+                IRubyObject verbose = runtime.getVerbose();
                 runtime.setVerbose(runtime.getNil());
             	runtime.defineGlobalConstant("DATA", new RubyFile(runtime, file, content));
                 runtime.setVerbose(verbose);
-            	result.setEndSeen(false);
+            	result.setEndOffset(-1);
             }
         } catch (SyntaxException e) {
-            StringBuffer buffer = new StringBuffer(100);
+            StringBuilder buffer = new StringBuilder(100);
             buffer.append(e.getPosition().getFile()).append(':');
-            buffer.append(e.getPosition().getEndLine()).append(": ");
+            buffer.append(e.getPosition().getEndLine() + 1).append(": ");
             buffer.append(e.getMessage());
             throw runtime.newSyntaxError(buffer.toString());
         } finally {
             RubyParserPool.getInstance().returnParser(parser);
-            tc.unsetCRef();
         }
         
         // If variables were added then we may need to grow the dynamic scope to match the static
@@ -114,8 +150,11 @@ public class Parser {
             result.getScope().growIfNeeded();
         }
 
-        // FIXME: We should move this into ParserSupport.addRootNode since actual parser should do this.
-        result.addAppendBeginNodes();
-        return result.getAST();
+        Node ast = result.getAST();
+        
+        totalTime += System.nanoTime() - startTime;
+        totalBytes += lexerSource.getOffset();
+
+        return ast;
     }
 }
