@@ -30,7 +30,6 @@ package org.jruby.runtime;
 import org.jruby.RubyModule;
 import org.jruby.ast.IterNode;
 import org.jruby.runtime.builtin.IRubyObject;
-import org.jruby.util.collections.SinglyLinkedList;
 
 /**
  * Represents the live state of a for or END construct in Ruby.  This is different from an
@@ -39,31 +38,33 @@ import org.jruby.util.collections.SinglyLinkedList;
  * stuff.  Also because of this we do not need to clone the block since it state does not change.
  * 
  */
-public class SharedScopeBlock extends Block {
-    private SharedScopeBlock(IterNode iterNode, IRubyObject self, Frame frame,
-            SinglyLinkedList cref, Visibility visibility, RubyModule klass,
-            DynamicScope dynamicScope) {
-        super(iterNode, self, frame, cref, visibility, klass, dynamicScope);
+public class SharedScopeBlock extends InterpretedBlock {
+    protected SharedScopeBlock(IterNode iterNode) {
+        super(iterNode, asArgumentType(getArgumentTypeWackyHack(iterNode)));
     }
     
-    public static Block createSharedScopeBlock(ThreadContext context, IterNode iterNode, DynamicScope dynamicScope, IRubyObject self) {
-        return new SharedScopeBlock(iterNode, self,
-                context.getCurrentFrame(),
-                context.peekCRef(),
+    public static Block newInterpretedSharedScopeClosure(ThreadContext context, IterNode iterNode, DynamicScope dynamicScope, IRubyObject self) {
+        Binding binding = new Binding(self,
+                context.getCurrentFrame().duplicate(),
                 context.getCurrentFrame().getVisibility(),
                 context.getRubyClass(),
                 dynamicScope);
+        BlockBody body = new SharedScopeBlock(iterNode);
+
+        return new Block(body, binding);
     }
     
-    protected void pre(ThreadContext context, RubyModule klass) {
-        context.preForBlock(this, klass);
+    @Override
+    protected Frame pre(ThreadContext context, RubyModule klass, Binding binding) {
+        return context.preForBlock(binding, klass);
     }
     
-    public IRubyObject call(ThreadContext context, IRubyObject[] args, IRubyObject replacementSelf) {
-        return yield(context, context.getRuntime().newArrayNoCopy(args), null, null, true);
+    public IRubyObject call(ThreadContext context, IRubyObject[] args, IRubyObject replacementSelf, Binding binding, Block.Type type) {
+        return yield(context, context.getRuntime().newArrayNoCopy(args), null, null, true, binding, type);
     }
     
-    public Block cloneBlock() {
-        return this;
+    @Override
+    public Block cloneBlock(Binding binding) {
+        return new Block(this, binding);
     }
 }
