@@ -41,6 +41,7 @@ import org.jruby.RubyInteger;
 import org.jruby.RubyKernel;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyString;
+import org.jruby.common.IRubyWarnings.ID;
 import org.jruby.runtime.ClassIndex;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.builtin.IRubyObject;
@@ -86,13 +87,13 @@ public class Sprintf {
     
     
     private static class Args {
-        Ruby runtime;
-        Locale locale;
-        IRubyObject rubyObject;
-        List rubyArray;
-        int length;
-        int unnumbered; // last index (+1) accessed by next()
-        int numbered;   // last index (+1) accessed by get()
+        private final Ruby runtime;
+        private final Locale locale;
+        private final IRubyObject rubyObject;
+        private final List rubyArray;
+        private final int length;
+        private int unnumbered; // last index (+1) accessed by next()
+        private int numbered;   // last index (+1) accessed by get()
         
         Args(Locale locale, IRubyObject rubyObject) {
             if (rubyObject == null) {
@@ -105,6 +106,7 @@ public class Sprintf {
                 this.length = rubyArray.size();
             } else {
                 this.length = 1;
+                this.rubyArray = null;
             }
             this.runtime = rubyObject.getRuntime();
         }
@@ -123,12 +125,12 @@ public class Sprintf {
             throw runtime.newArgumentError(message);
         }
         
-        final void warn(String message) {
-            runtime.getWarnings().warn(message);
+        final void warn(ID id, String message) {
+            runtime.getWarnings().warn(id, message);
         }
         
-        final void warning(String message) {
-            runtime.getWarnings().warning(message);
+        final void warning(ID id, String message) {
+            runtime.getWarnings().warning(id, message);
         }
         
         final IRubyObject next() {
@@ -184,7 +186,7 @@ public class Sprintf {
             }
 
             // basically just forcing a TypeError here to match MRI
-            obj = obj.convertToType(obj.getRuntime().getFixnum(), MethodIndex.TO_INT, "to_int", true);
+            obj = TypeConverter.convertToType(obj, obj.getRuntime().getFixnum(), MethodIndex.TO_INT, "to_int", true);
             return (int)((RubyFixnum)obj).getLongValue();
         }
         
@@ -206,6 +208,7 @@ public class Sprintf {
     private static class Buffer {
         byte[] buf;
         int size;
+        private boolean tainted = false;
          
         Buffer() {
             buf = new byte[INITIAL_BUFFER_SIZE];
@@ -273,6 +276,7 @@ public class Sprintf {
             return new ByteList(buf,0,size);
         }
         
+        @Override
         public final String toString() {
             return new String(buf,0,size);
         }
@@ -284,6 +288,17 @@ public class Sprintf {
     public static CharSequence sprintf(Locale locale, CharSequence format, IRubyObject args) {
         return rubySprintf(format, new Args(locale,args));
     }
+    
+    // Special form of sprintf that returns a RubyString and handles
+    // tainted strings correctly.
+    public static RubyString sprintf(Ruby runtime, Locale locale, CharSequence format, IRubyObject args) {
+        Buffer b = rubySprintfToBuffer(format, new Args(locale,args));
+        RubyString s = runtime.newString(b.toByteList());
+        if (b.tainted) {
+            s.setTaint(true);
+        }
+        return s;
+    }
 
     public static CharSequence sprintf(CharSequence format, IRubyObject args) {
         return rubySprintf(format, new Args(args));
@@ -306,6 +321,10 @@ public class Sprintf {
     }
     
     private static CharSequence rubySprintf(CharSequence charFormat, Args args) {
+        return rubySprintfToBuffer(charFormat, args).toByteList();
+    }
+    
+    private static Buffer rubySprintfToBuffer(CharSequence charFormat, Args args) {
         byte[] format;
         Buffer buf = new Buffer();
         
@@ -350,7 +369,16 @@ public class Sprintf {
             for ( ; incomplete && offset < length ; ) {
                 switch(fchar = format[offset]) {
                 default:
-                    if (isPrintable(fchar)) {
+                    if (fchar == '\0' && flags == FLAG_NONE) {
+                        // MRI 1.8.6 behavior: null byte after '%'
+                        // leads to "%" string. Null byte in
+                        // other places, like "%5\0", leads to error.
+                        buf.write('%');
+                        buf.write(fchar);
+                        incomplete = false;
+                        offset++;
+                        break;
+                    } else if (isPrintable(fchar)) {
                         raiseArgumentError(args,"malformed format string - %" + (char)fchar);
                     } else {
                         raiseArgumentError(args,ERR_MALFORMED_FORMAT);
@@ -530,6 +558,9 @@ public class Sprintf {
                     }
                     ByteList bytes = arg.asString().getByteList();
                     int len = bytes.length();
+                    if (arg.isTaint()) {
+                        buf.tainted = true;
+                    }
                     if ((flags & FLAG_PRECISION) != 0 && precision < len) {
                         len = precision;
                     }
@@ -579,7 +610,11 @@ public class Sprintf {
                             arg = RubyNumeric.str2inum(arg.getRuntime(),(RubyString)arg,0,true);
                             break;
                         default:
-                            arg = arg.convertToType(arg.getRuntime().getClass("Integer"), MethodIndex.TO_I, "to_i", true);
+                            if (arg.respondsTo("to_int")) {
+                                arg = TypeConverter.convertToType(arg, arg.getRuntime().getInteger(), MethodIndex.TO_INT, "to_int", true);
+                            } else {
+                                arg = TypeConverter.convertToType(arg, arg.getRuntime().getInteger(), MethodIndex.TO_I, "to_i", true);
+                            }
                             break;
                         }
                         type = arg.getMetaClass().index;
@@ -659,7 +694,7 @@ public class Sprintf {
                         }
                     } else if (negative) {
                         if (base == 10) {
-                            warning(args,"negative number for %u specifier");
+                            warning(ID.NEGATIVE_NUMBER_FOR_U, args,"negative number for %u specifier");
                             leadChar = '.';
                             len += 2;
                         } else {
@@ -922,7 +957,7 @@ public class Sprintf {
                             decDigits = nDigits - 1;
                             // precision for G/g includes integer digits
                             precision = Math.max(0,precision - 1);
-                            
+
                             if (precision < decDigits) {
                                 int n = round(digits,nDigits,precision,precision!=0);
                                 if (n > nDigits) {
@@ -931,16 +966,52 @@ public class Sprintf {
                                 decDigits = Math.min(nDigits - 1,precision);
                             }
                             exponent += nDigits - 1;
-                            if (precision > 0) {
-                                len += 1 + precision; // n.prec
-                            } else {
-                                len += 1; // n
-                                if ((flags & FLAG_SHARP) != 0) {
-                                    len++; // will have a trailing '.'
-                                }
-                            }
                             
-                            width -= len + 5; // 5 -> e+nnn / e-nnn
+                            boolean isSharp = (flags & FLAG_SHARP) != 0;
+
+                            // deal with length/width
+			    
+                            len++; // first digit is always printed
+
+                            // MRI behavior: Be default, 2 digits
+                            // in the exponent. Use 3 digits
+                            // only when necessary.
+                            // See comment for writeExp method for more details.
+                            if (exponent > 99)
+                            	len += 5; // 5 -> e+nnn / e-nnn
+                            else
+                            	len += 4; // 4 -> e+nn / e-nn
+
+                            if (isSharp) {
+                            	// in this mode, '.' is always printed
+                            	len++;
+                            }
+
+                            if (precision > 0) {
+                            	if (!isSharp) {
+                            	    // MRI behavior: In this mode
+                            	    // trailing zeroes are removed:
+                            	    // 1.500E+05 -> 1.5E+05 
+                            	    int j = decDigits;
+                            	    for (; j >= 1; j--) {
+                            	        if (digits[j]== '0') {
+                            	            decDigits--;
+                            	        } else {
+                            	            break;
+                            	        }
+                            	    }
+
+                            	    if (decDigits > 0) {
+                            	        len += 1; // '.' is printed
+                            	        len += decDigits;
+                            	    }
+                            	} else  {
+                            	    // all precision numebers printed
+                            	    len += precision;
+                            	}
+                            }
+
+                            width -= len;
 
                             if (width > 0 && (flags & (FLAG_ZERO|FLAG_MINUS)) == 0) {
                                 buf.fill(' ',width);
@@ -953,25 +1024,28 @@ public class Sprintf {
                                 buf.fill('0',width);
                                 width = 0;
                             }
+
                             // now some data...
                             buf.write(digits[0]);
-                            if (precision > 0) {
-                                buf.write(args.getDecimalSeparator()); // '.'
-                                if (decDigits > 0) {
-                                    buf.write(digits,1,decDigits);
-                                    precision -= decDigits;
-                                }
-                            } else if ((flags & FLAG_SHARP) != 0) {
-                                buf.write(args.getDecimalSeparator());
+
+                            boolean dotToPrint = isSharp
+                                    || (precision > 0 && decDigits > 0);
+
+                            if (dotToPrint) {
+                            	buf.write(args.getDecimalSeparator()); // '.'
                             }
-                            buf.write(expChar); // E or e
-                            buf.write(exponent >= 0 ? '+' : '-');
-                            if (exponent < 0) {
-                                exponent = -exponent;
+
+                            if (precision > 0 && decDigits > 0) {
+                            	buf.write(digits, 1, decDigits);
+                            	precision -= decDigits;
                             }
-                            buf.write(exponent / 100 + '0');
-                            buf.write(exponent % 100 / 10 + '0');
-                            buf.write(exponent % 10 + '0');
+
+                            if (precision > 0 && isSharp) {
+                            	buf.fill('0', precision);
+                            }
+
+                            writeExp(buf, exponent, expChar);
+
                             if (width > 0) {
                                 buf.fill(' ', width);
                             }
@@ -1158,17 +1232,31 @@ public class Sprintf {
                             decDigits = Math.min(nDigits - 1,precision);
                         }
                         exponent += nDigits - 1;
+
+                        boolean isSharp = (flags & FLAG_SHARP) != 0;
+
+                        // deal with length/width
+
+                        len++; // first digit is always printed
+
+                        // MRI behavior: Be default, 2 digits
+                        // in the exponent. Use 3 digits
+                        // only when necessary.
+                        // See comment for writeExp method for more details.
+                        if (exponent > 99)
+                            len += 5; // 5 -> e+nnn / e-nnn
+                        else
+                            len += 4; // 4 -> e+nn / e-nn
+
                         if (precision > 0) {
-                            len += 2 + precision; // n.prec
-                        } else {
-                            len += 1; // n
-                            if ((flags & FLAG_SHARP) != 0) {
-                                len++; // will have a trailing '.'
-                            }
+                            // '.' and all precision digits printed
+                            len += 1 + precision;
+                        } else  if (isSharp) {
+                            len++;  // in this mode, '.' is always printed
                         }
-                        
-                        width -= len + 5; // 5 -> e+nnn / e-nnn
-                        
+
+                        width -= len;
+
                         if (width > 0 && (flags & (FLAG_ZERO|FLAG_MINUS)) == 0) {
                             buf.fill(' ',width);
                             width = 0;
@@ -1195,14 +1283,9 @@ public class Sprintf {
                         } else if ((flags & FLAG_SHARP) != 0) {
                             buf.write(args.getDecimalSeparator());
                         }
-                        buf.write(expChar); // E or e
-                        buf.write(exponent >= 0 ? '+' : '-');
-                        if (exponent < 0) {
-                            exponent = -exponent;
-                        }
-                        buf.write(exponent / 100 + '0');
-                        buf.write(exponent % 100 / 10 + '0');
-                        buf.write(exponent % 10 + '0');
+
+                        writeExp(buf, exponent, expChar);
+
                         if (width > 0) {
                             buf.fill(' ', width);
                         }
@@ -1226,8 +1309,39 @@ public class Sprintf {
                 }
             }
         } // main while loop (offset < length)
-        
-        return buf.toByteList();
+
+        // MRI behavior: validate only the unnumbered arguments
+        if ((args.numbered == 0) && args.unnumbered < args.length) {
+            if (args.runtime.getDebug().isTrue()) {
+                args.raiseArgumentError("too many arguments for format string");
+            } else if (args.runtime.getVerbose().isTrue()) {
+                args.warn(ID.TOO_MANY_ARGUMENTS, "too many arguments for format string");
+            }
+        }
+
+        return buf;
+    }
+
+    private static void writeExp(Buffer buf, int exponent, byte expChar) {
+        // Unfortunately, the number of digits in the exponent is
+        // not clearly defined in Ruby documentation. This is a
+        // platform/version-dependent behavior. On Linux/Mac/Cygwin/*nix,
+        // two digits are used. On Windows, 3 digits are used.
+        // It is desirable for JRuby to have consistent behavior, and
+        // the two digits behavior was selected. This is also in sync
+        // with "Java-native" sprintf behavior (java.util.Formatter).
+        buf.write(expChar); // E or e
+        buf.write(exponent >= 0 ? '+' : '-');
+        if (exponent < 0) {
+            exponent = -exponent;
+        }
+        if (exponent > 99) {                                
+            buf.write(exponent / 100 + '0');
+            buf.write(exponent % 100 / 10 + '0');
+        } else {
+            buf.write(exponent / 10 + '0');
+        }
+        buf.write(exponent % 10 + '0');
     }
 
     // debugging code, keeping for now
@@ -1258,8 +1372,8 @@ public class Sprintf {
         args.raiseArgumentError(message);
     }
     
-    private static final void warning(Args args, String message) {
-        args.warning(message);
+    private static final void warning(ID id, Args args, String message) {
+        args.warning(id, message);
     }
     
     private static final void checkOffset(Args args, int offset, int length, String message) {
