@@ -27,8 +27,6 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.yaml;
 
-import java.util.Date;
-import java.util.Calendar;
 import java.util.HashMap;
 import java.util.Iterator;
 import java.util.List;
@@ -39,13 +37,20 @@ import java.util.regex.Pattern;
 
 import org.jvyamlb.Composer;
 import org.jvyamlb.Constructor;
-import org.jvyamlb.ConstructorException;
+import org.jvyamlb.exceptions.ConstructorException;
 import org.jvyamlb.ConstructorImpl;
 import org.jvyamlb.SafeConstructorImpl;
+import org.jvyamlb.Scanner;
+import org.jvyamlb.ScannerImpl;
+import org.jvyamlb.ComposerImpl;
+import org.jvyamlb.ParserImpl;
+import org.jvyamlb.ResolverImpl;
+import org.jvyamlb.YAML;
 
 import org.jvyamlb.nodes.Node;
 import org.jvyamlb.nodes.LinkNode;
 
+
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
 import org.jruby.RubyClass;
@@ -55,12 +60,16 @@ import org.jruby.RubyHash;
 import org.jruby.RubyString;
 import org.jruby.RubyStruct;
 import org.jruby.RubyRange;
+import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.builtin.IRubyObject;
 
 import org.jruby.util.ByteList;
 
+import org.joda.time.DateTime;
+import org.jruby.runtime.ThreadContext;
+
 /**
  * @author <a href="mailto:ola.bini@ki.se">Ola Bini</a>
  */
@@ -96,17 +105,54 @@ public class JRubyConstructor extends ConstructorImpl {
     private final Ruby runtime;
 
     public JRubyConstructor(final IRubyObject receiver, final Composer composer) {
+        this(receiver.getRuntime(), composer);
+    }
+
+    public JRubyConstructor(final Ruby runtime, final Composer composer) {
         super(composer);
-        this.runtime = receiver.getRuntime();
+        this.runtime = runtime;
     }
 
     public Object constructRubyScalar(final Node node) {
-        ByteList sc = (ByteList)super.constructScalar(node);
-        if(sc.length() > 0 && sc.charAt(0) == ':' && ((org.jvyamlb.nodes.ScalarNode)node).getStyle() == 0) {
-            return runtime.evalScript(sc.toString());
-        }
+        if(node instanceof org.jvyamlb.nodes.ScalarNode) {
+            ByteList sc = (ByteList)super.constructScalar(node);
+            if(sc.length() > 1 && sc.charAt(0) == ':' && ((org.jvyamlb.nodes.ScalarNode)node).getStyle() == 0) {
+                int first = sc.get(1);
+                int last = sc.get(sc.realSize-1);
+                if((first == '"' && last == '"') ||
+                   (first == '\'' && last == '\'')) {
 
-        return RubyString.newString(runtime,(ByteList)super.constructScalar(node));
+                    Scanner scn = new ScannerImpl(sc.makeShared(1, sc.realSize-1));
+                    Constructor ctor = new JRubyConstructor(runtime, new ComposerImpl(new ParserImpl(scn,YAML.config().version("1.0")),new ResolverImpl()));
+                    ctor.checkData();
+                    return ((RubyString)ctor.getData()).intern();
+                }
+
+                return runtime.newSymbol(new String(sc.bytes, sc.begin+1, sc.realSize-1));
+            }
+
+            return RubyString.newString(runtime,(ByteList)super.constructScalar(node));
+        } else {
+            // Assume it's a mapping node
+
+            Map val = (Map)(constructMapping(node));
+            RubyString str = (RubyString)val.get(runtime.newString("str"));
+
+            Map props = new HashMap();
+            for(Iterator iter = val.entrySet().iterator();iter.hasNext();) {
+                Map.Entry em = (Map.Entry)iter.next();
+                if(em.getKey().toString().startsWith("@")) {
+                    props.put(em.getKey(),em.getValue());
+                    iter.remove();
+                }
+            }
+            for(Iterator iter = props.entrySet().iterator();iter.hasNext();) {
+                Map.Entry em = (Map.Entry)iter.next();
+                str.instance_variable_set((IRubyObject)em.getKey(),(IRubyObject)em.getValue());
+            }
+
+            return str;
+        }
     }
 
     public Object constructPrivateType(final Node node) {
@@ -118,7 +164,10 @@ public class JRubyConstructor extends ConstructorImpl {
         } else {
             val = constructRubyScalar(node);
         }
-        return runtime.getModule("YAML").getConstant("PrivateType").callMethod(runtime.getCurrentContext(),"new",new IRubyObject[]{runtime.newString(node.getTag()),(IRubyObject)val});
+        return RuntimeHelpers.invoke(
+                runtime.getCurrentContext(),
+                runtime.fastGetModule("YAML").fastGetConstant("PrivateType"),
+                "new", runtime.newString(node.getTag()),(IRubyObject)val);
     }
 
     public Object constructRubySequence(final Node node) {
@@ -133,7 +182,7 @@ public class JRubyConstructor extends ConstructorImpl {
                 final IRubyObject ix = runtime.newFixnum(i);
                 addFixer((Node)(((LinkNode)oo).getValue()), new RecursiveFixer() {
                         public void replace(Node node, Object real) {
-                            arr.aset(new IRubyObject[]{ix, (IRubyObject)real});
+                            arr.aset(ix, (IRubyObject)real);
                         }
                     });
             } else {
@@ -162,7 +211,7 @@ public class JRubyConstructor extends ConstructorImpl {
 
     public static Object constructYamlOmap(final Constructor ctor, final Node node) {
         Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        RubyArray arr = (RubyArray)(runtime.getModule("YAML").getConstant("Omap").callMethod(runtime.getCurrentContext(),"new"));
+        RubyArray arr = (RubyArray)(runtime.fastGetModule("YAML").fastGetConstant("Omap").callMethod(runtime.getCurrentContext(),"new"));
         List l = (List)ctor.constructSequence(node);
         ctor.doRecursionFix(node, arr);
         for(Iterator iter = l.iterator();iter.hasNext();) {
@@ -188,7 +237,7 @@ public class JRubyConstructor extends ConstructorImpl {
         final Object _str = ((JRubyConstructor)ctor).constructRubyScalar(node);
         if(_str instanceof org.jruby.RubyString) {
             final org.jruby.RubyString str = (org.jruby.RubyString)_str;
-            return (str.getValue().length() == 0 && ((org.jvyamlb.nodes.ScalarNode)node).getStyle() == 0) ? str.getRuntime().getNil() : str;
+            return (str.getByteList().realSize == 0 && ((org.jvyamlb.nodes.ScalarNode)node).getStyle() == 0) ? str.getRuntime().getNil() : str;
         }
         return _str;
     }
@@ -206,16 +255,17 @@ public class JRubyConstructor extends ConstructorImpl {
     }
 
     public static Object constructYamlTimestamp(final Constructor ctor, final Node node) {
-        Calendar c = (Calendar)SafeConstructorImpl.constructYamlTimestamp(ctor,node);
-        return org.jruby.RubyTime.newTime(((JRubyConstructor)ctor).runtime,c);
+        Object[] value = (Object[])SafeConstructorImpl.constructYamlTimestamp(ctor,node);
+        DateTime dt = (DateTime)(value[0]);
+        org.jruby.RubyTime rt = org.jruby.RubyTime.newTime(((JRubyConstructor)ctor).runtime,dt);
+        rt.setUSec(((Integer)value[1]));
+        return rt;
     }
 
     public static Object constructYamlTimestampYMD(final Constructor ctor, final Node node) {
-        Date d = (Date)SafeConstructorImpl.constructYamlTimestamp(ctor,node);
-        Calendar c = Calendar.getInstance();
-        c.setTime(d);
+        DateTime dt = (DateTime)(((Object[])SafeConstructorImpl.constructYamlTimestamp(ctor,node))[0]);
         Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        return runtime.getClass("Date").callMethod(runtime.getCurrentContext(),"new",new IRubyObject[]{runtime.newFixnum(c.get(Calendar.YEAR)),runtime.newFixnum(c.get(Calendar.MONTH)+1),runtime.newFixnum(c.get(Calendar.DAY_OF_MONTH))});
+        return RuntimeHelpers.invoke(runtime.getCurrentContext(), runtime.fastGetClass("Date"), "new", runtime.newFixnum(dt.getYear()),runtime.newFixnum(dt.getMonthOfYear()),runtime.newFixnum(dt.getDayOfMonth()));
     }
 
     public static Object constructYamlInt(final Constructor ctor, final Node node) {
@@ -237,9 +287,46 @@ public class JRubyConstructor extends ConstructorImpl {
         return SafeConstructorImpl.constructJava(ctor,pref,node);
     }
 
+    public static Object constructRubyException(final Constructor ctor, final String tag, final Node node) {
+        final Ruby runtime = ((JRubyConstructor)ctor).runtime;
+        RubyModule objClass = runtime.getObject();
+        if(tag != null) {
+            final String[] nms = tag.split("::");
+            try {
+                for(int i=0,j=nms.length;i<j;i++) {
+                    objClass = (RubyModule)objClass.getConstant(nms[i]);
+                }
+            } catch(Exception e) {
+                // No constant available, so we'll fall back on YAML::Object
+                objClass = (RubyClass)runtime.fastGetModule("YAML").fastGetConstant("Object");
+                final RubyHash vars = (RubyHash)(((JRubyConstructor)ctor).constructRubyMapping(node));
+                return RuntimeHelpers.invoke(runtime.getCurrentContext(), objClass, "new", runtime.newString(tag), vars);
+            }
+        }
+        final RubyClass theCls = (RubyClass)objClass;
+        final RubyObject oo = (RubyObject)theCls.getAllocator().allocate(runtime, theCls);
+        final Map vars = (Map)(ctor.constructMapping(node));
+        ctor.doRecursionFix(node, oo);
+        for(final Iterator iter = vars.keySet().iterator();iter.hasNext();) {
+            final IRubyObject key = (IRubyObject)iter.next();
+            final Object val = vars.get(key);
+            if(val instanceof LinkNode) {
+                final String KEY = "@" + key.toString();
+                ctor.addFixer((Node)(((LinkNode)val).getValue()), new RecursiveFixer() {
+                        public void replace(Node node, Object real) {
+                            oo.setInstanceVariable(KEY,(IRubyObject)real);
+                        }
+                    });
+            } else {
+                oo.setInstanceVariable("@" + key.toString(),(IRubyObject)val);
+            }
+        }
+        return oo;
+    }
+
     public static Object constructRubyStruct(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        RubyModule sClass = runtime.getModule("Struct");
+        RubyModule sClass = runtime.fastGetModule("Struct");
         RubyClass struct_type;
         String[] nms = tag.split("::");
         for(int i=0,j=nms.length;i<j && sClass != null;i++) {
@@ -265,7 +352,7 @@ public class JRubyConstructor extends ConstructorImpl {
                 Map.Entry em = (Map.Entry)iter.next();
                 params[i] = ((RubyString)em.getKey()).intern();
             }
-            struct_type = RubyStruct.newInstance(runtime.getModule("Struct"),params,Block.NULL_BLOCK);
+            struct_type = RubyStruct.newInstance(runtime.fastGetModule("Struct"),params,Block.NULL_BLOCK);
         } else {
             struct_type = (RubyClass)sClass;
         }
@@ -286,23 +373,28 @@ public class JRubyConstructor extends ConstructorImpl {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
         if(theCls.respondsTo("yaml_new")) {
             final RubyHash vars = (RubyHash)(((JRubyConstructor)ctor).constructRubyMapping(node));
-            return theCls.callMethod(runtime.getCurrentContext(), "yaml_new", new IRubyObject[]{theCls, runtime.newString(node.getTag()), vars});
+            return RuntimeHelpers.invoke(runtime.getCurrentContext(), theCls, "yaml_new", theCls, runtime.newString(node.getTag()), vars);
         } else {
             final RubyObject oo = (RubyObject)theCls.getAllocator().allocate(runtime, theCls);
-            final Map vars = (Map)(ctor.constructMapping(node));
-            ctor.doRecursionFix(node, oo);
-            for(final Iterator iter = vars.keySet().iterator();iter.hasNext();) {
-                final IRubyObject key = (IRubyObject)iter.next();
-                final Object val = vars.get(key);
-                if(val instanceof LinkNode) {
-                    final String KEY = "@" + key.toString();
-                    ctor.addFixer((Node)(((LinkNode)val).getValue()), new RecursiveFixer() {
-                            public void replace(Node node, Object real) {
-                                oo.setInstanceVariable(KEY,(IRubyObject)real);
-                            }
-                        });
-                } else {
-                    oo.setInstanceVariable("@" + key.toString(),(IRubyObject)val);
+            if (oo.respondsTo("yaml_initialize")) {
+                RubyHash vars = (RubyHash)(((JRubyConstructor)ctor).constructRubyMapping(node));
+                RuntimeHelpers.invoke(runtime.getCurrentContext(), oo, "yaml_initialize", runtime.newString(node.getTag()), vars);
+            } else {
+                final Map vars = (Map)(ctor.constructMapping(node));
+                ctor.doRecursionFix(node, oo);
+                for(final Iterator iter = vars.keySet().iterator();iter.hasNext();) {
+                    final IRubyObject key = (IRubyObject)iter.next();
+                    final Object val = vars.get(key);
+                    if(val instanceof LinkNode) {
+                        final String KEY = "@" + key.toString();
+                        ctor.addFixer((Node)(((LinkNode)val).getValue()), new RecursiveFixer() {
+                                public void replace(Node node, Object real) {
+                                    oo.setInstanceVariable(KEY,(IRubyObject)real);
+                                }
+                            });
+                    } else {
+                        oo.setInstanceVariable("@" + key.toString(),(IRubyObject)val);
+                    }
                 }
             }
             return oo;
@@ -311,7 +403,7 @@ public class JRubyConstructor extends ConstructorImpl {
 
     public static Object constructRuby(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        RubyModule objClass = runtime.getModule("Object");
+        RubyModule objClass = runtime.getObject();
         if(tag != null) {
             final String[] nms = tag.split("::");
             try {
@@ -320,9 +412,9 @@ public class JRubyConstructor extends ConstructorImpl {
                 }
             } catch(Exception e) {
                 // No constant available, so we'll fall back on YAML::Object
-                objClass = (RubyClass)runtime.getModule("YAML").getConstant("Object");
+                objClass = (RubyClass)runtime.fastGetModule("YAML").fastGetConstant("Object");
                 final RubyHash vars = (RubyHash)(((JRubyConstructor)ctor).constructRubyMapping(node));
-                return objClass.callMethod(runtime.getCurrentContext(), "new", new IRubyObject[]{runtime.newString(tag), vars});
+                return RuntimeHelpers.invoke(runtime.getCurrentContext(), objClass, "new", runtime.newString(tag), vars);
             }
         }
         final RubyClass theCls = (RubyClass)objClass;
@@ -332,12 +424,14 @@ public class JRubyConstructor extends ConstructorImpl {
     public static Object constructRubyRegexp(final Constructor ctor, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
         String s1 = ctor.constructScalar(node).toString();
-        return runtime.evalScript(s1);
+        // This should be fixed in some way
+        return runtime.evalScriptlet(s1);
     }
 
     public static Object constructRubyRange(final Constructor ctor, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        if(node instanceof org.jvyamlb.nodes.ScalarNode) {
+        ThreadContext context = runtime.getCurrentContext();
+        if (node instanceof org.jvyamlb.nodes.ScalarNode) {
             String s1 = ctor.constructScalar(node).toString();
             String first;
             String second;
@@ -352,22 +446,22 @@ public class JRubyConstructor extends ConstructorImpl {
                 first = s1.substring(0,ix);
                 second = s1.substring(ix+2);
             }
-            IRubyObject fist = runtime.getModule("YAML").callMethod(runtime.getCurrentContext(),"load",runtime.newString(first));
-            IRubyObject sic = runtime.getModule("YAML").callMethod(runtime.getCurrentContext(),"load",runtime.newString(second));
-            return RubyRange.newRange(runtime, fist, sic, exc);
+            IRubyObject fist = runtime.fastGetModule("YAML").callMethod(context,"load",runtime.newString(first));
+            IRubyObject sic = runtime.fastGetModule("YAML").callMethod(context,"load",runtime.newString(second));
+            return RubyRange.newRange(runtime, context, fist, sic, exc);
         } else {
             final Map vars = (Map)(ctor.constructMapping(node));
             IRubyObject beg = (IRubyObject)vars.get(runtime.newString("begin"));
             IRubyObject end = (IRubyObject)vars.get(runtime.newString("end"));
             boolean excl = ((IRubyObject)vars.get(runtime.newString("excl"))).isTrue();
-            return RubyRange.newRange(runtime, beg, end, excl);
+            return RubyRange.newRange(runtime, context, beg, end, excl);
         }
     }
 
     public static Object findAndCreateFromCustomTagging(final Constructor ctor, final Node node) {
         String tag = node.getTag();
         Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        IRubyObject _cl = runtime.getModule("YAML").callMethod(runtime.getCurrentContext(), "tagged_classes").callMethod(runtime.getCurrentContext(),"[]", runtime.newString(tag));
+        IRubyObject _cl = runtime.fastGetModule("YAML").callMethod(runtime.getCurrentContext(), "tagged_classes").callMethod(runtime.getCurrentContext(),"[]", runtime.newString(tag));
         if(!(_cl instanceof RubyClass)) {
             return null;
         }
@@ -380,7 +474,7 @@ public class JRubyConstructor extends ConstructorImpl {
 
     public static Object constructRubyInt(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        RubyModule objClass = runtime.getModule("Object");
+        RubyModule objClass = runtime.getObject();
         if(tag != null) {
             final String[] nms = tag.split("::");
             for(int i=0,j=nms.length;i<j;i++) {
@@ -394,9 +488,25 @@ public class JRubyConstructor extends ConstructorImpl {
         return oo;
     }
 
+    public static Object constructRubyString(final Constructor ctor, final String tag, final Node node) {
+        final Ruby runtime = ((JRubyConstructor)ctor).runtime;
+        RubyModule objClass = runtime.getObject();
+        if(tag != null) {
+            final String[] nms = tag.split("::");
+            for(int i=0,j=nms.length;i<j;i++) {
+                objClass = (RubyModule)objClass.getConstant(nms[i]);
+            }
+        }
+        final RubyClass theCls = (RubyClass)objClass;
+        final RubyObject oo = (RubyObject)theCls.getAllocator().allocate(runtime, theCls);
+        final IRubyObject val = (IRubyObject)constructYamlStr(ctor, node);
+        oo.callInit(new IRubyObject[]{val},org.jruby.runtime.Block.NULL_BLOCK);
+        return oo;
+    }
+
     public static Object constructRubyMap(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        RubyModule objClass = runtime.getModule("Object");
+        RubyModule objClass = runtime.getObject();
         if(tag != null) {
             final String[] nms = tag.split("::");
             for(int i=0,j=nms.length;i<j;i++) {
@@ -408,14 +518,14 @@ public class JRubyConstructor extends ConstructorImpl {
         final Map vars = (Map)(ctor.constructMapping(node));
         for(final Iterator iter = vars.keySet().iterator();iter.hasNext();) {
             final IRubyObject key = (IRubyObject)iter.next();
-            oo.callMethod(oo.getRuntime().getCurrentContext(),MethodIndex.ASET, "[]=", new IRubyObject[]{key,(IRubyObject)vars.get(key)});
+            RuntimeHelpers.invoke(oo.getRuntime().getCurrentContext(), oo, "[]=", key, (IRubyObject)vars.get(key));
         }
         return oo;
     }
 
     public static Object constructRubySequence(final Constructor ctor, final String tag, final Node node) {
         final Ruby runtime = ((JRubyConstructor)ctor).runtime;
-        RubyModule objClass = runtime.getModule("Object");
+        RubyModule objClass = runtime.getObject();
         if(tag != null) {
             final String[] nms = tag.split("::");
             for(int i=0,j=nms.length;i<j;i++) {
@@ -426,7 +536,7 @@ public class JRubyConstructor extends ConstructorImpl {
         final RubyObject oo = (RubyObject)theCls.getAllocator().allocate(runtime, theCls);
         final List vars = (List)(ctor.constructSequence(node));
         for(final Iterator iter = vars.iterator();iter.hasNext();) {
-            oo.callMethod(oo.getRuntime().getCurrentContext(),MethodIndex.OP_LSHIFT, "<<", new IRubyObject[]{(IRubyObject)iter.next()});;
+            RuntimeHelpers.invoke(oo.getRuntime().getCurrentContext(), oo, "<<", (IRubyObject)iter.next());
         }
         return oo;
     }
@@ -469,7 +579,12 @@ public class JRubyConstructor extends ConstructorImpl {
             });
         addConstructor("tag:yaml.org,2002:timestamp",new YamlConstructor() {
                 public Object call(final Constructor self, final Node node) {
-                    return constructYamlTimestamp(self,node);
+                    java.util.regex.Matcher match = SafeConstructorImpl.YMD_REGEXP.matcher(node.getValue().toString());
+                    if(match.matches()) {
+                        return constructYamlTimestampYMD(self,node);
+                    } else {
+                        return constructYamlTimestamp(self,node);
+                    }
                 }
             });
         addConstructor("tag:yaml.org,2002:timestamp#ymd",new YamlConstructor() {
@@ -521,6 +636,11 @@ public class JRubyConstructor extends ConstructorImpl {
                     return constructRubyMap(self,pref,node);
                 }
             });
+        addMultiConstructor("tag:ruby.yaml.org,2002:hash:",new YamlMultiConstructor() {
+                public Object call(final Constructor self, final String pref, final Node node) {
+                    return constructRubyMap(self,pref,node);
+                }
+            });
         addMultiConstructor("tag:yaml.org,2002:int:",new YamlMultiConstructor() {
                 public Object call(final Constructor self, final String pref, final Node node) {
                     return constructRubyInt(self,pref,node);
@@ -531,6 +651,21 @@ public class JRubyConstructor extends ConstructorImpl {
                     return constructRubySequence(self,pref,node);
                 }
             });
+        addMultiConstructor("tag:ruby.yaml.org,2002:array:",new YamlMultiConstructor() {
+                public Object call(final Constructor self, final String pref, final Node node) {
+                    return constructRubySequence(self,pref,node);
+                }
+            });
+        addMultiConstructor("tag:yaml.org,2002:str:",new YamlMultiConstructor() {
+                public Object call(final Constructor self, final String pref, final Node node) {
+                    return constructRubyString(self,pref,node);
+                }
+            });
+        addMultiConstructor("tag:ruby.yaml.org,2002:string:",new YamlMultiConstructor() {
+                public Object call(final Constructor self, final String pref, final Node node) {
+                    return constructRubyString(self,pref,node);
+                }
+            });
         addMultiConstructor("tag:yaml.org,2002:ruby/object:",new YamlMultiConstructor() {
                 public Object call(final Constructor self, final String pref, final Node node) {
                     return constructRuby(self,pref,node);
@@ -556,5 +691,10 @@ public class JRubyConstructor extends ConstructorImpl {
                     return constructRubyStruct(self,pref,node);
                 }
             });
+        addMultiConstructor("tag:ruby.yaml.org,2002:exception:",new YamlMultiConstructor() {
+                public Object call(final Constructor self, final String pref, final Node node) {
+                    return constructRuby(self,pref,node);
+                }
+            });
     }
 }// JRubyConstructor
