@@ -29,10 +29,15 @@ package org.jruby.yaml;
 
 import java.io.IOException;
 
+import java.util.Iterator;
+import java.util.Map;
+import java.util.HashMap;
+
+import org.jruby.RubyClass;
+import org.jruby.RubyModule;
 import org.jruby.RubyHash;
 import org.jruby.RubyArray;
 import org.jruby.RubyString;
-
 import org.jruby.runtime.builtin.IRubyObject;
 
 import org.jruby.javasupport.JavaEmbedUtils;
@@ -43,6 +48,7 @@ import org.jvyamlb.Representer;
 import org.jvyamlb.YAMLConfig;
 import org.jvyamlb.YAMLNodeCreator;
 import org.jvyamlb.nodes.Node;
+import org.jvyamlb.nodes.MappingNode;
 
 import org.jruby.util.ByteList;
 
@@ -95,15 +101,28 @@ public class JRubyRepresenter extends SafeRepresenterImpl {
         }
     }
 
+    @Override
+    public Node representMapping(final String tag, final Map mapping, final boolean flowStyle) throws IOException {
+        Map value = new HashMap();
+        final Iterator iter = (mapping instanceof RubyHash) ? ((RubyHash)mapping).directEntrySet().iterator() : mapping.entrySet().iterator();
+        while(iter.hasNext()) {
+            Map.Entry entry = (Map.Entry)iter.next();
+            value.put(representData(entry.getKey()),representData(entry.getValue()));
+        }
+        return new MappingNode(tag,value,flowStyle);
+    }
+
     protected boolean ignoreAliases(final Object data) {
         return (data instanceof IRubyObject && ((IRubyObject)data).isNil()) || super.ignoreAliases(data);
     }
 
     public static class IRubyObjectYAMLNodeCreator implements YAMLNodeCreator {
         private final IRubyObject data;
+        private final RubyClass outClass;
 
         public IRubyObjectYAMLNodeCreator(final Object data) {
             this.data = (IRubyObject)data;
+            this.outClass = ((RubyClass)(((RubyModule)this.data.getRuntime().getModule("YAML").getConstant("JvYAML"))).getConstant("Node"));
         }
 
         public String taguri() {
@@ -111,7 +130,9 @@ public class JRubyRepresenter extends SafeRepresenterImpl {
         }
 
         public Node toYamlNode(final Representer representer) throws IOException {
-            if(data.getMetaClass().searchMethod("to_yaml") instanceof org.jruby.internal.runtime.methods.SimpleCallbackMethod) {
+            if(data.getMetaClass().searchMethod("to_yaml") == data.getRuntime().getObjectToYamlMethod() ||
+               data.getMetaClass().searchMethod("to_yaml").isUndefined() // In this case, hope that it works out correctly when calling to_yaml_node. Rails does this.
+               ) {
                 // to_yaml have not been overridden
                 Object val = data.callMethod(data.getRuntime().getCurrentContext(), "to_yaml_node", JavaEmbedUtils.javaToRuby(data.getRuntime(),representer));
                 if(val instanceof Node) {
@@ -121,6 +142,11 @@ public class JRubyRepresenter extends SafeRepresenterImpl {
                 }
             } else {
                 IRubyObject val = data.callMethod(data.getRuntime().getCurrentContext(), "to_yaml", JavaEmbedUtils.javaToRuby(data.getRuntime(),representer));
+
+                if(!outClass.isInstance(val)) {
+                    throw val.getRuntime().newTypeError("wrong argument type " + val.getMetaClass().getRealClass() + " (expected YAML::JvYAML::Node)");
+                }
+
                 IRubyObject value = val.callMethod(data.getRuntime().getCurrentContext(),"value");
                 IRubyObject style = val.callMethod(data.getRuntime().getCurrentContext(),"style");
                 IRubyObject type_id = val.callMethod(data.getRuntime().getCurrentContext(),"type_id");
