@@ -41,10 +41,8 @@ import org.jruby.ast.ArrayNode;
 import org.jruby.ast.ConstNode;
 import org.jruby.ast.LocalVarNode;
 import org.jruby.ast.Node;
-import org.jruby.ast.OptNNode;
 import org.jruby.ast.PostExeNode;
 import org.jruby.ast.RegexpNode;
-import org.jruby.ast.visitor.rewriter.ReWriteVisitor;
 import org.jruby.lexer.yacc.IDESourcePosition;
 import org.jruby.util.ByteList;
 
@@ -75,10 +73,6 @@ public class TestReWriteVisitor extends TestCase {
         return new ReWriteVisitor(new StringWriter(), "");
     }
 
-    public void testVisitOptNNode() {
-        assertNull(getVisitor().visitOptNNode(new OptNNode(emptyPosition, null)));
-    }
-
     public void testVisitPostExeNode() {
         assertNull(getVisitor().visitPostExeNode(new PostExeNode(emptyPosition, null)));
     }
@@ -120,11 +114,10 @@ public class TestReWriteVisitor extends TestCase {
             stream.close();
             BufferedReader reader = new BufferedReader(new FileReader(fileName));
             assertEquals(reader.readLine(), testString);
+            reader.close();
         } finally {
             new java.io.File(fileName).delete();
         }
     }
 }
 
-
-
