@@ -36,6 +36,7 @@ test_equal("", s[6..2])
 t = ""
 test_equal(nil, t[6..2])
 test_equal(nil, t[-2..-4])
+test_equal("", t[0...-1])
 test_equal("ell", s[/[aeiow](.)\1/])
 test_equal("ell", s[/[aeiow](.)\1/, 0])
 test_equal("l", s[/[aeiow](.)\1/, 1])
@@ -135,6 +136,36 @@ test_equal(1, 'B' <=> 'A')
 test_equal(nil, 'A' <=> 3)
 test_equal(nil, 'A' <=> 3.to_f)
 
+##### casecmp #####
+s = 'abc'
+# compare string to self
+test_equal 0, s.casecmp(s)
+
+# another, but equal string
+test_equal 0, 'abc'.casecmp('abc')
+
+# case insensitively equal string
+test_equal 0, 'Abc'.casecmp('aBC')
+test_equal 0, 'ABc'.casecmp('ABC')
+
+# smaller string of same length
+test_equal -1, 'Abc'.casecmp('aCc')
+test_equal -1, 'aBc'.casecmp('BbC')
+test_equal 1, 'Abc'.casecmp('aAc')
+test_equal 1, 'bBc'.casecmp('AbC')
+
+# string of different length
+test_equal 1, 'Abcd'.casecmp('aBc')
+test_equal -1, 'Abc'.casecmp('aBcD')
+
+# with empty strings
+test_equal 0, ''.casecmp('')
+test_equal -1, ''.casecmp('a')
+test_equal 1, 'a'.casecmp('')
+
+# type mismatch
+test_exception(TypeError) { 'a'.casecmp(:a) }
+test_exception(TypeError) { 'foo'.casecmp(nil) }
 
 ##### <</concat ######
 s = "a"
@@ -234,6 +265,10 @@ test_exception(ArgumentError)  { "hi".rjust(11, "") }
 # oct should return zero in appropriate cases
 test_equal(0, "b".oct)
 test_equal(0, "".oct)
+test_equal(500, "0d500".oct)
+test_equal(7, "0o7".oct)
+test_equal(11, "0d11".oct)
+test_equal(255, "0377bad".oct)
 
 ##### replace #####
 t = "hello"
@@ -332,11 +367,35 @@ s = o.dup
 test_equal("Bar", s.slice!("Bar"))
 test_equal("Foo", s)
 
+s = o.dup
+res = "Exception thrown"
+test_no_exception {
+  res = s.slice!(0.5, 2.5)
+}
+test_equal("Fo", res)
+test_equal("oBar", s)
+
+s = o.dup
+obj = Object.new
+def obj.to_int() 2 end
+res = "Exception thrown"
+test_no_exception {
+  res = s.slice!(obj, obj)
+}
+test_equal("oB", res)
+test_equal("Foar", s)
+
 ##### split ######
 
 test_equal(["1", "2", "3"], "1x2x3".split('x'))
 test_equal(["1", "2", "3"], "1   2     3".split(' '))
 test_equal(["1", "2", "3"], "  1   2     3".split(' '))
+test_equal(%w(a b c), "abc".split(''))
+test_equal(%w(a), "a".split(''))
+test_equal([], "".split(''))
+test_equal([], "".split(' '))
+test_equal([], "".split('a'))
+test_equal(%w(a bc), "abc".split('',2))
 # explicit Regex will not ignore leading whitespace
 test_equal(["", "1", "2", "3"], "  1   2     3".split(/\s+/))
 test_equal(["1", "2", "3"], "1 2 3".split())
@@ -348,6 +407,7 @@ test_equal(["", "a", "b", "c"], "/a/b/c".split("/"))
 test_equal(["a", "b", "c"], "abc".split(//))
 test_equal(["/home", "/jruby"], "/home/jruby".split(%r<(?=/)>))
 test_equal(["///home", "///jruby"], "///home///jruby".split(%r<(?=///)>))
+test_equal(["run"], "run".split('_to_'))
 
 ##### sub #####
 
@@ -372,6 +432,13 @@ test_equal("AAAA0000", "ZZZ9999".succ)
 test_succ!("AAAA0000", "ZZZ9999")
 test_equal("**+", "***".succ)
 test_succ!("**+", "***")
+# JRUBY-1750
+test_equal("*10", "*9".succ)
+test_succ!("*10", "*9")
+test_equal("*AA", "*Z".succ)
+test_succ!("*AA", "*Z")
+test_equal("100*AA", "99*ZZ".succ)
+test_succ!("100*AA", "99*ZZ")
 
 ##### sum #####
 
@@ -426,7 +493,7 @@ test_equal(294977, "1100101".to_i(8))
 test_equal(1100101, "1100101".to_i(10))
 test_equal(17826049, "1100101".to_i(16))
 test_equal(199066177, "1100101".to_i(24))
-
+test_equal(500, "0d500".to_i)
 
 ##### to_sym (see intern) #####
 
@@ -452,6 +519,72 @@ test_equal("05", '%02d' % '5')
 test_equal("05", '%02g' % '5')
 test_equal("05", '%02G' % '5')
 test_equal("  ", '%2s' % nil)
+test_equal("%\000", "%\0" % [])
+test_equal("%\000x he1\005llo", "%\0x he%x\05llo" % [1])
+
+# underscore right after 0:
+test_equal("ff", '%x' % "0_3_7_7")
+test_equal("255", "%u" % "0_3_7_7")
+
+# regression tests for E/e/G/g patterns
+test_equal("1.000000E+01", "%E" % 10)
+test_equal("1.000000e+99", "%e" % 1e99)
+test_equal(" 1.000000E+100", "% E" % 1e100)
+test_equal("1.000202E+06", "%E" % 1000202)
+
+test_equal("1.000000E+05", "%E" % 1e5)
+test_equal("1.000000E+06", "%E" % 1000000)
+test_equal("1.000000e+99", "%e" % 1e99)
+test_equal(" 1.000000E+100", "% E" % 1e100)
+test_equal("1.000202E+06", "%E" % 1000202)
+test_equal("  1.00E+03", "%10.2E" % 1001)
+test_equal("1.0010E+03", "%10.4E" % 1001)
+test_equal("  1.00E+03","%#10.2E" % 1001)
+test_equal("1.00E+03", "%#7.2E" % 1001)
+test_equal("1.0E+03", "%#7.1E" % 1001)
+test_equal("2.0001E+05", "%.4E" % 200011)
+test_equal("2.0001E+05", "%#.4E" % 200011)
+test_equal("1.000E+03", "%7.3E" % 1000)
+test_equal("1.0000000000E+03", "%-#7.10E" % 1000)
+test_equal("  1.11E+03", "%10.2E" % 1111)
+test_equal("1.00E+04", "%#.2E" % 1E4)
+test_equal("1.00000E+05", "%#.5E" % 1E5)
+test_equal("1.00000E+05         ", "%-#20.5E" % 1E5)
+test_equal("         1.00000E+05", "%#20.5E" % 1E5)
+test_equal("00001.01000E+06", "%0#15.5E" % 10.1E5)
+
+test_equal("100000", "%G" % 1e5)
+test_equal("1E+06", "%G" % 1000000)
+test_equal("1e+99", "%g" % 1e99)
+test_equal(" 1E+100", "% G" % 1e100)
+test_equal("1.0002E+06", "%G" % 1000202)
+test_equal("     1E+03", "%10.2G" % 1001)
+test_equal("   1.0E+03","%#10.2G" % 1001)
+test_equal("1.0E+03", "%#7.2G" % 1001)
+test_equal(" 1.E+03", "%#7.1G" % 1001)
+test_equal("2E+05", "%.4G" % 200011)
+test_equal("2.000E+05", "%#.4G" % 200011)
+test_equal("  1E+03", "%7.3G" % 1000)
+test_equal("1000.000000", "%-#7.10G" % 1000)
+test_equal("   1.1E+03", "%10.2G" % 1111)
+test_equal("1.0E+04", "%#.2G" % 1E4)
+test_equal("1.0000E+05", "%#.5G" % 1E5)
+test_equal("1.0000E+05          ", "%-#20.5G" % 1E5)
+test_equal("          1.0000E+05", "%#20.5G" % 1E5)
+test_equal("000001.0100E+06", "%0#15.5G" % 10.1E5)
+
+# tainted args
+str = "string".taint
+def str.inspect() "string".taint end
+test_equal(true, ("%s" % str).tainted?)
+test_equal(true, ("%p" % str).tainted?)
+
+test_equal(true, "hello".taint.crypt("other").tainted?)
+test_equal(true, "hello".crypt("other".taint).tainted?)
+test_equal(true, "hello".taint.dump.tainted?)
+test_equal(true, "hello".taint.inspect.tainted?)
+test_equal(true, ("%s".taint % "hello").tainted?)
+test_equal(true, "hello".taint.inspect.tainted?)
 
 # test that extensions of the base classes are typed correctly
 class StringExt < String
@@ -467,6 +600,12 @@ test_equal(2, StringExt.new[2])
 
 test_equal("foa3VCPbMb8XQ", "foobar".crypt("foo"))
 
+# regression test for String#crypt
+test_no_exception {
+  "hello".crypt("\x80\x80")
+  "hello".crypt("\xFF\xFF")
+}
+
 test_exception(TypeError) { "this" =~ "that" }
 
 # UTF behavior around inspect, to_s, and split
@@ -495,6 +634,44 @@ $KCODE = old_code
 # unpack("U*") should raise ArgumentError when the string is not valid UTF8
 test_exception(ArgumentError) { x2.unpack("U*") }
 
+# unpack with Q/q patterns
+test_equal([18446744073709551615], ("\xFF" * 8).unpack('Q5'))
+test_equal([-1, nil, nil, nil, nil], ("\xFF" * 8).unpack('q5'))
+
+# unpack with Z* and Z patterns
+test_equal(["abc ", "abc "],    "abc \0abc \0".unpack('Z*Z*'))
+test_equal(["abc "],            "abc \0abc \0".unpack('Z10'))
+test_equal(["abc ", "c "],      "abc \0abc \0".unpack('Z7Z*'))
+test_equal(["abc ", ""],        "abc \0abc \0".unpack('Z50Z*'))
+test_equal(["abc ", ""],        "abc \0\0\0abc \0".unpack('Z*Z*'))
+test_equal(["abc "],            "abc \0\0\0\0".unpack('Z*'))
+test_equal(["abc ", ""],        "abc \0\0\0\0".unpack('Z*Z*'))
+test_equal([""],                "\0".unpack('Z*'))
+test_equal([""],                "\0\0".unpack('Z*'))
+test_equal([""],                "\0\0abc".unpack('Z*'))
+test_equal([""],                "\0\0abc\0\0".unpack('Z*'))
+
+# unpack with X* pattern
+test_equal(["a", "b", "b"], "abc".unpack("aaX*a"))
+test_exception(ArgumentError) { "a".unpack("X*") }
+
+# pack/unpack with "U" pattern
+numbers = [0, 1, 15, 16, 127, 127, 255, 1024, 2048, 2**16 -1, 2**16, 2**16 + 1, 2**30]
+numbers.each do |n|
+  test_equal([n], [n].pack("U").unpack("U"))
+end
+
+# unpack with "@" pattern
+test_equal(['a'], "abcdef".unpack("@5@a"))
+test_equal(['a', 'd'], "abcd".unpack("a@*a"))
+test_equal(['a'], "abcdef".unpack("@-777a"))
+test_exception(ArgumentError) { "a".unpack("@10") }
+
+# unpack with "M" pattern
+test_equal(["", ""], "=5".unpack('Ma'))
+test_equal(["abc"], "abc=".unpack('M'))
+test_equal(["a", "*", "", ""], "a=*".unpack('MMMM'))
+
 # and just for kicks, make sure we're returning appropriate byte values for each_byte!
 
 bytes = []
@@ -543,6 +720,20 @@ f = FooStr.new("AAAA")
 
 test_equal("AAAA", [f].join(','))
 
+# multiplication should return proper subclass
+class MyTestString < String; end
+my_str = MyTestString.new
+test_equal(my_str.class, (my_str * 3).class)
+
+# dump should return proper subclass
+test_equal(my_str.class, my_str.dump.class)
+
+# inspect should *not* return subclass
+test_equal(String, my_str.inspect.class)
+
+# crypt should *not* return subclass
+test_equal(String, my_str.crypt("aa").class)
+
 # test coercion for multiple methods
 class Foo
   def to_int
@@ -621,3 +812,24 @@ test_exception(TypeError) { "s" << 256 }
 test_equal("s\001", "s" << 1)
 test_exception(NoMethodError) { +"s" }
 
+# JRUBY-1195
+"%mhello".scan(/%m/) { |v| "m" =~ /x/ }
+test_equal("hello", $');
+
+class Foo
+    def to_s
+        nil
+    end
+end
+
+test_exception(TypeError){String(Foo.new)}
+
+test_no_exception do
+  a = "abc"
+  b = a[2..2]
+  b.each { |c| c + "" }
+end
+
+# JRUBY-1778
+"hello.".gsub('.', "!")
+test_ok("hello".match('.') != nil, "match after gsub must not return nil")
